/* $Header: /home/dl2/nygren/cvs/pan/src/applib/appcodeobj.c,v 1.4 1997/12/10 06:58:02 nygren Exp $
 *
 * FILE:    appcodeobj.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for reading and generating PAN code objects from and
 * to software segment streams.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/pancodeobj.h>
#include <common/sha1.h>

#include <applib/appcodeobj.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <unistd.h>


static int doesfileexist(char *basename, char *prefix, char *suffix);


/* Searches for and loads a PAN code object with the given filename.
 * Searches in the current directory, followed by the 
 * PAN_CODEOBJ_PATH environmental variable.
 * The returned panAppCodeobj should be added to a swseg ring (or should
 * otherwise be unreferenced when no longer in use).
 */
panAppCodeobj *
panAppCodeobj_load(char *filename)
{
  char *prefix, *suffix, *pan_codeobj_path;
  char *pathcomp, *copath=NULL;	/* A directory in the path */
  char infilename[MAXNAMLEN];
  int infd, filelen;
  panSwseg *cosws;
  
  prefix = "."; suffix="";
  if ((filelen = doesfileexist(filename,prefix,suffix)) > 0)
    goto load_found_file;
       
  suffix = panAppCodeobj_FILE_EXTENSION;
  if ((filelen = doesfileexist(filename,prefix,suffix)) > 0)
    goto load_found_file;

  pan_codeobj_path = getenv("PAN_CODEOBJ_PATH");
  if (!pan_codeobj_path) return NULL;
  panPrint_debug("panAppCodeobj_load: Looking through path <%s>\n",
		 pan_codeobj_path);
  copath = panMemory_alloc(strlen(pan_codeobj_path)+1);
  strcpy(copath, pan_codeobj_path);
  pathcomp = strtok(copath, ":");
  while (pathcomp)
    {
      panPrint_debug("panAppCodeobj_load: looking at <%s>\n", pathcomp);
      prefix = pathcomp; suffix="";
      if ((filelen = doesfileexist(filename,prefix,suffix)) > 0)
	goto load_found_file;
      
      suffix = panAppCodeobj_FILE_EXTENSION;
      if ((filelen = doesfileexist(filename,prefix,suffix)) > 0)
	goto load_found_file;
      
      pathcomp = strtok(NULL, ":");
    }

  if (copath) panMemory_free(copath);

  panPrint_debug("panAppCodeobj_load: couldn't find <%s>\n", filename);

  return NULL;

load_found_file:

  strcpy(infilename,prefix);
  strcat(infilename,"/");
  strcat(infilename,filename);
  strcat(infilename,suffix);
  panPrint_debug("panAppCodeobj_load: loading <%s>\n", infilename);
  
  /* Now open and create a swseg by mmapping in the file */
  infd = open(infilename, O_RDONLY);
  if (-1 == infd)
    { perror("open"); 
    panPrint_warn("Couldn't open <%s> for reading\n", infilename); 
    panMemory_free(copath);
    return NULL; }

  cosws = panSwseg_create_byMmap(filelen,
				 PROT_READ, MAP_PRIVATE, infd, 0);
  if (!cosws) 
    { panPrint_warn("Couldn't mmap <%s>\n", infilename); 
    panMemory_free(copath);
    return NULL; }

  close(infd);			/* doesn't need to stay open to 
				 * keep the mapping */
  
  if (copath) panMemory_free(copath);

  return (panAppCodeobj*)cosws;
}


/* Inserts a code object into a node.  If pin is 1, pins the code object
 * into the node's code cache.  Returns how many bytes were sent.  */
int
panAppCodeobj_insertIntoNode(panNodeLink *nl, panAppCodeobj *aco, int pin)
{
  static panAppCodeobj *coinsertion_co;
  static panCodeobjName coinsertion_name;
  panNodeAddr lna;
  panSwseg *sws;
  panStream _ps, *ps = &_ps;
  int caplen;

  if (!coinsertion_co) 
    {
      coinsertion_co = panAppCodeobj_load("insertco");
      if (!coinsertion_co) 
	{
	  panPrint_error("panAppCodeobj_insertIntoNode:"
			 " couldn't load insertco\n");
	  return 0;
	}
      panAppCodeobj_getCodeobjName(&coinsertion_name, coinsertion_co);
    }

  sws = panSwseg_create_byMalloc(panCodeobjHead_SIZE+4+panNodeAddr_SIZE);
  panSwseg_chain(sws, aco);
  panStream_init(ps, sws, 0);
  
  panCodeobjName_write(ps, &coinsertion_name);
  panNodeLink_getLocalAddr(&lna, nl);
  panNodeAddr_write(ps, &lna);
  panStream_write_net_int32(ps, pin);
  panNodeAddr_write(ps, &lna);

  panStream_init(ps, sws, 0);
  caplen = panStream_getBytesRemaining(ps);
  panStream_init(ps, sws, 0);

  return panNodeLink_send(nl, ps, caplen);
}

/* Looks for a codeobj name in aco and stores it into coname.  
 * Returns 1 on success, 0 on failure. */
int
panAppCodeobj_getCodeobjName(panCodeobjName *coname, panAppCodeobj *aco)
{
  panStream ps;

  if (!coname || !aco) return 0;
  panStream_init(&ps, aco, panCodeobjHead_NAME_OFF);
  return (panCodeobjName_SIZE == panCodeobjName_read(coname, &ps, 1));
}


/* Returns -1 if a file doesn't exist.  Otherwise, 
 * returns the length of the file in bytes.  Uses a filename
 * composed from "prefix"+"/"+"basename"+"suffix" */
static int
doesfileexist(char *basename, char *prefix, char *suffix)
{
  char filename[MAXNAMLEN];
  struct stat statbuf;	

  strcpy(filename,prefix);
  strcat(filename,"/");
  strcat(filename,basename);
  strcat(filename,suffix);
  
  panPrint_debug("doesfileexist: looking for <%s>\n", filename);

  if (0 != stat(filename, &statbuf))
    return -1;
  else 
    return statbuf.st_size;
}
