/* $Header: /home/dl2/nygren/cvs/pan/src/applib/arch/linux2/nodelink.c,v 1.1 1997/12/18 02:33:20 nygren Exp $
 *
 * FILE:    nodelink.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides an application interface to connect an application to a
 * pan KNode.  Contains routines for connecting an app to a PAN node
 * and for sending/receiving capsules.
 *
 * We're currently using AF_PAN sockets.
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/conventions.h>
#include <common/nodeaddr.h>
#include <applib/nodelink.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>


/* The nodelink structure specific to this architecture (move to
 * a header at some point?) */
struct _panNodeLink 
{
  int sock;			/* The AF_PAN socket bound
				 * to the application */
  int status;			/* Is this link connected? */
#define panNodeLink_STATUS_CONNECTED    0
#define panNodeLink_STATUS_DISCONNECTED 1
  panNodeAddr        localaddr;	/* Address of the local node */
};


/* For filling in writes */
#define filler_data_size 1024
static u_int8_t filler_data[1024];

/* For constructing an iovec for sending data to
 * a node.  We use a static structure to avoid
 * having to allocate one each time, especially since this
 * is much larger than we need in order to be safe. */
#define DATA_SEND_IOVEC_SIZE 128
static struct iovec data_send_iovec[DATA_SEND_IOVEC_SIZE];
static int data_send_iovec_pos; /* The next free slot in data_send_iovec */


/* Writes length bytes from a swseg stream into an iovec by making
 * slots in the iovec point to regions of memory.  Incriments
 * iovpos, which contains the next free slot location.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns 0 on error.
 */
panStream_GEN_SCAN_FN(
      static panStream_write_toIOVec(struct iovec *iov, int *iovpos,
				     int iovsize,
				     panStream *ps, unsigned length),
      { if (!iov || !iovpos || !ps) return 0; },
      {
	if (*iovpos>=iovsize) 
	  { panPrint_error("panStream_write_toIOVec: overflow of iovec\n");
	  return 0; }
	iov[*iovpos].iov_base = ps->pos;
	iov[*iovpos].iov_len  = numscan;
	*iovpos += 1;
      })

/* Reads length bytes from a socket into a swseg stream using read.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns -1 on error.
 */
panStream_GEN_SCAN_FN(
      static panStream_read_fromSock(panStream *ps, int sock, unsigned length),
      { if ((sock<0) || !ps) return 0; },
      {	
	int toread=numscan; 
	int readhere=0; 
	int bread=0;
        while (toread>0) {
	  bread = read(sock, ps->pos+readhere, toread);
	  if (bread<0) 
	    { perror("read"); return -1; }
	  toread-=bread;  readhere+=bread;
	} })


/* Creates a socket connected to a pan node in the kernel.
 * nodename is currently ignored.
 */
panNodeLink *
panNodeLink_create(char *nodename)
{
  panNodeLink *nl;
  struct sockaddr_pan spanaddr;
  int spanaddr_len;

  nl = (panNodeLink*)panMemory_alloc(sizeof(panNodeLink));
  if (!nl)
    { panPrint_error("panNodeLink_create: couldn't alloc memory\n");
    return NULL; }

  if ((nl->sock = socket(AF_PAN, SOCK_STREAM, 0)) < 0)
    { perror("socket");
      panPrint_error("panNodeLink_create: couldn't create socket\n");
      panMemory_free(nl);
      return NULL; }

  spanaddr_len = sizeof(struct sockaddr_pan);
  if (getsockname(nl->sock, (struct sockaddr*)&spanaddr, 
		  &spanaddr_len) != 0)
    { 
      perror("getsockname");
      panPrint_error("panNodeLink_create: couldn't get socket's name!\n");
      close(nl->sock); 
      panMemory_free(nl);
      return NULL; 
    }
  panNodeAddr_dup(&(nl->localaddr), &(spanaddr.span_addr));
  panPrint_debug("panNodeLink_create: local addr set to ");
  panNodeAddr_print(&(nl->localaddr));   panPrint_debug("\n");
  return nl;
}


/* Inserts a capsule that is length bytes long and starts at datastart 
 * into the node at the other end of the NodeLink.  Advances datastart
 * and returns the number of bytes delivered or -1 if an error occurs.
 */
int
panNodeLink_send(panNodeLink *nl, panStream *datastart, int length)
{
  int nsent, tosend = length, tosendthistime, nstreamadv, totalsending=0;
  
    if (length == 0) return 0;

  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_send: passed invalid nodelink\n");
    return -1; }

  data_send_iovec_pos = 0;
  
  nsent = panStream_write_toIOVec(data_send_iovec, &data_send_iovec_pos,
				  DATA_SEND_IOVEC_SIZE, datastart, length);
  totalsending += nsent;
  nstreamadv = nsent;
  tosend = length-nsent;

  /* If the stream was shorter than length, fill the rest with garbage
   * to avoid desyncronizing. */
  while (tosend>0)
    {
      tosendthistime = tosend>filler_data_size?filler_data_size:tosend;
      
      panPrint_debug("SENDING %d (filler)\n", tosendthistime);

      if (data_send_iovec_pos >= DATA_SEND_IOVEC_SIZE)
	{ panPrint_error("panNodeLink_send: iovec overflow\n");
	return -1; }
      data_send_iovec[data_send_iovec_pos].iov_base = filler_data;
      data_send_iovec[data_send_iovec_pos].iov_len  = tosendthistime;
      totalsending += tosendthistime;
      data_send_iovec_pos++;
    }

  panPrint_debug("panNodeLink_send: sending %d bytes in %d vecs\n",
		 totalsending, data_send_iovec_pos);
  nsent = writev(nl->sock, data_send_iovec, data_send_iovec_pos);
  if (nsent != totalsending)
    { perror ("write");
    panPrint_error("panNodeLink_send: error sending capsule body\n");
    panNodeLink_disconnect(nl);
    nl->status = panNodeLink_STATUS_DISCONNECTED;
    return -1; }
  

  return nstreamadv;
}



/* Retrieves a datagram delivered to the app by a capsule in the node.
 * Will block until a full capsule is received if block is non-zero.
 * If part of a capsule has already been read with
 * panNodeLink_read, will return the rest of it. 
 * Returns NULL on error. */
panSwseg *
panNodeLink_retr(panNodeLink *nl, int block)
{
  int navail, capsize;			/* Bytes available to be read */
  int nread;			/* Number of bytes read */
  panSwseg *capdata;
  
  panPrint_debug("panNodeLink_retr: entered\n");

  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_retr: passed invalid applink\n");
    return NULL; }

  if (ioctl (nl->sock, FIONREAD, (void *) &navail) == -1) 
    { perror ("ioctl");
    panPrint_error("panNodeLink_retr: failed to FIONREAD\n");
    panNodeLink_disconnect(nl);
    return NULL; }
  panPrint_debug("panNodeLink_retr: FIONREAD returned %d\n", navail);

  if (!navail)
    {
      fd_set readfds;

      if (!block) return 0;
      /* else, wait for data to become available */
      do {
	FD_ZERO(&readfds);
	FD_SET(nl->sock, &readfds);
	if (-1 == select(nl->sock+1, &readfds, NULL, NULL, NULL)) 
	  { perror("select");
	    panPrint_warn("panNodeLink_retr: select returned an error\n");
	    return NULL; }
	if (ioctl (nl->sock, FIONREAD, (void *) &navail) == -1) 
	  { perror ("ioctl");
	  panPrint_error("panNodeLink_retr: failed to FIONREAD\n");
	  panNodeLink_disconnect(nl);
	  return NULL; }	
      } while (!navail);      
    }

  /* If we're here, then data is available.  Now we just need
   * to read navail bytes into a capsule */
  capsize = navail;

  capdata = panSwseg_create_byMalloc(capsize);
  if (!capdata)
    { panPrint_error("panNodeLink_retr: couldn't create swseg\n");
    return NULL; }

  nread = read(nl->sock, capdata->data, capsize);

  if ((nread == -1) || (nread != capsize))
    { perror ("read");
    panPrint_error("panNodeLink_retr: error reading capdata."
		   "  Expected %d, got %d.\n", capsize, nread);
    panSwseg_unref(capdata);
    panNodeLink_disconnect(nl);
    return NULL; }

  panPrint_debug("panNodeLink_retr: returning data of size %d\n", capsize);

  return capdata;
}


/* Reads data delivered by a capsule into a swseg stream.
 * Will block until a nbytes are read.
 * Returns the number of bytes read or -1 on error. */
int
panNodeLink_read(panStream *ps, panNodeLink *nl, int nbytes)
{
  int toread;
  int nread;			/* Number of bytes read */

  return 0;			/* Not yet implemented */

  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_retr: passed invalid applink\n");
    return 0; }

  nread = panStream_read_fromSock(ps, nl->sock, toread);
  if (nread == -1)
    { 
      perror ("read");
      panPrint_error("panNodeLink_retr: error reading capsize\n");
      panNodeLink_disconnect(nl);
      return -1; 
    }	  
  return nread;
}


/* Returns 1 if a panNodeLink is still connected.
 * Returns 0 if it is no longer connected. */
int 
panNodeLink_isConnected(panNodeLink *nl)
{
  return nl->status == panNodeLink_STATUS_CONNECTED;
}


/* Returns the socket file descriptor associated with a panNodeLink
 * (or -1 if this doing this is not supported for this architecture).
 */
int 
panNodeLink_getSockFD(panNodeLink *nl)
{
  return nl->sock;
}

/* Returns the local address of the node at the other end of the applink.
 */
int 
panNodeLink_getLocalAddr(panNodeAddr *na, panNodeLink *nl)
{
  memcpy(na, &nl->localaddr, sizeof(panNodeAddr));
  return 1;
}


/* Disconnects a nodelink. */
void
panNodeLink_disconnect(panNodeLink *nl)
{
  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_disconnect: passed invalid applink\n");
    return; }

  close(nl->sock);
  nl->status = panNodeLink_STATUS_DISCONNECTED;
}


/* Destroys an applink (and frees assocated memory).  It must
 * have been disconnected first. */
void
panNodeLink_destroy(panNodeLink *nl)
{
  if (!nl)
    { panPrint_error("panNodeLink_destroy: passed invalid applink\n");
    return; }
  if (nl->status == panNodeLink_STATUS_CONNECTED)
    panNodeLink_disconnect(nl);
  panMemory_free(nl);
}




