/* $Header: /home/dl2/nygren/cvs/pan/src/applib/arch/unixuser/nodelink.c,v 1.5 1997/12/13 05:10:23 nygren Exp $
 *
 * FILE:    nodelink.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides an application interface to connect an application to a
 * pan UNode.  Contains routines for connecting an app to a PAN node
 * and for sending/receiving capsules.
 *
 * We're currently using UNIX domain socket streams where
 * a 4-byte integer capsule length is sent, followed by the capsule
 * itself (this repeats).  The first thing received is the node
 * sending its address (of size panNodeAddr_SIZE).
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/uio.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/conventions.h>
#include <common/nodeaddr.h>
#include <applib/nodelink.h>

#include <sys/un.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>


/* The nodelink structure specific to this architecture (move to
 * a header at some point?) */
struct _panNodeLink 
{
  int sock;			/* The UNIX domain socket bound
				 * to the application */
  int status;			/* Is this link connected? */
#define panNodeLink_STATUS_CONNECTED    0
#define panNodeLink_STATUS_DISCONNECTED 1
  
  int capsize;   		/* The number of bytes coming through
				 * in the next capsule from the application.
				 * If the next thing to do is to read
				 * another 4-byte capsule length, then has
				 * the value -1.  */
  struct sockaddr_un nodesockaddr; /* Contains the address (filename) of the 
				    * node connected at the
				    * other end of the socket */
  struct sockaddr_un appsockaddr; /* Contains the address (filename) of the 
				   * app connected at this
				   * end of the socket */
  panNodeAddr        localaddr;	/* Address of the local node */
};


/* For filling in writes */
#define filler_data_size 1024
static u_int8_t filler_data[1024];

/* For constructing an iovec for sending data to
 * a node.  We use a static structure to avoid
 * having to allocate one each time, especially since this
 * is much larger than we need in order to be safe. */
#define DATA_SEND_IOVEC_SIZE 128
static struct iovec data_send_iovec[DATA_SEND_IOVEC_SIZE];
static int data_send_iovec_pos; /* The next free slot in data_send_iovec */


/* Writes length bytes from a swseg stream into an iovec by making
 * slots in the iovec point to regions of memory.  Incriments
 * iovpos, which contains the next free slot location.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns 0 on error.
 */
panStream_GEN_SCAN_FN(
      static panStream_write_toIOVec(struct iovec *iov, int *iovpos,
				     int iovsize,
				     panStream *ps, unsigned length),
      { if (!iov || !iovpos || !ps) return 0; },
      {
	if (*iovpos>=iovsize) 
	  { panPrint_error("panStream_write_toIOVec: overflow of iovec\n");
	  return 0; }
	iov[*iovpos].iov_base = ps->pos;
	iov[*iovpos].iov_len  = numscan;
	*iovpos += 1;
      })

/* Reads length bytes from a socket into a swseg stream using read.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns 0 on error.
 */
panStream_GEN_SCAN_FN(
      static panStream_read_fromSock(panStream *ps, int sock, unsigned length),
      { if ((sock<0) || !ps) return 0; },
      {	if (numscan != read(sock, ps->pos, numscan))
	{ perror("read"); return 0; }
      })



/* Connects to a PAN Node that is listening on a socket 
 *        /tmp/.PAN-unix/node.$portnum
 * Nodename should be the $portnum of the node to connect to.
 * (A string is used to allow different naming schemes to be used 
 * depending on the architecture type).
 */
panNodeLink *
panNodeLink_create(char *nodename)
{
  char sockfname[UNIX_PATH_MAX]; /* The name of the socket to listen on */
  char nodeaddrname[512];
  int  bread, portnum;
  panNodeLink *nl;

  nl = (panNodeLink*)panMemory_alloc(sizeof(panNodeLink));
  if (!nl)
    { panPrint_error("panNodeLink_create: couldn't alloc memory\n");
    return NULL; }

  portnum = atoi(nodename);
  if (portnum <= 0)
    { panPrint_error("panNodeLink_create: invalid nodename/portnumber %s\n",
		     nodename);
    return NULL; }

  snprintf(sockfname, UNIX_PATH_MAX-1,
	   "%s/node.%d", panNodeAppIF_SOCKDIR_NAME, portnum);
  nl->nodesockaddr.sun_family = AF_UNIX;
  strcpy(nl->nodesockaddr.sun_path, sockfname);

  snprintf(sockfname, UNIX_PATH_MAX-1,
	   "%s/app.%d.XXXXXX", panNodeAppIF_SOCKDIR_NAME, portnum);
  if (!mktemp(sockfname)) { perror("mktemp"); return NULL; }
  unlink(sockfname);
  nl->appsockaddr.sun_family = AF_UNIX;
  strcpy(nl->appsockaddr.sun_path, sockfname);

  if ((nl->sock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
    { perror("socket");
      panPrint_error("panNodeLink_create: couldn't create socket\n");
      panMemory_free(nl);
      return NULL; }
  
  if (bind(nl->sock, (struct sockaddr*)&nl->appsockaddr, sizeof(struct sockaddr_un)) < 0)
    { perror("bind");
      panPrint_error("panNodeLink_create: couldn't bind to socket %s\n",
		     nl->appsockaddr.sun_path);
      close(nl->sock); panMemory_free(nl);
      return NULL; }
    
  if (connect(nl->sock, (struct sockaddr*)&nl->nodesockaddr, sizeof(struct sockaddr_un)) < 0)
    { perror("connect");
      panPrint_error("panNodeLink_create: couldn't connect to socket %s\n",
		     nl->nodesockaddr.sun_path);
      close(nl->sock); panMemory_free(nl);
      return NULL; }

  bread = read(nl->sock,  &nl->localaddr, sizeof(nl->localaddr));
  if (bread != sizeof(nl->localaddr))
    { perror("read");
    panPrint_error("panNodeLink_create: couldn't read localaddr from socket %s -- needed %d, got %d\n",
		     nl->nodesockaddr.sun_path, sizeof(nl->localaddr), bread);
      close(nl->sock); panMemory_free(nl);
      return NULL; }
  panNodeAddr_stringify(nodeaddrname, 512, &nl->localaddr);
  panPrint_debug("panNodeLink_create: connected to %s\n", nodeaddrname);

  nl->capsize = -1;

  return nl;
}


/* Inserts a capsule that is length bytes long and starts at datastart 
 * into the node at the other end of the NodeLink.  Advances datastart
 * and returns the number of bytes delivered or -1 if an error occurs.
 */
int
panNodeLink_send(panNodeLink *nl, panStream *datastart, int length)
{
  int nsent, tosend = length, tosendthistime, nstreamadv, totalsending=0;
  u_int32_t datlen=length;
  
    if (length == 0) return 0;

  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_send: passed invalid nodelink\n");
    return -1; }

  panPrint_debug("SENDING %d (val=%d)\n", 4, datlen);
  nsent = write(nl->sock, &datlen, 4);
  if (nsent != 4)
    { perror ("write");
    panPrint_error("panNodeLink_send: error sending datlen \n");
    panNodeLink_disconnect(nl);
    return -1; }

  data_send_iovec_pos = 0;
  
  nsent = panStream_write_toIOVec(data_send_iovec, &data_send_iovec_pos,
				  DATA_SEND_IOVEC_SIZE, datastart, length);
  totalsending += nsent;
  nstreamadv = nsent;
  tosend = length-nsent;

  /* If the stream was shorter than length, fill the rest with garbage
   * to avoid desyncronizing. */
  while (tosend>0)
    {
      tosendthistime = tosend>filler_data_size?filler_data_size:tosend;
      
      panPrint_debug("SENDING %d (filler)\n", tosendthistime);

      if (data_send_iovec_pos >= DATA_SEND_IOVEC_SIZE)
	{ panPrint_error("panNodeLink_send: iovec overflow\n");
	return -1; }
      data_send_iovec[data_send_iovec_pos].iov_base = filler_data;
      data_send_iovec[data_send_iovec_pos].iov_len  = tosendthistime;
      totalsending += tosendthistime;
      data_send_iovec_pos++;
    }

  panPrint_debug("panNodeLink_send: sending %d bytes in %d vecs\n",
		 totalsending, data_send_iovec_pos);
  nsent = writev(nl->sock, data_send_iovec, data_send_iovec_pos);
  if (nsent != totalsending)
    { perror ("write");
    panPrint_error("panNodeLink_send: error sending capsule body\n");
    panNodeLink_disconnect(nl);
    nl->status = panNodeLink_STATUS_DISCONNECTED;
    return -1; }
  

  return nstreamadv;
}



/* Retrieves a datagram delivered to the app by a capsule in the node.
 * Will block until a full capsule is received if block is non-zero.
 * If part of a capsule has already been read with
 * panNodeLink_read, will return the rest of it. 
 * Returns NULL on error. */
panSwseg *
panNodeLink_retr(panNodeLink *nl, int block)
{
  int navail;			/* Bytes available to be read */
  int nread;			/* Number of bytes read */
  panSwseg *capdata;

  panPrint_debug("panNodeLink_retr: entered\n");

  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_retr: passed invalid applink\n");
    return NULL; }

  if (ioctl (nl->sock, FIONREAD, (void *) &navail) == -1) 
    { perror ("ioctl");
    panPrint_error("panNodeLink_retr: failed to FIONREAD\n");
    return NULL; }

  if (nl->capsize == -1)
    {
      panPrint_debug("panNodeLink_retr: about to read capsize\n");
      if ((navail < 4) && !block)
	return NULL;
      if (read(nl->sock, &nl->capsize, 4) != 4) 
	{ perror ("read");
	panPrint_error("panNodeLink_retr: error reading capsize\n");
	panNodeLink_disconnect(nl);
	return NULL; }
      navail -= 4;
    }

  panPrint_debug("panNodeLink_retr: capsize %d, navail %d\n", 
		 nl->capsize, navail);

  if (nl->capsize < 0)
    { panPrint_error("panNodeLink_retr: capsize shouldn't be %d\n", 
		     nl->capsize);
    return NULL; }

  if ((navail < nl->capsize) && !block)
    return NULL;

  capdata = panSwseg_create_byMalloc(nl->capsize);
  if (!capdata)
    { panPrint_error("panNodeLink_retr: couldn't create swseg\n");
    return NULL; }

  nread = read(nl->sock, capdata->data, nl->capsize);

  if ((nread == -1) || (nread != nl->capsize))
    { perror ("read");
    panPrint_error("panNodeLink_retr: error reading capdata."
		   "  Expected %d, got %d.\n", nl->capsize, nread);
    panSwseg_unref(capdata);
    panNodeLink_disconnect(nl);
    return NULL; }

  panPrint_debug("panNodeLink_retr: returning data of size %d\n", nl->capsize);

  nl->capsize = -1;
  
  return capdata;
}


/* Reads data delivered by a capsule into a swseg stream.
 * Will block until a nbytes are read.
 * Returns the number of bytes read). */
int
panNodeLink_read(panStream *ps, panNodeLink *nl, int nbytes)
{
  int toread;
  int nread, totalread=0;			/* Number of bytes read */

  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_retr: passed invalid applink\n");
    return 0; }

  while (nbytes>0)
    {
      if (nl->capsize == -1)
	{
	  if (read(nl->sock, &nl->capsize, 4) != 4) 
	    { perror ("read");
	    panPrint_error("panNodeLink_retr: error reading capsize\n");
	    panNodeLink_disconnect(nl);
	    return totalread; }	  
	}
      
      if (nl->capsize == 0) 
	{
	  nl->capsize = -1;
	  continue;
	}

      toread = (nl->capsize>nbytes)?(nbytes):(nl->capsize); /* min */
      
      nread = panStream_read_fromSock(ps, nl->sock, toread);
      if (nread == -1)
	{ perror ("read");
	panPrint_error("panNodeLink_retr: error reading capsize\n");
	panNodeLink_disconnect(nl);
	return totalread; }	  

      totalread += nread;
      nbytes -= nread;
      nl->capsize -= nread;

      if (nread < toread)
	{ perror ("read");
	panPrint_error("panNodeLink_retr: error reading capsize\n");
	panNodeLink_disconnect(nl);
	return totalread; }	  

      if (nl->capsize == 0) nl->capsize = -1;
    }

  return totalread;
}


/* Returns 1 if a panNodeLink is still connected.
 * Returns 0 if it is no longer connected. */
int 
panNodeLink_isConnected(panNodeLink *nl)
{
  return nl->status == panNodeLink_STATUS_CONNECTED;
}


/* Returns the socket file descriptor associated with a panNodeLink
 * (or -1 if this doing this is not supported for this architecture).
 */
int 
panNodeLink_getSockFD(panNodeLink *nl)
{
  return nl->sock;
}

/* Returns the local address of the node at the other end of the applink.
 */
int 
panNodeLink_getLocalAddr(panNodeAddr *na, panNodeLink *nl)
{
  memcpy(na, &nl->localaddr, sizeof(panNodeAddr));
  return 1;
}


/* Disconnects a nodelink. */
void
panNodeLink_disconnect(panNodeLink *nl)
{
  if (!nl || (nl->status == panNodeLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeLink_disconnect: passed invalid applink\n");
    return; }

  close(nl->sock);
  nl->status = panNodeLink_STATUS_DISCONNECTED;
}


/* Destroys an applink (and frees assocated memory).  It must
 * have been disconnected first. */
void
panNodeLink_destroy(panNodeLink *nl)
{
  if (!nl)
    { panPrint_error("panNodeLink_destroy: passed invalid applink\n");
    return; }
  if (nl->status == panNodeLink_STATUS_CONNECTED)
    panNodeLink_disconnect(nl);
  unlink(nl->appsockaddr.sun_path);
  panMemory_free(nl);
}




