/* $Header: /home/dl2/nygren/cvs/pan/src/apps/flood/floodcodeobj.c,v 1.2 1998/02/18 20:00:59 nygren Exp $
 *
 * FILE:    floodcapsule.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Rudimentry flood capsule for experimental use.
 * See floodcap.psg for the specification of the capsule contents.
 *
 * Flood works by sending a flood of capsules to a destination node.
 * The first capsule in the flood indicates that record keeping should
 * be started.  Subsequently, reports are periodically sent back to
 * the source of the flood giving information about the rate at which
 * capsules are arriving.  This can be used to calculate the bandwidth of
 * the connection between two nodes.  By only sending reports periodically,
 * we avoid having to worry about the interference generated by collisions.
 *
 * NOTE: Doesn't do much error-checking yet.
 * 
 */

#include <common/pan-copyright.h>
#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pniapplink.h>
#include <pni/pnicodeobj.h>
#include <pni/pnimisc.h>
#include <codeobj/portmap/portmap.h>
#include <pni/pniutil_hashtab.h>
#include "floodcap_ser_co.h"


/* Hash table keeps track of multiple flows.
 * Entries are keyed on <sourceaddr,portno> pairs.
 */
#define pcoFlood_FlowHashTab_Size 11  /* Ideally a prime */
#define pcoFlood_FlowHashTab_KeySize (pniNodeAddr_SIZE+4)
static pniHashTab *pcoFlood_FlowHashTab   = NULL; /* Not created until used */

/* Keeps track of a flood flow. */
typedef struct _pcoFloodFlow {
  pniNodeAddr    sourceaddr;	/* Where the flood flow is coming from */
  int            portnum;	/* Where the flood flow is coming from */
  pniTimestamp   starttime;	/* When flood flow started */
  int            capcount;	/* How many capsules have arrived */
  int            bytecount;	/* How much data has arrived */
  int            updateperiod;	/* How frequently to send updates */
} pcoFloodFlow;


/* Very simple hashing function */
static unsigned int
pcoFlood_hashFn(char *key)
{
  int hash, i;

  if (!key) 
    return 0;
  for (hash=0, i=0; i<pniNodeAddr_SIZE; i++)
    {
      hash ^= (key[i])<<i;
    }
  return (hash % pcoFlood_FlowHashTab_Size);
}



/* Starts a new flood by creating an entry in the hash table */
static pcoFloodFlow *
pcoFlood_startFlow(pcoFloodCapsule *floodcap)
{
  pniSwseg     *ffsws;
  pcoFloodFlow *ff;
  char         *key;

  if (!pcoFlood_FlowHashTab)
    {
      /* Create the hashtable if this is the first time */
      pcoFlood_FlowHashTab = pniHashTab_create(pcoFlood_FlowHashTab_Size, 
					       pcoFlood_FlowHashTab_KeySize, 
					       pcoFlood_hashFn);
      if (!pcoFlood_FlowHashTab) 
	return NULL;
      pniSwseg_makePersistent(pniCodeobj_This, pcoFlood_FlowHashTab);
    }

  ffsws = pniSwseg_create_byMalloc(sizeof(pcoFloodFlow));
  if (!ffsws) return NULL;
  ff = (pcoFloodFlow*)ffsws->data;

  pniNodeAddr_dup(&ff->sourceaddr, &floodcap->sourceaddr);
  ff->portnum = floodcap->portnum;
  ff->capcount  = 0;
  ff->bytecount = 0;
  ff->updateperiod = floodcap->capcount;
  pniTimestamp_getCurrent(&ff->starttime);
  
  key = (char*)&(floodcap->sourceaddr);	/* source followed by port */
  if (NULL != pniHashTab_lookup(pcoFlood_FlowHashTab, key))
    {
      pniLog_warn("---> Flood flow already in hashtab!\n");
      return NULL;
    }

  pniHashTab_insert(pcoFlood_FlowHashTab, key, ffsws);

  return ff;
}



/* Process capsules as they arrive at the node */
void 
accept(pniStream *capstart, int caplength)
{
  pniStream   _ps, *ps=&_ps;
  pniNodeAddr curaddr;
  pniAppLink  *theapp;
  pcoFloodCapsule floodcap;
  pcoFloodFlow *ff;

  pniStream_dup(ps, capstart);
  pcoFloodCapsule_read(&floodcap, ps);
  pniStream_dup(ps, capstart);

  floodcap.hopcount++;
  pniLog_debug("---> At node with hopcount %d, portnum %d, type %d\n", 
	 floodcap.hopcount, floodcap.portnum, floodcap.type);

  pniNodeAddr_getLocalAddr(&curaddr);

  if (floodcap.type == pcoFloodCapsule_TYPE_START 
      || floodcap.type == pcoFloodCapsule_TYPE_DATA
      || floodcap.type == pcoFloodCapsule_TYPE_MULTI
      || floodcap.type == pcoFloodCapsule_TYPE_END)
    {
      /* Get to the destination */
      if (0 != pniNodeAddr_compare(&floodcap.destaddr, &curaddr))
	{
	  if (floodcap.type == pcoFloodCapsule_TYPE_MULTI)
	    {
	      int multictr;
	      pniSwseg  *dupsws;
	      pniStream oldcap, newcap;
	      pniLog_debug("---> HEADING TO DESTINATION!!!\n");
	      floodcap.type = pcoFloodCapsule_TYPE_DATA;
	      pcoFloodCapsule_write(ps, &floodcap);
	      /* Send all but one of the multidata capsules */
	      for (multictr=0; multictr<floodcap.capcount-1; multictr++)
		{
#if 0
		  dupsws=pniSwseg_create_byMalloc(caplength);
		  if (!dupsws) return;
		  pniStream_init(&newcap, dupsws, 0);
		  pniStream_dup(&oldcap, capstart);
		  pniStream_copy(&newcap, &oldcap, caplength);
		  pniStream_init(&newcap, dupsws, 0);
#else
		  /* This violates invariants! */
		  pniStream_dup(&newcap, capstart);
#endif
		  pniNetwork_sendCapsule(&floodcap.destaddr,
					 &newcap, caplength);
		}
	      /* Send the remaining one */
	      pniNetwork_sendCapsule(&floodcap.destaddr, capstart, caplength);
	      return;
	    }
	  else  /* DATA or START */
	    {
	      pniLog_debug("---> HEADING TO DESTINATION!!!\n");
	      pcoFloodCapsule_write(ps, &floodcap);
	      pniNetwork_sendCapsule(&floodcap.destaddr, capstart, caplength);
	      return;
	    }
	}
      else 
	{
	  /* Start or data capsule at the destination */
	  if (floodcap.type == pcoFloodCapsule_TYPE_END)
	    {
	      char *key;
	      key = (char*)&(floodcap.sourceaddr);/* source followed by port */
	      pniLog_debug("---> removing flood flow\n");
	      pniHashTab_remove(pcoFlood_FlowHashTab, key);
	      pniLog_debug("---> done removing flood flow\n");
	      return;
	    }
	  else if (floodcap.type == pcoFloodCapsule_TYPE_START)
	    {
	      if (!(ff = pcoFlood_startFlow(&floodcap)))
		{
		  pniLog_warn("pcoFlood_startFlow failed!\n");
		  return;
		}
	    }
	  else /* Just a data capsule */
	    {
	      char *key;
	      pniSwseg *ffsws;
	      key = (char*)&(floodcap.sourceaddr);/* source followed by port */
	      if (!(ffsws = pniHashTab_lookup(pcoFlood_FlowHashTab, key)))
		{
		  pniLog_warn("---> Flood flow not in FlowHashTab...\n");
		  return;	/* Drop the capsule */
		}	      
	      ff = (pcoFloodFlow*)ffsws->data;
	    }
	  /* Now for either START or DATA */
	  ff->capcount++;
	  ff->bytecount += caplength;
	  
	  if ((ff->capcount % ff->updateperiod) == 0)
	    {
	      /* Send an update */
	      pniTimestamp curtime;

	      pniTimestamp_getCurrent(&curtime);
	      floodcap.type = pcoFloodCapsule_TYPE_UPDATE;
	      floodcap.capcount  = ff->capcount; 
	      floodcap.bytecount = ff->bytecount;
	      pniTimestamp_subtract(&floodcap.duration,
				    &curtime, &ff->starttime);
	      pniLog_warn("--> Responding with capcount %d, bytecount %d, starttime %d.%d  endtime %d.%d  duration %d.%d\n", 
			  ff->capcount, ff->bytecount, 
			  ff->starttime.sec, ff->starttime.usec, 
			  curtime.sec, curtime.usec, 
			  floodcap.duration.sec, floodcap.duration.usec);

	      pcoFloodCapsule_write(ps, &floodcap);
	      pniNetwork_sendCapsule(&floodcap.sourceaddr, capstart,
				     pcoFloodCapsule_SIZE);
	      return;
	    }	    
	}    
      return;
    }

  else /* pcoFloodCapsule_TYPE_UPDATE */
    {
      /* Go back to the source then deliver to the app there */

      if (0 == pniNodeAddr_compare(&floodcap.sourceaddr, &curaddr))
	{
	  pniLog_debug("---> BACK AT SOURCE!!!\n");
	  pcoFloodCapsule_write(ps, &floodcap);

	  theapp = pcoPortmap_mapPortnumToApplink(floodcap.portnum);
	  if (!theapp)
	    { pniLog_warn("---> ERROR: couldn't find port %d\n", 
			  floodcap.portnum); 
	    return; }
	  pniAppLink_deliverData(theapp, capstart, caplength);
	  return;
	}
      else 
	{
	  pniLog_debug("---> HEADING HOME!!!\n");
	  pcoFloodCapsule_write(ps, &floodcap);
	  pniNetwork_sendCapsule(&floodcap.sourceaddr, capstart, caplength);
	  return;
	}      
    }
}

