/* $Header: /home/dl2/nygren/cvs/pan/src/apps/ping/pingapp.c,v 1.6 1998/01/20 23:43:03 nygren Exp $
 *
 * FILE:    pingapp.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Trivial "ping" application.  Uses pingcapsule and puts 
 * a 4-byte sequence number and a time-stamp into the payload.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <getopt.h>

#include <applib/panapplib.h>
#include <codeobj/portmap/portmap.h>
#include "pingcap_ser_app.h"

#define fail_if(pred,pusage,fmt,arg...) \
   do { if ((pred)) { \
              if(!(pusage)) fprintf(stderr, "%s:%d: ",__FILE__,__LINE__);\
	      fprintf(stderr, fmt, ##arg);\
              if (pusage) print_usage(); \
	      exit(-1); } } while(0)

#define print_verbose(fmt,arg...) \
   if (opt_verbose) \
        fprintf(stderr, fmt,##arg)

#define print_info(fmt,arg...) \
   if (!opt_quiet) \
        printf(fmt, ##arg)


/* Arguments and options */
int         opt_quiet=0;	/* Only output final results */
int         opt_verbose=0;	/* Be very verbose in printing stuff */
int         opt_count=1;	/* Number of packets to send */
int         opt_concurrent=1;	/* The number of capsules to send at once.
				 * 0 = flood ping. */
int         opt_capsize=56;	/* The amount of data to add after header */
int         opt_stampatnode=0;	/* Do timestamping in the node */
char       *opt_localnode="0";	/* Local user-space node to connect to */
char       *opt_destaddr=NULL; /* Destination address */
char       *progname;

/* global variables */
panNodeAddr localaddr, destaddr;
panNodeLink    *pnl;

void ping_send_capsule(panSwseg *capsws, palPingCapsule *pingcap, int seqno);
int  ping_receive_capsule(panTimestamp *rtt);


void 
print_usage(void)
{
  printf("Usage: %s [-hqvS] -c <count> -C <concurrent>\n", progname);
  printf("                      -s <size> -L <localnode>\n");
  printf("  -q              quiet output\n");
  printf("  -v              verbose output\n");
  printf("  -S              timestamp in local node\n");
  printf("  -c <count>      number to send\n");
  printf("  -C <concurrent> concurrent number to have in transit at once\n"); 
  printf("  -s <size>       total size of capsule (header+payload) not\n");
  printf("                  including underlying headers.\n");
  printf("  -L <localnode>  local user-space node to connect to\n");
}


void
main(int argc, char **argv)
{
  panAppCodeobj  *ping_co;
  panCodeobjName  ping_coname;
  panSwseg       *capsws;
  panTimestamp    rtt, maxrtt={0,0}, minrtt={100000000,0};
  panTimestamp    starttime, endtime, runtime;
  unsigned int    usecaccum=0;
  int32_t         portnum, retval;
  palPingCapsule  pingcap; 

  int             curopt;	 /* For getopt */
  int             intransit=0; 	 /* Number of capsules in transit */
  int             numreceived=0; /* Number of capsules received */
  int             numsent=0;     /* Number of capsules sent */
  int             seqno=1;       /* Sequence number */
  double          kBps;

  progname=argv[0];

  while (EOF != (curopt = getopt(argc, argv, "hqvSc:C:s:L:")))
    {
      switch (curopt) 
	{
	case 'h':
	case '?':
	  print_usage(); exit(0);
	  break;
	case 'q':
	  opt_quiet=1;
	  break;
	case 'v':
	  opt_verbose=1;
	  break;
	case 'S':
	  opt_stampatnode=1;
	  break;
	case 'c':
	  opt_count=atoi(optarg);
	  break;
	case 'C':
	  opt_concurrent=atoi(optarg);
	  break;
	case 's':
	  opt_capsize=atoi(optarg);
	  break;
	case 'L':
	  opt_localnode=optarg;
	  break;
	default:
	  print_usage(); exit(-1);
	}
    }

  fail_if((optind != argc-1), 1, "no destination provided\n");
  opt_destaddr = argv[optind];

  print_verbose("Connecting to the node <%s>...\n", opt_localnode);
  pnl = panNodeLink_create(opt_localnode);
  fail_if(!pnl,0, "Couldn't create nodelink!\n");
  panNodeLink_getLocalAddr(&localaddr, pnl);
  print_verbose("Yay!  Connected to local node %s\n",
		panNodeAddr_stringify(NULL,0,&localaddr));

  fail_if(!(panNodeAddr_parse(&destaddr, opt_destaddr)),
	  0,"Couldn't parse dest address!\n");

  portnum = 0;			/* Don't care which... */
  retval = palPortmap_bind(pnl, &portnum, 1/*insertco*/);
  fail_if((retval != 0), 0,"Couldn't bind to port (retval %d)!\n",retval);
  print_verbose("Bound to port number %d\n", portnum);

  ping_co = panAppCodeobj_load("pingcapsule");
  fail_if(!ping_co,0,"Couldn't load codeobject!\n");

  fail_if(   !panAppCodeobj_getCodeobjName(&ping_coname, ping_co),
	  0, "Couldn't get coname!\n");

  fail_if(   !panAppCodeobj_insertIntoNode(pnl, ping_co, 1),
	  0, "Couldn't insert codeobject!\n");

  if (opt_capsize<palPingCapsule_SIZE) 
    opt_capsize=palPingCapsule_SIZE;

  /* Build up the base capsule header than will be used for all capsules */
  capsws = panSwseg_create_byMalloc(opt_capsize);

  panCodeobjName_dup(&(pingcap.coname), &ping_coname);
  panNodeAddr_dup(&(pingcap.cohomeaddr), &localaddr);
  panNodeAddr_dup(&(pingcap.sourceaddr), &localaddr);
  panNodeAddr_dup(&(pingcap.destaddr), &destaddr);
  pingcap.status      = palPingCapsule_STATUS_JUST_INSERTED;
  pingcap.hopcount    = 0;
  pingcap.portnum     = portnum;
  pingcap.stampatnode = opt_stampatnode;

  /* Fill in when sending:
   *    if (!stampatnode) panTimestamp_getCurrent(&pingcap.senttime);
   *    pingcap.seqno      = seqno++;  
   */

  panTimestamp_getCurrent(&starttime);
  /* Enter into a loop until we get all our capsules back.
   * At any point, sent out enough capsules to get us up
   * to the concurrent number, then read capsules. */
  while (numreceived < opt_count)
    {
      while ((intransit < opt_concurrent) && (numsent < opt_count))
	{
	  ping_send_capsule(capsws, &pingcap, seqno++);
	  intransit++;
	  numsent++;
	}
      /* Receive a capsule */
      if (ping_receive_capsule(&rtt))
	{	  
	  if (panTimestamp_compare(&rtt, &maxrtt) > 0)
	    panTimestamp_dup(&maxrtt, &rtt);
	  if (panTimestamp_compare(&rtt, &minrtt) < 0)
	    panTimestamp_dup(&minrtt, &rtt);
	  usecaccum += rtt.sec*1000000 + rtt.usec;

	  intransit--;
	  numreceived++;
	}
    }
  panTimestamp_getCurrent(&endtime);

  /* Print summary */
  printf("--- ping statistics ---\n");
  printf("%d packets transmitted, %d packets received, %1.0f%% packet loss\n",
	 numsent, numreceived, 
	 100.0*((double)(numsent-numreceived))/(double)numsent);
  printf("round-trip min/avg/max = %s / %d / ",
	 panTimestamp_stringify(NULL, 0, &minrtt, 0),
	 usecaccum / numreceived);
  printf("%s usecs\n",
	 panTimestamp_stringify(NULL, 0, &maxrtt, 0));	 
  panTimestamp_subtract(&runtime, &endtime, &starttime);
  kBps = ((double)(numreceived*opt_capsize)
    / ((double)runtime.sec)+(((double)runtime.usec)/1e6)) / 1000.0;
  printf("bytes = %d,  runtime = %s secs,   bandwidth = %1.2f KBps = %1.2f Kbps\n", 
	 (numreceived*opt_capsize),
	 panTimestamp_stringify(NULL, 0, &runtime, 
				panTimestamp_STRINGIFY_FMT_INSECS),
	 kBps, 8.0*kBps);

  print_verbose("Doing a portmapper release...\n");

  palPortmap_release(pnl, portnum);

  print_verbose("Shutting down...\n");

  panNodeLink_disconnect(pnl);

  print_verbose("Destroying...\n");
  panNodeLink_destroy(pnl);
  
}


void 
ping_send_capsule(panSwseg *capsws, palPingCapsule *pingcap, int seqno)
{
  panStream       _ps, *ps=&_ps;

  if (!opt_stampatnode)
    panTimestamp_getCurrent(&pingcap->senttime);
  pingcap->seqno      = seqno;  
  panStream_init(ps, capsws, 0);
  palPingCapsule_write(ps, pingcap);
  panStream_init(ps, capsws, 0);
  panNodeLink_send(pnl, ps, opt_capsize);  
}


int
ping_receive_capsule(panTimestamp *rtt)
{
  panStream      _ps, *ps=&_ps;
  panSwseg      *returnedcapsws;
  int            retval;
  palPingCapsule pingreply;
  panTimestamp   curtime;

  print_verbose("Waiting to get reply back...\n"); 
  returnedcapsws = panNodeLink_retr(pnl, 1);
  panTimestamp_getCurrent(&curtime);
  print_verbose("Got reply back!\n");
  panStream_init(ps, returnedcapsws, 0);

  if ((retval=palPingCapsule_read(&pingreply, ps)) < palPingCapsule_SIZE)
    { printf("ERROR: got short reply to ping header (%d bytes)\n",retval); }

  /* Calculate the rtt */
  if (opt_stampatnode)
    {
      panTimestamp_dup(rtt, &pingreply.senttime);
    }
  else
    {
      panTimestamp   curtime;
      panTimestamp_getCurrent(&curtime);      
      panTimestamp_subtract(rtt, &curtime, &pingreply.senttime);
    }

  print_verbose("Got capsule back with status %d, hop count %d, portnum %d, "
		"seq %d, timestamp 0x%x 0x%x\n", 
		pingreply.status, pingreply.hopcount, 
		pingreply.portnum, pingreply.seqno,
		pingreply.senttime.sec, pingreply.senttime.usec);

  print_info("%d bytes from %s: seqno=%d hops=%d time=%s usecs\n",
	     returnedcapsws->length, 
	     panNodeAddr_stringify(NULL,0, &pingreply.sourceaddr),
	     pingreply.seqno, pingreply.hopcount, 
	     panTimestamp_stringify(NULL, 0, rtt, 0));

  panSwseg_unref(returnedcapsws);
  return 1;
}
