/* $Header: /home/dl2/nygren/cvs/pan/src/codeobj/dynamicload/dynamicload.c,v 1.4 1998/01/20 23:43:32 nygren Exp $
 *
 * FILE:    dynamicload.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Capsule code for implementing dynamic loading of code objects.
 * One of these capsules gets sent towards the code object home
 * in an attempt to find the code object.  On finding it,
 * it ships it back to the host that requested it.
 *
 * Capsule contains:
 * 
 *    SIZE        DESCRIPTION
 *    ----        -----------
 *    24 bytes    DLCodeobjName --     Name of this dynamicload code object
 *    16 bytes    DLCodeobjHomeAddr -- Home for this dynamicload code object
 *    16 bytes    TargetHomeAddr --    The home address of the code object
 *                                     we're fetching 
 *    16 bytes    ClientAddr --        The address of the node we're fetching
 *                                     this code object for.
 *     4 bytes    Status/Size (int32)  If 0, we're still looking for the 
 *                                     code object.  Otherwise, the size
 *                                     of the code object that we're
 *                                     shipping back.
 *  This is followed by either:
 * 
 *    24 bytes    TargetCOName         The name of the code object we're
 *                                     looking for.
 * 
 *  or: 
 *
 *     N bytes    TargetCodeobj        The code object that we've retrieved
 *
 * NOTE: Doesn't do error-checking yet.
 *
 */

#include <common/pan-copyright.h>
#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pnicodeobj.h>
#include <pni/pnimisc.h>


/* Forwards a dynamic load capsule towards its
 * destination, or drops it if it reaches it.
 */
void
transport_towards_target(pniStream *capstart, int caplength)
{
  pniStream       _ps, *ps=&_ps;
  pniNodeAddr     targetaddr, curaddr;

  pniStream_dup(ps, capstart);

  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE);
  pniNodeAddr_read(&targetaddr, ps);

  pniNodeAddr_getLocalAddr(&curaddr);
  if (0 == pniNodeAddr_compare(&targetaddr, &curaddr))
    {
      pniLog_error("===>> dynamicload: got to codeobj home and couldn't find!\n");
      pniLog_error("===>> dynamicload: should send back error message.\n");
      return;
    }
  pniLog_debug("dynamicload: sending capsule towards targetaddr\n");
  pniNetwork_sendCapsule(&targetaddr, capstart, caplength);
}


/* Constructs a reply containing the code object and sends it
 * back towards the source. */
void
send_codeobj_back(pniStream *capstart, int caplength, 
		  pniSwseg *codeobj,   int colength)
{
  pniStream       _ps, *ps=&_ps;
  pniNodeAddr     clientaddr;
  pniSwseg        *headsws;	/* Overlaying swseg */
  pniSwseg        *capsws;	/* Swseg for the capsule header */
  unsigned         capoff;

  pniStream_dup(ps, capstart);

  /* Read the client address */
  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE + pniNodeAddr_SIZE);
  pniNodeAddr_read(&clientaddr, ps);

  /* Write in the length of the code object */
  if (!colength)
    { pniLog_error("===> dynamicload: got bad colength!!!\n"); return; }
  pniStream_write_net_int32(ps, colength);

  /* Make an overlay of the capsule header */
  capsws = pniStream_getState(capstart, &capoff);
  headsws  = pniSwseg_overlay(capsws, capoff, 
			      pniCapsuleHeader_SIZE + 2*pniNodeAddr_SIZE + 4);
  /* Now chain the header overlay in front of the codeobj */
  pniSwseg_chain(headsws, codeobj);

  pniStream_init(ps, headsws, 0);

  pniLog_debug("dynamicload: sending capsule towards clientaddr\n");
  pniNetwork_sendCapsule(&clientaddr, ps, 
			 pniCapsuleHeader_SIZE + 2*pniNodeAddr_SIZE
			 + 4 + colength);
}


/* Accepts and evaluates a dynamicload capsule */
void 
accept(pniStream *capstart, int caplength)
{

  pniStream       _ps, *ps=&_ps;
  int             status_or_size, colength;
  pniSwseg       *codeobj;
  pniCodeobjName  coname;

  pniStream_dup(ps, capstart);
  /* Seek forward to the status/size field */
  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE + 2*pniNodeAddr_SIZE);

  status_or_size = pniStream_read_net_int32(ps);

  if (status_or_size == 0) 
    {
      /****************************************************************/
      /* Still looking...  Check to see if this node has the codeobj. */
      /****************************************************************/

      pniCodeobjName_read(&coname, ps, 0);
      codeobj = pniCodeobj_lookup(&coname, &colength);

      if (codeobj)
	{
	  /* We've found the code object!  Construct and send a reply. */
	  send_codeobj_back(capstart, caplength, codeobj, colength);
	}
      else
	{
	  /* Keep looking by sending this capsule on towards its destination */
	  transport_towards_target(capstart, caplength);
	}
    }
  else
    {
      /*******************************************************************/
      /* Transport the code object back and load it into nodes as we go. */
      /*******************************************************************/
      
      pniNodeAddr     clientaddr, curaddr, targetaddr;
      
      pniStream_dup(ps, capstart);
      
      pniStream_seekFwd(ps, pniCapsuleHeader_SIZE);
      pniNodeAddr_read(&targetaddr, ps);
      pniNodeAddr_read(&clientaddr, ps);
      
      pniNodeAddr_getLocalAddr(&curaddr);

      if (0 != pniNodeAddr_compare(&clientaddr, &curaddr))
	{
	  pniLog_debug("dynamicload: sending codeobject back to client\n");
	  pniNetwork_sendCapsule(&clientaddr, capstart, caplength);
	}
      
      /* Load the code object into the node */
      pniStream_seekFwd(ps, 4);
      pniLog_debug("dynamicload: loading codeobj into node\n");
      pniCodeobj_load(ps, &targetaddr);
    }
  
}

