/* $Header: /home/dl2/nygren/cvs/pan/src/codeobj/portmap/portmap.c,v 1.4 1998/01/23 07:41:14 nygren Exp $
 *
 * FILE:    portmap.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides routines for mapping port numbers to applications.
 * Applications can insert a capsule in order to request a port
 * number which then gets allocated and put into the port
 * number hash table.  Also exports functions to other 
 * code objects to allow them to access the mappings.
 * It would be nice if there was a way to do this somewhat
 * securely.
 *
 *
 * Capsules inserted by applications are of the form:
 * 
 *    SIZE        DESCRIPTION
 *    ----        -----------
 *    24 bytes    CodeobjName           Name of this code object
 *    16 bytes    CodeobjHomeAddr       Home for this code object
 *     4 bytes    Operation             Operation to perform
 *                                        1 == bind to port
 *                                        2 == delete port
 *     4 bytes    PortNum               Port number to bind to or delete.
 *                                      (0 to bind to an arbitrary port).
 *     N bytes    Arbitrary Data        Ignored for now but incorporated
 *                                      into response.
 * 
 * Responses are sent back to the sending application using the
 * same data format.  Arbitrary data is preserved, portnum
 * is replaced with the actual port number we bound to,
 * and operation is replaced with a result code:
 *
 *      0 == operation succeeded (portnum in PortNum)
 *     -1 == couldn't bind to port -- already in use
 *     -2 == couldn't bind to port -- out of resources
 *     -3 == couldn't perform operation -- permision denied
 *     -4 == couldn't find port -- doesn't exist
 *     -5 == no applink to bind
 *     -6 == passed an invalid argument
 */

#include <common/pan-copyright.h>
#include <pni/pniswseg.h>
#include <pni/pniswsstream.h>
#include <pni/pninetwork.h>
#include <pni/pnicodeobj.h>
#include <pni/pniapplink.h>
#include <pni/pnimisc.h>
#include <pni/pniutil_hashtab.h>
#include "portmap.h"


#define pcoPortmap_HashTabSize 97              /* Ideally, a prime not 
				                   * near a power of 2 */
static pniHashTab *pcoPortmap_HashTab     = NULL; /* Not created until used */
static int         pcoPortmap_NextPortnum = 16384; /* Next to allocate */


/* Just does division hashing... */
static unsigned int
pcoPortmap_hashFn(char *key)
{
  int ikey;

  if (!key) 
    return 0;
  ikey = *((int*)key);
  return (ikey % pcoPortmap_HashTabSize);
}


/* Binds portnum to the currently running AppLink.
 * Returns a return code and puts the port actually
 * bound to into portnum.  If portnum is 0, binds to
 * the next available port. 
 */
int 
pcoPortmap_bindPortnumToApplink(int *portnum)
{
  pniAppLink *alink;
  char *key;

  if (!portnum) 
    return pcoPortmap_RVAL_EINVAL;

  if (!pcoPortmap_HashTab)
    {
      /* Create the hashtable if this is the first time */
      pcoPortmap_HashTab = pniHashTab_create(pcoPortmap_HashTabSize, 4, 
					     pcoPortmap_hashFn);
      if (!pcoPortmap_HashTab) 
	return pcoPortmap_RVAL_EFULL;
      pniSwseg_makePersistent(pniCodeobj_This, pcoPortmap_HashTab);
    }

  alink = pniAppLink_getSender();
  if (!alink || !pniAppLink_isConnected(alink)) 
    return pcoPortmap_RVAL_ENOALINK;
  if (*portnum == 0) 
    *portnum = pcoPortmap_NextPortnum++;
  key = (char*)portnum;

  if (NULL != pniHashTab_lookup(pcoPortmap_HashTab, key))
    return pcoPortmap_RVAL_EINUSE;

  pniHashTab_insert(pcoPortmap_HashTab, key, alink);

  return pcoPortmap_RVAL_SUCCESS;
}


/* Maps a port number to an applink and returns the applink or NULL if
 * not found */
pniAppLink *
pcoPortmap_mapPortnumToApplink(int portnum)
{
  char *key = (char*)&portnum;
  pniAppLink *alink;

  if (!pcoPortmap_HashTab)
    return NULL;
  alink = pniHashTab_lookup(pcoPortmap_HashTab, key);
  if (!pniAppLink_isConnected(alink))
    {
      pniHashTab_remove(pcoPortmap_HashTab, key);
      return NULL;
    }  
  return alink;
}


/* Deletes a port number to applink mapping.  The applink
 * must match the currently running applink.  Returns
 * an RVAL.
 */
int
pcoPortmap_deletePortnumMapping(int portnum)
{
  char *key = (char*)&portnum;
  pniAppLink *applink, *runningapplink;

  if (!pcoPortmap_HashTab)
    return pcoPortmap_RVAL_ENOTFOUND;
  
  applink = (pniAppLink*)pniHashTab_lookup(pcoPortmap_HashTab, key);
  if (!applink) 
    return pcoPortmap_RVAL_ENOTFOUND;
  runningapplink = pniAppLink_getSender();
  if (pniAppLink_isConnected(applink)
      && (!runningapplink || ((runningapplink != applink))))
    return pcoPortmap_RVAL_EPERM;
  pniHashTab_remove(pcoPortmap_HashTab, key);
  return pcoPortmap_RVAL_SUCCESS;
}


/* Delivers a response to the current application.
 * Advances capstart, though.  */
static void
deliver_response(pniStream *capstart, int caplength,
		 int retval, int portnum)
{
  pniStream       _ps, *ps=&_ps;
  pniAppLink     *alink;

  pniStream_dup(ps, capstart);
  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE);
  pniStream_write_net_int32(ps, retval);
  pniStream_write_net_int32(ps, portnum);

  alink = pniAppLink_getSender();
  pniAppLink_deliverData(alink, capstart, caplength);
}


/* Accepts and evaluates a dynamicload capsule */
void 
accept(pniStream *capstart, int caplength)
{
  pniStream       _ps, *ps=&_ps;
  int operation, portnum, retval;

  pniStream_dup(ps, capstart);
  pniStream_seekFwd(ps, pniCapsuleHeader_SIZE);

  operation = pniStream_read_net_int32(ps);
  portnum   = pniStream_read_net_int32(ps);

  if (operation == pcoPortmap_OP_BIND)
    {
      pniLog_debug("--> portmap: binding a port to an applink\n");
      retval = pcoPortmap_bindPortnumToApplink(&portnum);
      deliver_response(capstart, caplength, retval, portnum);
      return;
    }
  else if (operation == pcoPortmap_OP_DELETE)
    {
      pniLog_debug("--> portmap: deleting the binding to port %d\n", portnum);
      retval = pcoPortmap_deletePortnumMapping(portnum);
      deliver_response(capstart, caplength, retval, portnum);
      return;      
    }
  else
    {
      deliver_response(capstart, caplength, pcoPortmap_RVAL_EINVAL, 0);
      return;
    }
}

