/* $Header: /home/dl2/nygren/cvs/pan/src/common/arch/linux2/nodeaddr.c,v 1.3 1998/01/23 07:41:43 nygren Exp $
 *
 * FILE:    nodeaddr.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Contains platform-dependant routines for handling node addresses.
 *
 */


#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/nodeaddr.h>

#ifndef __KERNEL__
#include <ctype.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif



/* Compares two panNodeAddr's for sorting purposes.
 * Returns 0 if they're equal, -1 if na1<na2, and 1 if na1>na2.
 */
int
panNodeAddr_compare(panNodeAddr *na1, panNodeAddr *na2)
{
  u_int32_t 
    *a1=(u_int32_t*)(&(na1->data[0])),
    *a2=(u_int32_t*)(&(na2->data[0]));

  if (*a1 < *a2) return -1;
  else if (*a1 > *a2) return 1;
  else return 0;
}


/*
 * Reads a panNodeAddr in from a stream and advances the stream by 
 * a total of 16 bytes (128 bits).  Returns the number of bytes read.
 */
unsigned
panNodeAddr_read(panNodeAddr *na, panStream *ps)
{
  unsigned bread=0;			/* Bytes read */

  bread += panStream_read(na->data, ps, panNodeAddr_SIZE);
  memset(na->data+4, 0, panNodeAddr_SIZE-4); /* pad with 0's */
  return bread;
}


/*
 * Writes a panNodeAddr to a stream and advances the stream by 
 * a total of 16 bytes (128 bits).  Returns the number of bytes written.
 */
unsigned
panNodeAddr_write(panStream *ps, panNodeAddr *na)
{
  unsigned bwrote=0;			/* Bytes read */

  bwrote += panStream_write(ps, na->data, panNodeAddr_SIZE);
  return bwrote;
}


/* Initializes a panNodeAddr to have IP address addr and port port.
 * Value should be passed in network byte order.
 */
void
panNodeAddr_init(panNodeAddr *na, u_int32_t addr)
{
  memset(((char*)na)+4, 0, sizeof(panNodeAddr)-4); /* pad with 0's */
  *((u_int32_t*)(&(na->data[0]))) = addr;
}


/* Prints out the value of a panNodeAddr */
void 
panNodeAddr_print(panNodeAddr *na)
{
  unsigned char *addr;

  addr = na->data;
  panPrint_debug("%d.%d.%d.%d",addr[0],addr[1],addr[2],addr[3]);
}


/* Copies the contents of a node address into a string
 * and returns it.  If buf is null, uses a static buffer.
 */
#ifdef __KERNEL__
char *
panNodeAddr_stringify(char *buf, int buflen, panNodeAddr *na)
{
  static char safebuf[16];
  unsigned char *addr;

  addr = na->data;
  if (buf && buflen<16)
    {
      return "__use_a_bigger_buffer__";
    }
  else
    {
      if (!buf) buf=safebuf;
      sprintf(buf, "%u.%u.%u.%u",
	       0xff&addr[0], 0xff&addr[1], 0xff&addr[2], 0xff&addr[3]);
      return buf;
    }
}

#else
char *
panNodeAddr_stringify(char *buf, int buflen, panNodeAddr *na)
{
  static char safebuf[64];
  unsigned char *addr;

  addr = na->data;
  if (!buf) { buf=safebuf; buflen=63; }
  snprintf(buf, buflen, "%u.%u.%u.%u",
	   0xff&addr[0], 0xff&addr[1], 0xff&addr[2], 0xff&addr[3]);
  return buf;
}
#endif



#ifndef __KERNEL__

/* Takes a textual representation of a node address and
 * stores it into the node address na.  For UNodes, 
 * addresses should be in the form "hostname:portnum".
 * Returns 1 on success, 0 on failure.
 */
extern int 
panNodeAddr_parse(panNodeAddr *na, char *nodename)
{
  char            nodenamebuf[256];
  char           *hostname_str;
  struct hostent *hentry;
  u_int32_t       hostaddr;

  strncpy(nodenamebuf, nodename, 256-2);
  hostname_str = nodenamebuf;
  
  hentry = gethostbyname(hostname_str);
  if (!hentry)
    { perror("gethostbyname");
    panPrint_warn("panNodeAddr_parse: couldn't find host <%s> in <%s>\n", 
		  hostname_str, nodename);
    return 0; }

  assert((hentry->h_length == 4) && (hentry->h_addrtype == AF_INET));
  memcpy(&hostaddr, hentry->h_addr, hentry->h_length);

  panNodeAddr_init(na, hostaddr);

  return 1;
}


#endif /* __KERNEL__ */
