/* $Header: /home/dl2/nygren/cvs/pan/src/common/arch/unixuser/nodeaddr.c,v 1.5 1998/01/23 07:41:50 nygren Exp $
 *
 * FILE:    nodeaddr.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Contains platform-dependant routines for handling node addresses.
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/nodeaddr.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>




/* Compares two panNodeAddr's for sorting purposes.
 * Returns 0 if they're equal, -1 if una1<una2, and 1 if una1>una2.
 */
int
panNodeAddr_compare(panNodeAddr *una1, panNodeAddr *una2)
{
  u_int32_t 
    *a1=(u_int32_t*)(&(una1->data[0])),
    *a2=(u_int32_t*)(&(una2->data[0]));
  u_int16_t *p1, *p2;

  if (*a1 < *a2) return -1;
  else if (*a1 > *a2) return 1;
  p1=(u_int16_t*)(&(una1->data[4]));
  p2=(u_int16_t*)(&(una2->data[4]));
  if (*p1 < *p2) return -1;
  else if (*p1 > *p2) return 1;
  /* Else, *a1==*a2 and *p1==*p2 */
  return 0;
}


/*
 * Reads a panNodeAddr in from a stream and advances the stream by 
 * a total of 16 bytes (128 bits).  Returns the number of bytes read.
 */
unsigned
panNodeAddr_read(panNodeAddr *una, panStream *ps)
{
  unsigned bread=0;			/* Bytes read */

  bread += panStream_read(una->data, ps, panNodeAddr_SIZE);
  memset(una->data+6, 0, panNodeAddr_SIZE-6); /* pad with 0's */
  return bread;
}


/*
 * Writes a panNodeAddr to a stream and advances the stream by 
 * a total of 16 bytes (128 bits).  Returns the number of bytes written.
 */
unsigned
panNodeAddr_write(panStream *ps, panNodeAddr *una)
{
  unsigned bwrote=0;			/* Bytes read */

  bwrote += panStream_write(ps, una->data, panNodeAddr_SIZE);
  return bwrote;
}


/* Initializes a panNodeAddr to have IP address addr and port port.
 * Both values should be passed in network byte order.
 */
void
panNodeAddr_init(panNodeAddr *una, u_int32_t addr, u_int16_t port)
{
  *((u_int32_t*)(&(una->data[0]))) = addr;
  *((u_int16_t*)(&(una->data[4]))) = port;
  memset(una->data+6, 0, panNodeAddr_SIZE-6);
}


/* Prints out the value of a panNodeAddr */
void 
panNodeAddr_print(panNodeAddr *una)
{
  unsigned char *addr;
  u_int16_t *port;

  addr = una->data;
  port = (u_int16_t*)&(una->data[4]);
  panPrint_debug("%d.%d.%d.%d:%d",addr[0],addr[1],addr[2],addr[3],
	       ntohs(*port));
}

/* Copies the contents of a node address into a string. */
char *
panNodeAddr_stringify(char *buf, int buflen, panNodeAddr *una)
{
  static char safebuf[64];
  unsigned char *addr;
  u_int16_t *port;

  addr = una->data;
  port = (u_int16_t*)&(una->data[4]);
  if (!buf) { buf=safebuf; buflen=63; }
  snprintf(buf, buflen, "%d.%d.%d.%d:%d",
	   addr[0], addr[1], addr[2], addr[3], ntohs(*port));
  return buf;
}


/* Takes a textual representation of a node address and
 * stores it into the node address na.  For UNodes, 
 * addresses should be in the form "hostname:portnum".
 * Returns 1 on success, 0 on failure.
 */
extern int 
panNodeAddr_parse(panNodeAddr *na, char *nodename)
{
  char            nodenamebuf[256];
  char           *portnum_str, *hostname_str, *colon;
  struct hostent *hentry;
  u_int32_t       hostaddr;
  u_int16_t       portnum;

  strncpy(nodenamebuf, nodename, 256-2);
  colon = strchr(nodenamebuf, ':');
  if (!colon) 
    { panPrint_warn("panNodeAddr_parse: invalid nodeaddr name <%s>\n",
		    nodename);
    return 0; }
  *colon = '\0';
  portnum_str = colon+1;
  hostname_str = nodenamebuf;
  
  hentry = gethostbyname(hostname_str);
  if (!hentry)
    { perror("gethostbyname");
    panPrint_warn("panNodeAddr_parse: couldn't find host <%s> in <%s>\n", 
		  hostname_str, nodename);
    return 0; }

  assert((hentry->h_length == 4) && (hentry->h_addrtype == AF_INET));
  memcpy(&hostaddr, hentry->h_addr, hentry->h_length);

  if (!isdigit(*portnum_str))
    { panPrint_warn("panNodeAddr_parse: invalid port number <%s> in <%s>\n", 
		  portnum_str, nodename);
    return 0; }
  portnum = atoi(portnum_str);

  panNodeAddr_init(na, hostaddr, htons(portnum));

  return 1;
}
