/* $Header: /home/dl2/nygren/cvs/pan/src/common/pancodeobj.c,v 1.5 1997/12/10 06:58:21 nygren Exp $
 *
 * FILE:    pancodeobj.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for reading and generating PAN code objects from and
 * to software segment streams.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/pancodeobj.h>
#include <common/sha1.h>


/* Returns the type of a code object name */
u_int16_t
panCodeobjName_getType(panCodeobjName *con)
{
  u_int16_t *shortdata;
  shortdata = (u_int16_t*)(con->data);
  return ntohs(shortdata[0]);
}


/* Returns the version of a code object name */
u_int16_t
panCodeobjName_getVers(panCodeobjName *con)
{
  u_int16_t *shortdata;
  shortdata = (u_int16_t*)(con->data);
  return ntohs(shortdata[1]);
}

/* Sets the type of a code object name to type */
void
panCodeobjName_setType(panCodeobjName *con, u_int16_t type)
{
  u_int16_t *shortdata;
  shortdata = (u_int16_t*)(con->data);
  shortdata[0] = htons(type);
}

/* Sets the version of a code object name to vers */
void
panCodeobjName_setVers(panCodeobjName *con, u_int16_t vers)
{
  u_int16_t *shortdata;
  shortdata = (u_int16_t*)(con->data);
  shortdata[1] = htons(vers);
}

/* Sets the hash of a code object name to digest (which must be 
 * an array of panCodeobjName_HASHSIZE bytes). */
void
panCodeobjName_setHash(panCodeobjName *con, u_int8_t *digest)
{
  int i;
  u_int8_t *bytedata;
  bytedata = (u_int8_t*)(con->data);
  for (i=0; i<panCodeobjName_HASHSIZE; i++)
    bytedata[i+4] = digest[i]; 
}


/* Compares the hash of a code object name to digest (which must be 
 * an array of panCodeobjName_HASHSIZE bytes). Returns 1 if
 * they match or 0 if they don't match.  */
int
panCodeobjName_compareHash(panCodeobjName *con, u_int8_t *digest)
{
  int i;
  u_int8_t *bytedata;
  bytedata = (u_int8_t*)(con->data);
  for (i=0; i<panCodeobjName_HASHSIZE; i++)
    if (bytedata[i+4] != digest[i])
      return 0;
  return 1;
}


/* Writes a code object name to a panStream from a panCodeobjName 
 * and advances the stream.  Returns the number of bytes written.
 */
int
panCodeobjName_write(panStream *ps, panCodeobjName *coname)
{
  if (!coname || !ps) return 0;

  return panStream_write(ps, (u_int8_t*)(coname->data), 
			 panCodeobjName_SIZE);
}

/* Reads a code object name from a panStream into a panCodeobjName 
 * and advances the stream.  If testvalidity is true, checks to
 * see if the type and version are valid.
 */
int
panCodeobjName_read(panCodeobjName *coname, panStream *ps, int testvalidity)
{
  if (!coname || !ps) return 0;
  
  if (panCodeobjName_SIZE != panStream_read((u_int8_t*)(coname->data), 
					    ps, panCodeobjName_SIZE))
    { panPrint_warn("short read of panCodeobjName\n"); return 0; }

  if (testvalidity)
    {
      if (panCodeobjName_getType(coname) != panCodeobjName_TYPE_SHA1)
	{ panPrint_warn("Currently only SHA1 (0x%02x) panCodeobjNames are supported (got 0x%02x)\n", panCodeobjName_getType(coname), panCodeobjName_TYPE_SHA1);
	return 0; }
  
      if (panCodeobjName_getVers(coname) != panCodeobjName_VERS_SHA1)
	{ panPrint_warn("Warning SHA1 version mismatch (got 0x%4hx, current is 0x%4hx\n", panCodeobjName_getVers(coname), panCodeobjName_VERS_SHA1);
	return 0; }
    }

  return panCodeobjName_SIZE;
}

/* Scans through a code object and calculates its size.
 * Also ensures that the entire code object is available.
 * Does *not* advance ps.
 * Returns -1 if the software segment is shorter than the code object
 * is somehow corrupt.
 */
int
panCodeobj_getLength(const panStream *startps)
{
  unsigned  totalsize, sectsize;
  u_int16_t secttype;
  int       numsect, i; 
  panStream _ps, *ps;

  ps = &_ps;
  panStream_dup(ps, startps);

  totalsize = panCodeobjHead_SIZE;
  
  if (panCodeobjHead_NUMSECT_OFF
      != panStream_seekFwd(ps, panCodeobjHead_NUMSECT_OFF))
    { panPrint_error("panCodeobj_getLength: short header\n"); return -1; }

  numsect = panStream_read_net_uint16(ps);
  
  if (2 != panStream_seekFwd(ps, 2))
    { panPrint_error("panCodeobj_getLength: short header\n"); return -1; }
  
  for (i=0; i<numsect; i++)
    {
        secttype = panStream_read_net_uint16(ps);
        sectsize = panStream_read_net_uint16(ps);
	totalsize += 4 + sectsize;
	if (sectsize != panStream_seekFwd(ps, sectsize))
	  { panPrint_error("panCodeobj_getLength: short section\n");
	  return -1; }
    }
  return totalsize;
}


/* Reads the header of a code object from a panStream that is
 * located at the beginning of the code object.  When it
 * returns, the stream will have been advanced to the end
 * of the code object header.
 *
 * If testvalidity is true, also does some simple tests on the validity
 * of the code object being read in (magic number, etc.)
 * 
 * Returns 1 on success, 0 on an error.
 *
 */
int
panCodeobjHead_read(panCodeobjHead *coh, panStream *ps,
		    int testvalidity)
{
  int i;

  if (!ps || !coh) return 0;

  for (i=0; i<4; i++)
    {
      coh->magic[i] = panStream_read_uint8(ps);
      if (testvalidity && coh->magic[i] != panCodeobjHead_MAGIC[i])
	{ panPrint_warn("Magic number of codeobj does not match.\n");
	return 0; }
    }

  if (0 == panCodeobjName_read(&(coh->name), ps, testvalidity))
    return 0; 

  coh->code_type = panStream_read_net_uint16(ps);
  coh->code_vers = panStream_read_net_uint16(ps);
  coh->api_type  = panStream_read_net_uint16(ps);
  coh->api_vers  = panStream_read_net_uint16(ps);

  coh->num_sections  = panStream_read_net_uint16(ps);
  coh->padding       = panStream_read_net_uint16(ps);

  /* May want to do some verification of version numbering... */

  return 1;
}

/* Reads numsections sections into an array of code object sections
 * from a panStream and advances the stream to the end of them.
 * Returns the number of code objects read (or zero if an error occurs).
 */
int
panCodeobjSection_readArray(panCodeobjSection cosects[], panStream *ps, 
			    int numsections)
{
  int csi;			/* Index into cosects array */

  for (csi=0; csi < numsections; csi++)
    {
      cosects[csi].type   = panStream_read_net_uint16(ps);
      cosects[csi].length = panStream_read_net_uint16(ps);
      panStream_dup(&(cosects[csi].body), ps);
      if ((cosects[csi].type != panCodeobjSection_BINCODE)
	  && (cosects[csi].type != panCodeobjSection_DEPENDENCIES)
	  && (cosects[csi].type != panCodeobjSection_SYMBOLICNAME))
	{
	  panPrint_warn("panCodeobjSection_readArray: unknown section type \"%d\"\n", cosects[csi].type);
	}
      if (cosects[csi].length != panStream_seekFwd(ps, cosects[csi].length))
	{
	  panPrint_warn("panCodeobjSection_readArray: short segment read\n");
	  return csi;
	}
    }
  return csi;
}


/* Passed a code object name and a panStream that points to
 * the start of a code object that is cosize bytes long.
 * Hashes the code object and compares the hash to con.
 * If they match, returns 1.  Otherwise, returns 0.
 */
int 
panCodeobjName_verify(panCodeobjName *con, panStream *ps, int cosize)
{
  sha1_ctx   sha1context;
  u_int8_t   digestbuf[SHA1_SIZE];
  int i;

  panStream_seekFwd(ps, panCodeobjHead_STARTHASH_OFF); 

  if (panCodeobjName_getType(con) == panCodeobjName_TYPE_SHA1)
    {
      if ((panCodeobjName_getVers(con) & 0xff00) != 0x0100)
	{ panPrint_warn("don't know how to deal with SHA1 panCodeobjName version 0x%04x\n", panCodeobjName_getVers(con)); return(0); }
      sha1_init(&sha1context);
      panStream_sha1_update(&sha1context, ps, 
			    cosize-panCodeobjHead_STARTHASH_OFF);
      sha1_final(digestbuf, &sha1context);
#if PAN_DEBUG && 1
      {
	char namstrbuf[panCodeobjName_STRINGIFY_STRLEN], *pnsb=namstrbuf;
	pnsb += sprintf(pnsb, "%04x%04x",
			panCodeobjName_getType(con), 
			panCodeobjName_getVers(con));
	for (i=0; i<5; i++)
	  pnsb += sprintf(pnsb, " %02x%02x%02x%02x", 
			  digestbuf[i*4+0], digestbuf[i*4+1], 
			  digestbuf[i*4+2], digestbuf[i*4+3]);
	panPrint_debug("Generated hash during verify:\n");
	panPrint_debug("     [ %s ]\n", namstrbuf);
      }
#endif
      if (!panCodeobjName_compareHash(con, digestbuf))
	{ panPrint_warn("\tpanCodeobjName hash mismatch while verifying\n"); return(0); }	  
    }
  else
    {
      panPrint_warn("unknown panCodeobjName type 0x%04x\n", 
		    panCodeobjName_getType(con));
      return(0);
    }
  return(1);
}

/* Writes a string representing a code object name from con 
 * into buf in the form:
 *        "XXXXXXXX XXXXXXXX XXXXXXXX XXXXXXXX XXXXXXXX XXXXXXXX"
 */
void
panCodeobjName_stringify(char *buf, panCodeobjName *con)
{  
  int i;
  u_int8_t *bytedata;
  bytedata = (u_int8_t*)(con->data);
  for (i=0; i<6; i++)
    {
      buf += sprintf(buf, "%s%02x%02x%02x%02x", i?" ":"",
		      bytedata[i*4+0], bytedata[i*4+1], 
		      bytedata[i*4+2], bytedata[i*4+3]);
    }
  buf[panCodeobjName_STRINGIFY_STRLEN-1] = '\0';
}

/* Convert types to strings */
char *
panCodeobjName_type2string(u_int16_t type)
{
  if (type == panCodeobjName_TYPE_SHA1)
    return("SHA1");
  else if (type == panCodeobjName_TYPE_MD5)
    return("MD5");
  else
    return("unknown");
}
char *
panCodeobjHead_codetype2string(u_int16_t type)
{
  if (type == panCodeobjHead_CODETYPE_IX86BIN)
    return("ix86bin");
  else
    return("unknown");
}
char *
panCodeobjHead_apitype2string(u_int16_t type)
{
  if (type == panCodeobjHead_APITYPE_PAN)
    return("PAN");
  else if (type == panCodeobjHead_APITYPE_ANTS)
    return("ANTS");
  else
    return("unknown");
}

