/* $Header: /home/dl2/nygren/cvs/pan/src/common/pancodeobj.h,v 1.4 1997/12/01 22:15:59 nygren Exp $
 *
 * FILE:    pancodeobj.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Specifies the format of code objects that PAN transmits across the
 * network.  This format must clearly specify packing and byte
 * ordering in order to provide code mobility.
 * 
 * FILES DEPENDING ON THIS FORMAT:
 *
 *    panotool/pancodeobj_util.c
 */

#ifndef COMMON_PANCODEOBJ_H_INC
#define COMMON_PANCODEOBJ_H_INC

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swsstream.h>
#include <common/sha1.h>

/* These structures are meant for reference and 
 * storage.  No assumptions should be made about how the compiler
 * packs the data into this structure.  On the network, all
 * data must be in network byte order.  int8's are byte aligned,
 * int32's are 4-byte aligned, and int16's are 2-byte aligned.
 */

#define panCodeobjName_HASHSIZE 20  /* The length of the hash of a code 
				     * object (in bytes).  */
#define panCodeobjName_NUMWORDS 6   /* The number of 4-byte words in a 
				     * PAN code object's name */

#if panCodeobjName_HASHSIZE != (SHA1_SIZE)
#error "Code object name size changed from 20 bytes (sha1) & requires changes!"
#endif

#if panCodeobjName_HASHSIZE+4 != panCodeobjName_NUMWORDS*4
#error "panCodeobjName_NUMWORDS is invalid"
#endif


/* Code object names contain a type, a version, and a cryptographic
 * hash of the code object.  Initially, the hash is a 160-bit SHA-1 hash.  
 * This unpacked version shows what the fields are.  Instead of accessing
 * this, use panCodeobj and the accessor functions.  */
struct _panCodeobjNameUnpacked
{
#define panCodeobjName_TYPE_OFF   (0)
  u_int16_t   type;  /* The type of the code object name */
#define panCodeobjName_TYPE_SHA1 0x5348
#define panCodeobjName_TYPE_MD5  0x4d44  /* last 4 bytes of md5 name 
					  * will be padded with zeros */

#define panCodeobjName_VERS_OFF   (panCodeobjName_TYPE_OFF + 2)
  u_int16_t   vers;     /* The version number (major,minor) */
#define panCodeobjName_VERS_SHA1 0x0102	/* The current version */
#define panCodeobjName_VERS_MD5  0x0102	/* The current version */

#define panCodeobjName_HASH_OFF   (panCodeobjName_VERS_OFF + 2)
  u_int8_t    hash[panCodeobjName_HASHSIZE]; /* The hash of the code object */
};

/* Code object names are 24 bytes long but accessed as
 * 4-byte words to allow for fast-access.  No byte-order 
 * conversion is done (the array of u_int32_t's here just overlays an
 * array of u_int8_t's in network byteorder).  Use the 
 * access functions for all operations.
 */
typedef struct _panCodeobjName {
  u_int32_t       data[panCodeobjName_NUMWORDS];
} panCodeobjName;

#define panCodeobjName_SIZE (panCodeobjName_NUMWORDS*4) 
                                 /* The length of the hash name of a code 
				  * object (in bytes).  The first 4 bytes
				  * are a type and a version.  */

/* The header for code objects transmitted on the wire.
 * The header is followed by a number of sections
 * Containing the body of the code object. */
typedef struct _panCodeobjHead
{
#define panCodeobjHead_MAGIC_OFF   (0)
  int8_t         magic[4];	/* A four byte magic number */
#define panCodeobjHead_MAGIC "PAN\001"

#define panCodeobjHead_NAME_OFF (panCodeobjHead_MAGIC_OFF + 4)
  panCodeobjName name;		/* The name of this code object (the
				 * hash of everything following this point) */

#define panCodeobjHead_CODETYPE_OFF (panCodeobjHead_NAME_OFF + panCodeobjName_SIZE)
#define panCodeobjHead_STARTHASH_OFF panCodeobjHead_CODETYPE_OFF
  u_int16_t      code_type;	/* The type of code contained in this
				 * code object. */
#define panCodeobjHead_CODETYPE_IX86BIN 0x01

#define panCodeobjHead_CODEVERS_OFF  (panCodeobjHead_CODETYPE_OFF + 2)
  u_int16_t      code_vers;	/* Version of this code type (major,minor) */
#define panCodeobjHead_CODEVERS_IX86BIN 0x0102 /* Current version */

#define panCodeobjHead_APITYPE_OFF (panCodeobjHead_CODEVERS_OFF + 2)
  u_int16_t      api_type;	/* The type of API and run-time-system that
				 * this code object is designed to run in. */
#define panCodeobjHead_APITYPE_PAN   0x01
#define panCodeobjHead_APITYPE_ANTS  0x02

#define panCodeobjHead_APIVERS_OFF (panCodeobjHead_APITYPE_OFF + 2)
  u_int16_t      api_vers;	/* Version of this api type (major,minor) */
#define panCodeobjHead_APIVERS_PAN 0x0102 /* Current version */

#define panCodeobjHead_NUMSECT_OFF (panCodeobjHead_APIVERS_OFF + 2)
  u_int16_t      num_sections;  /* Number of sections that follow the header */

#define panCodeobjHead_PAD_OFF (panCodeobjHead_NUMSECT_OFF + 2)
  u_int16_t      padding;       /* For future use.  Must be set to 0. */

} panCodeobjHead;
#define panCodeobjHead_SIZE (panCodeobjHead_PAD_OFF + 2)

/* The header is followed by a series of sections.  Each section has a
 * u_int16_t section type followed by a u_int16_t for the number of
 * bytes in the body of the section.  The body of the section contains
 * that number of bytes (which must be a multiple of 4).
 */
typedef struct _panCodeobjSection {
  u_int16_t       type;		/* The type of the section */
  u_int16_t       length;	/* The number of bytes in the section
				 * (not including the header) */
  panStream       body;         /* The body of the section (really an array
				 * of length "length").  In this header,
				 * body is a panStream that is initially
				 * set to the start of the section. */
} panCodeobjSection;
#define panCodeobjSectionHead_SIZE 4

/* SECTION TYPES AND DESCRIPTIONS OF CONTENTS: */

/* The BINCODE section contains binary executable code in code
 * objects of type IX86BIN.  In the current version,
 * this binary code may be object files in any of the 
 * following formats:
 *    a.out
 *    ELF
 *    COFF
 *    ar archives
 * The code contained should be null-padded to have a length that
 * is a multiple of 4.
 */
#define panCodeobjSection_BINCODE 0x01

/* The DEPENDENCIES section contains a listing of the names
 * of code objects upon which this code object depends.
 * For a section of length N, there will be (N / panCodeobjName_SIZE)
 * dependencies contained within the section.
 */
#define panCodeobjSection_DEPENDENCIES 0x02

/* The SYMBOLICNAME section contains a textual name for this code
 * object for debugging purposes.  The body of the section should
 * be a null-terminated string padded with NULL's to have a length
 * that's a multiple of 4.
 */
#define panCodeobjSection_SYMBOLICNAME 0x03

/* Prototypes for routines for reading from code objects */

extern int 
panCodeobjName_read(panCodeobjName *coname, panStream *ps, int testvalidity);

extern int
panCodeobjName_write(panStream *ps, panCodeobjName *coname);

extern int 
panCodeobjHead_read(panCodeobjHead *coh, panStream *ps, int testvalidity);

extern int 
panCodeobjSection_readArray(panCodeobjSection cosects[], panStream *ps,
			    int numsections);

extern int
panCodeobj_getLength(const panStream *startps);

extern int 
panCodeobjName_verify(panCodeobjName *con, panStream *ps, int cosize);

/* Returns 1 if they're equal, 0 if they differ */
#define panCodeobjName_equalP(con1, con2) (0==memcmp((con1),(con2),  \
						  sizeof(panCodeobjName)))

#define panCodeobjName_dup(dst, src) (memcpy((dst), (src), \
					     sizeof(panCodeobjName)))

#define panCodeobjName_STRINGIFY_STRLEN (8*6+5+1)
extern void
panCodeobjName_stringify(char *buf, panCodeobjName *con);

extern u_int16_t
panCodeobjName_getType(panCodeobjName *con);
extern u_int16_t
panCodeobjName_getVers(panCodeobjName *con);
extern void
panCodeobjName_setType(panCodeobjName *con, u_int16_t type);
extern void
panCodeobjName_setVers(panCodeobjName *con, u_int16_t vers);
extern void
panCodeobjName_setHash(panCodeobjName *con, u_int8_t *digest);
extern int
panCodeobjName_compareHash(panCodeobjName *con, u_int8_t *digest);

extern char *
panCodeobjName_type2string(u_int16_t type);
extern char *
panCodeobjHead_codetype2string(u_int16_t type);
extern char *
panCodeobjHead_apitype2string(u_int16_t type);

#endif /* COMMON_PANCODEOBJ_H_INC */
