/* $Header: /home/dl2/nygren/cvs/pan/src/common/swseg.c,v 1.8 1997/12/18 02:33:22 nygren Exp $
 *
 * FILE:    swseg.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for software segment data structures and routines.
 * See PAN/doc/software-segments.txt for a detailed description
 * of software segments.  Be sure to read that file before trying to
 * use software segments.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#ifndef __KERNEL__
#include <sys/types.h>
#include <sys/mman.h>
#endif

/* Finalizes a panSwseg_TYPE_MALLOCED swseg by freeing the data and
 * the swseg itself.  Also unreferences any subsequent swseg's. 
 */
void
panSwsegMalloced_finalize(panSwseg *sws)
{
  if (sws)
    {
      if (sws->type != panSwseg_TYPE_MALLOCED)
	{ panPrint_error("panSwsegMalloced_finalize called with segment of type %d\n", sws->type); return; }
      if (sws->data) panMemory_free(sws->data);
      if (sws->next) panSwseg_unref(sws->next);
      panMemory_free(sws);
    }
}

/* Finalizes a panSwseg_TYPE_STATIC swseg by freeing 
 * the swseg itself.  Also unreferences any subsequent swseg's. 
 */
void
panSwsegStatic_finalize(panSwseg *sws)
{
  if (sws)
    {
      if (sws->type != panSwseg_TYPE_STATIC)
	{ panPrint_error("panSwsegStatic_finalize called with segment of type %d\n", sws->type); return; }
      if (sws->next) panSwseg_unref(sws->next);
      panMemory_free(sws);
    }
}

/* Finalizes a panSwseg_TYPE_OVERLAY swseg by unreferencing the overlain swseg
 * and freeing the swseg itself.  Also unreferences any subsequent swseg's.
 */
void
panSwsegOverlay_finalize(panSwseg *sws)
{
  if (sws)
    {
      if (sws->type != panSwseg_TYPE_OVERLAY)
	{ panPrint_error("panSwsegOverlay_finalize called with segment of type %d\n", sws->type); return; }
      if (((panSwsegOverlay*)sws)->overlain)
	panSwseg_unref(((panSwsegOverlay*)sws)->overlain); 
      if (sws->next)
	panSwseg_unref(sws->next);
      panMemory_free(sws);
    }
}

#ifndef __KERNEL__

/* Finalizes a panSwseg_TYPE_MMAPPED swseg by unmapping the memory and
 * freeing the swseg itself.  
 */
void
panSwsegMmaped_finalize(panSwseg *sws)
{
  if (sws)
    {
      if (sws->type != panSwseg_TYPE_MMAPPED)
	{ panPrint_error("panSwsegMmaped_finalize called with segment of type %d\n", sws->type); return; }
      if (((panSwsegMmaped*)sws)->mmapaddr)
	munmap(((panSwsegMmaped*)sws)->mmapaddr, 
	       ((panSwsegMmaped*)sws)->mmaplen);
      if (sws->next)
	panSwseg_unref(sws->next);
      panMemory_free(sws);
    }
}

#endif /* __KERNEL__ */


/* Finalizes a panSwseg_TYPE_CONTAINER swseg by freeing the data and
 * the swseg itself.  Also unreferences any subsequent swseg's. 
 * Runs any finalization routines that the container may have.
 */
void
panSwsegContainer_finalize(panSwseg *sws)
{
  if (sws)
    {
      if (sws->type != panSwseg_TYPE_CONTAINER)
	{ panPrint_error("panSwsegContainer_finalize called with segment of type %d\n", sws->type); return; }
      if (((panSwsegContainer*)sws)->finalize)
	(((panSwsegContainer*)sws)->finalize)(sws);
      if (sws->data) panMemory_free(sws->data);
      if (sws->next) panSwseg_unref(sws->next);
      panMemory_free(sws);
    }
}


/* Mallocs length bytes of memory and creates a software segment of
 * type panSwseg_TYPE_MALLOCED that contains the memory and has an
 * initial refcount of 1.  The returned swseg is guarenteed to contain
 * a single contiguous region of length bytes as its data field. 
 * Returns NULL if couldn't allocate the
 * memory.  */
panSwseg *
panSwseg_create_byMalloc(unsigned length)
{
  panSwseg *newsws;

  newsws = panSwseg_CAST(panMemory_alloc(sizeof(panSwsegMalloced)));
  if (!newsws)
    { panPrint_error("panSwseg_create_byMalloc: malloc returned null\n");
    return NULL; }
  newsws->data = panMemory_alloc(length);
  if (NULL == newsws->data)
    { panPrint_error("panSwseg_create_byMalloc: malloc returned null\n");
    return NULL; }
  newsws->length = length;
  newsws->type = panSwseg_TYPE_MALLOCED;
  newsws->numref = 1;
  newsws->next = NULL;
  newsws->finalize = panSwsegMalloced_finalize;

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(length)
      || ADDR_NOT_ALIGNED_P(newsws->data))
    panPrint_warn("panSwseg_create_byMalloc: length (%d) or data (%p) not aligned properly\n", length, newsws->data);
#endif

  return newsws;
}


/* Creates a swseg of type 
 * type panSwseg_TYPE_STATIC that contains datastart and has an
 * initial refcount of 1.  The returned swseg will be of length datalen.
 */
panSwseg *
panSwseg_create_static(char *datastart, int datalen)
{
  panSwseg *newsws;

  newsws = panSwseg_CAST(panMemory_alloc(sizeof(panSwseg)));
  if (!newsws)
    { panPrint_error("panSwseg_create_static: malloc returned null\n");
    return NULL; }
  newsws->data = datastart;
  newsws->length = datalen;
  newsws->type = panSwseg_TYPE_STATIC;
  newsws->numref = 1;
  newsws->next = NULL;
  newsws->finalize = panSwsegStatic_finalize;

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(newsws->length)
      || ADDR_NOT_ALIGNED_P(newsws->data))
    panPrint_warn("panSwseg_create_static: length (%d) or data (%p) not aligned properly\n", newsws->length, newsws->data);
#endif

  return newsws;
}



/* Chains s1 and s2 together so that s2 appears at the
 * end of s1.  If s1 already has a next swseg, attempts
 * to chain s2 onto the end of it (and does so recursively). 
 */
void
panSwseg_chain(panSwseg *s1, panSwseg *s2)
{
  panSwseg *swscur;

  assert(s1 && s2);
  while (s1->next != NULL) s1 = s1->next;

#define panSwseg_VERIFY_NO_CYCLES 
#if defined(PANDEBUG) || defined(panSwseg_VERIFY_NO_CYCLES)
  for (swscur = s2; swscur; swscur = swscur->next)
    {
      if (swscur == s1)
	{
	  panPrint_warn("WARNING: attempt to create cycle with panSwseg_chain!\n");
	}
    }
#endif

  panSwseg_ref(s2);
  s1->next = s2;
}

/* Returns the number of swsegs in the chain starting at sws.
 */
int
panSwseg_getChainLength(panSwseg *sws)
{
  int chainlen;

  for (chainlen=0; sws; sws = sws->next)
    {
      chainlen++;
    }
  return chainlen;
}

/* Returns the total number of bytes in the chain.
 */
unsigned
panSwseg_getBytesInChain(panSwseg *sws)
{
  int numbytes;

  for (numbytes=0; sws; sws = sws->next)
    {
      numbytes += sws->length;
    }
  return numbytes;
}

/* Returns the total size of the data in a software segment
 * (including any swsegs contained within it or chained after it).
 * There should probably be a method on swsegs to implement this
 * and getBytesInChain...  (Although this can be a wrapper for 
 * the method).
 */
unsigned
panSwseg_getSize(panSwseg *sws)
{
  int numbytes;

  for (numbytes=0; sws; sws = sws->next)
    {
      switch (sws->type)
	{
	case panSwseg_TYPE_SIMPLE:
	case panSwseg_TYPE_OVERLAY:
	case panSwseg_TYPE_MALLOCED:
	case panSwseg_TYPE_MMAPPED:
	  numbytes += sws->length;
	  break;
	case panSwseg_TYPE_APPLINK:
	  numbytes += panSwsegApplink_RESOURCE_SIZE;
	  break;
	case panSwseg_TYPE_CONTAINER:
	  if (((panSwsegContainer*)sws)->getsize)
	    numbytes += (((panSwsegContainer*)sws)->getsize)(sws);
	  break;
	default:
	  panPrint_error("panSwseg_getSize: unknown swseg type (%d)\n",
			 sws->type);
	}
    }
  return numbytes;
}

/* Creates a swseg overlaying sws that starts off bytes into sws
 * and extends for length bytes.  This may require creating
 * a chain of software segments in order to overlay a chain.
 */
panSwseg *
panSwseg_overlay(panSwseg *sws, unsigned off, unsigned length)
{
  panSwsegOverlay *olsws;

  if (!sws) return NULL;

  /* Seek forward to the swseg to start in */
  while (off >= sws->length)
    {
      off -= sws->length;
      sws = sws->next;
      if (!sws) return NULL;
    }

  olsws = panMemory_allocStruct(panSwsegOverlay);
  if (!olsws)
    { panPrint_error("panSwseg_overlay: malloc returned null\n");
    return NULL; }
  
  /* Create the first overlay segment */
  olsws->sws.data     = sws->data + off;
  olsws->sws.type     = panSwseg_TYPE_OVERLAY;
  olsws->sws.numref   = 1;
  olsws->sws.next     = NULL;
  olsws->sws.finalize = panSwsegOverlay_finalize;
  olsws->overlain     = sws;
  panSwseg_ref(sws);

  /* Fit as much in as we can and chain if we need to */
  if (off+length > sws->length) /* Need to chain on more overlays */
    {
      olsws->sws.length   = sws->length - off;
      if (sws->next)
	{
	  olsws->sws.next = panSwseg_overlay(sws->next, 0, 
					     length - olsws->sws.length);
	  panSwseg_ref(olsws->sws.next);	  
	}
    }
  else				/* It all fits into a single segment */
    {
      olsws->sws.length   = length;
    }

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(olsws->sws.length)
      || ADDR_NOT_ALIGNED_P(olsws->sws.data))
    panPrint_warn("panSwseg_overlay: length (%d) or data (%p) not aligned properly\n", olsws->sws.length, olsws->sws.data);
#endif

  return panSwseg_CAST(olsws);
}


#ifndef __KERNEL__

/* Creates a swseg that contains the region from fd mapped from
 * an offset of offset with a length of len.  See the mmap man 
 * page for more details of the arguments.
 * On finalization, the memory will be unmapped.
 */
panSwseg *
panSwseg_create_byMmap(size_t len, int prot, int flags, int fd, off_t offset)
{
  panSwsegMmaped *mmsws;
  caddr_t         mmapdest;

  mmsws = panMemory_allocStruct(panSwsegMmaped);
  if (!mmsws)
    { panPrint_error("panSwseg_create_byMmap: malloc returned null\n");
    return NULL; }

  mmapdest = mmap(0, len, prot, flags, fd, offset);
  if (mmapdest == (caddr_t)-1)
    { panPrint_error("panSwseg_create_byMmap: mmap returned null\n");
    panMemory_free(mmsws); return NULL; }

  /* Create the first overlay segment */
  mmsws->sws.data     = (u_int8_t*)mmapdest;
  mmsws->sws.type     = panSwseg_TYPE_MMAPPED;
  mmsws->sws.numref   = 1;
  mmsws->sws.length   = len;
  mmsws->sws.next     = NULL;
  mmsws->sws.finalize = panSwsegMmaped_finalize;
  mmsws->fd           = fd;
  mmsws->mmapaddr     = mmapdest;
  mmsws->mmaplen      = len;
  

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(mmsws->sws.length)
      || ADDR_NOT_ALIGNED_P(mmsws->sws.data))
    panPrint_warn("panSwseg_create_byMmap: length (%d) or data (%p) not aligned properly\n", mmsws->sws.length, mmsws->sws.data);
#endif

  return panSwseg_CAST(mmsws);
}

#endif /* __KERNEL__ */



/* Mallocs length bytes of memory and creates a software segment of
 * type panSwseg_TYPE_CONTAINER that contains the memory and has an
 * initial refcount of 1.  The returned swseg is guarenteed to contain
 * a single contiguous region of length bytes as its data field. 
 * Returns NULL if couldn't allocate the memory.  */
panSwseg *
panSwsegContainer_create(unsigned length)
{
  panSwsegContainer *newsws;

  newsws = panMemory_allocStruct(panSwsegContainer);
  if (!newsws)
    { panPrint_error("panSwsegContainer_create: malloc returned null\n");
    return NULL; }
  newsws->sws.data = panMemory_alloc(length);
  if (NULL == newsws->sws.data)
    { panPrint_error("panSwsegContainer_create: malloc returned null\n");
    return NULL; }
  newsws->sws.length = length;
  newsws->sws.type = panSwseg_TYPE_CONTAINER;
  newsws->sws.numref = 1;
  newsws->sws.next = NULL;
  newsws->sws.finalize = panSwsegContainer_finalize;

  newsws->getsize = NULL;
  newsws->release = NULL;
  newsws->finalize = NULL;

#if PAN_DEBUG
  if (ADDR_NOT_ALIGNED_P(length)
      || ADDR_NOT_ALIGNED_P(newsws->sws.data))
    panPrint_warn("panSwsegContainer_create: length (%d) or data (%p) not aligned properly\n", length, newsws->sws.data);
#endif

  return (panSwseg*)newsws;
}



/* Incriments the reference count on sws.
 */
void
panSwseg_ref(panSwseg *sws)
{
  panCritRegion_ENTER();	/* Just to be safe... */
  if (sws) sws->numref++;
  panCritRegion_LEAVE();
}

/* Decriments the reference count on sws and frees if reaches 0.
 */
void
panSwseg_unref(panSwseg *sws)
{
  if (!sws) return;
  {
    panCritRegion_ENTER();
    if (sws->numref == 1)
      {
	sws->numref=0;
	panCritRegion_LEAVE();
	sws->finalize(sws);
      }
    else if (sws->numref == 0)
      {
	panPrint_error("panSwseg_unref: "
		       "unreferencing a swseg with 0 numref\n");
	panCritRegion_LEAVE();
      }
    else
      {
	sws->numref--;
	panCritRegion_LEAVE();
      }
  }
}



