/* $Header: /home/dl2/nygren/cvs/pan/src/common/swseg.h,v 1.8 1997/12/13 05:12:25 nygren Exp $
 *
 * FILE:    swseg.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Header file for software segment data structures and routines.
 * See PAN/doc/software-segments.txt for a detailed description
 * of software segments.  Be sure to read that file before trying to
 * use software segments.
 *
 */

#ifndef PAN_COMMON_SWSEG_H_INC
#define PAN_COMMON_SWSEG_H_INC

#include <common/pan-copyright.h>
#include <common/sysconfig.h>


/* The structure for a basic software segment.  Other types of software
 * segments can subclass this by adding fields to the end.  */
typedef struct _panSwseg {

  u_int8_t            *data;	/* The address of the region of memory
				 * that the swseg overlays/handles. */

  unsigned             length;	/* The length of the software segment's
				 * region in bytes. */

  unsigned short       type;	/* The type of this software segment */

  unsigned short       numref;	/* The reference count (how many subsystems 
				 * and software segments are referencing 
				 * this swseg) */

  struct _panSwseg    *next;	/* The next software segment 
				 * in a chain of swsegs (or NULL). */

  void (*finalize)(struct _panSwseg *); /* A routine to finalize this type of
				 * swseg.  Passed a pointer to the software  
				 * segment. Must unreference any swsegs this
				 * one references or is chained to and free the
				 * memory for the swseg, and, if 
				 * approprate, free the underlying memory. 
				 * We may want to replace both this
				 * pointer and the type field 
				 * with a single pointer to a class 
				 * structure and dispatch table.  */
} panSwseg;

/* Types of software segments.  These values may be stored
 * in the "type" field. */

/* A simple panSwseg with no extra fields.  The finalizer must
 * unreference the swseg pointed to by next and free the software
 * segment, but shouldn't touch underlying memory.  This should almost
 * never be used.  */
#define panSwseg_TYPE_SIMPLE 0


/* An overlay software segment overlays a subset of another software
 * segments data.  On finalization it should also unreference the
 * software segment that is overlaid.  */
#define panSwseg_TYPE_OVERLAY 1
typedef struct _panSwsegOverlay {
  panSwseg           sws;	/* The main body of the swseg */
  panSwseg          *overlain;  /* The software segment that this overlays */
} panSwsegOverlay;


/* A software segment containing memory allocated by malloc (or vmalloc
 * or whatever else the heap memory allocation routine is).  The segment
 * data should be freed at the same that the swseg is freed.  */
#define panSwseg_TYPE_MALLOCED 2
typedef struct _panSwsegMalloced {
  panSwseg           sws;	/* The main body of the swseg */
} panSwsegMalloced;


/* A software segment that overlays mmapped memory.  The memory
 * should be freed and the file descriptor closed when the 
 * software segment is finalized.  */
#define panSwseg_TYPE_MMAPPED 3
typedef struct _panSwsegMmaped {
  panSwseg           sws;	/* The main body of the swseg */
  int                fd;	/* The file descriptor */
  caddr_t            mmapaddr;	/* The addresses at which the mapped
				 * region starts. */
  size_t             mmaplen;	/* The size of the mmapped region. */
} panSwsegMmaped;


/* A software segment containing a static region of memory
 * (such as some data in the program's text segment.
 * The segment data should be treated as read-only.
 */
#define panSwseg_TYPE_STATIC 4


/* A applink **reference-only** software segment is used by nodes to
 * keep track of references to application links.  This way, we can
 * keep track of whether anyone is using the connection, allowing us
 * to eventually expunge closed connections.  */
#define panSwseg_TYPE_APPLINK 5
#define panSwsegApplink_RESOURCE_SIZE 128 /* How many bytes to count as */


/* A software segment that can contain other software segments
 * in an application-dependant manner.  The application is responsible
 * for doing reference counting and ensuring that there are no cycles.
 */
#define panSwseg_TYPE_CONTAINER 6
typedef struct _panSwsegContainer {
  panSwseg           sws;	/* The main body of the swseg */
  int (*getsize)(struct _panSwseg *); /* A routine that returns the total
					* size of all the swsegs in the
					* container. */
  int (*release)(struct _panSwseg *, int); /* A routine that takes the
					    * container and releases at least
					    * specified total size worth
					    * of swsegs.  Returns the total
					    * size of swsegs that were 
					    * released. */
  void (*finalize)(struct _panSwseg *); /* An application-provided 
					 * routine to finalize the
					 * contents of this container. */
} panSwsegContainer;


/* Only exists within Linux kernel nodes.  Contains a reference
 * to a kernel skbuff.  */
#define panSwseg_TYPE_SKBUFF 7



/* Casts a subtype of a swseg into a panSwseg. */
#define panSwseg_CAST(sws)  ((panSwseg*)(sws))

/* Prototypes for panSwseg routines */

extern panSwseg *
panSwseg_create_byMalloc(unsigned length);

extern panSwseg *
panSwseg_create_static(char *datastart, int datalen);

#ifndef __KERNEL__
extern panSwseg *
panSwseg_create_byMmap(size_t len, int prot, int flags, int fd, off_t offset);
#endif

extern void
panSwseg_chain(panSwseg *s1, panSwseg *s2);

extern int
panSwseg_getChainLength(panSwseg *sws);

extern unsigned
panSwseg_getBytesInChain(panSwseg *sws);

extern unsigned
panSwseg_getSize(panSwseg *sws);

extern panSwseg *
panSwseg_overlay(panSwseg *sws, unsigned off, unsigned length);

extern panSwseg *
panSwsegContainer_create(unsigned length);

extern void
panSwseg_ref(panSwseg *sws);

extern void
panSwseg_unref(panSwseg *sws);


/****** Software segment rings ******/
/*
 * Software segment rings contain a collection of one or more sections.
 * A section is a region of memory containing a number of entries.
 * Entries may either be unused, may contain pointers to swsegs,
 * or may be members of the free list in which case they'll
 * contain pointers to the next entry in the freelist. 
 * If an entry is "unused", all entries that follow it will
 * also be unused.  The freelist is a linked list of 
 * entries that are available for use.
 *
 * When a swseg is put into the ring, it'll be put into the head
 * entry of the freelist.  If the freelist is empty, it will
 * be put into the next unused entry in the last section.
 * A new section will be created if needed.
 *
 * When a swseg is removed, all entries are searched through
 * until it is found (upto the first unused entry) and the 
 * entry it was found in is added to the free list.
 *
 * Consequences:  The space used is O(max size of swsring)
 *                Time for adding to the ring is O(1)
 *                Time for removing from the ring is O(max size of ring)
 *
 * Note that a ring will never shrink past its maximum size.  However,
 * a ring will not grow in size if there are any slots available in
 * the freelist.
 *
 */

typedef struct _panSwsegRingEntry {
  int type;			/* Is this a swseg or in the freelist? */
#define  panSwsegRingEntry_TYPE_UNUSED   0
#define  panSwsegRingEntry_TYPE_SWSEG    1
#define  panSwsegRingEntry_TYPE_FREELIST 2
  union 
  {
    panSwseg *sws;		        /* A pointer to a swseg */
    struct _panSwsegRingEntry *flnext;	/* The next entry in the freelist */
  } e;
} panSwsegRingEntry;

typedef struct _panSwsegRingSection {
  struct _panSwsegRingSection *next; /* Pointer to the next section */
  int                          size; /* Size of this section */
  panSwsegRingEntry        *entries; /* Pointer to array of entries */
} panSwsegRingSection;

typedef struct _panSwsegRing {
  panSwsegRingSection *first;	/* First section */
  panSwsegRingSection *last;	/* Last section (where to get new entries) */
  int nextfree;			/* Next free entry in the last section */
  panSwsegRingEntry *freelist; 	/* First element in the free list */
} panSwsegRing;



extern panSwsegRing *
panSwsegRing_create(int initsize);

extern void
panSwsegRing_add(panSwsegRing *swsring, panSwseg *sws);

extern int
panSwsegRing_remove(panSwsegRing *swsring, panSwseg *sws, int all);

extern void
panSwsegRing_clearAll(panSwsegRing *swsring);

extern void
panSwsegRing_destroy(panSwsegRing *swsring);

extern int
panSwsegRing_numEntries(panSwsegRing *swsring);


#endif /* PAN_COMMON_SWSEG_H_INC */
