/* $Header: /home/dl2/nygren/cvs/pan/src/common/swsegring.c,v 1.4 1997/12/18 02:33:22 nygren Exp $
 *
 * FILE:    swsegring.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for software segment ring data structures and routines.
 * See the swseg.h header file for prototypes and for documentation
 * of the data structure semantics and invariants.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <sys/types.h>



/* 
 * Creates a new, empty software segment ring with the specified
 * initial number of entries and returns it.
 */
panSwsegRing *
panSwsegRing_create(int initsize)
{
  panSwsegRing *sr;
  
  if (initsize <= 1) 
    { 
      panPrint_error("panSwsegRing_create: min init size is 1, passed %d\n",
		     initsize);
      return NULL;
    }
  sr = panMemory_alloc(sizeof(panSwsegRing));
  if (!sr)
    { panPrint_error("panSwsegRing_create: couldn't malloc!\n"); return NULL; }
  sr->first = panMemory_alloc(sizeof(panSwsegRingSection)
			      + initsize*sizeof(panSwsegRingEntry));
  if (!sr->first)
    { panPrint_error("panSwsegRing_create: couldn't malloc!\n"); return NULL; }
  sr->last = sr->first;
  sr->nextfree = 0;
  sr->freelist = NULL;

  sr->first->size = initsize;
  sr->first->next = NULL;
  sr->first->entries = (panSwsegRingEntry*)(sr->first+1);
  memset(sr->first->entries, 0, sr->first->size*sizeof(panSwsegRingEntry));

  return sr;
}


/* 
 * Adds a software segment sws to the software segment ring.  
 * Allocates more memory if needed.
 */
void
panSwsegRing_add(panSwsegRing *ring, panSwseg *sws)
{
  assert(ring && sws);

  /* Use an entry from the freelist if possible */
  if (ring->freelist)
    {
      panSwsegRingEntry *nextfree = ring->freelist->e.flnext;
      assert (ring->freelist->type == panSwsegRingEntry_TYPE_FREELIST);
      ring->freelist->e.sws = sws;
      ring->freelist->type = panSwsegRingEntry_TYPE_SWSEG;
      ring->freelist = nextfree;
      return;
    }
  
  /* Create a new section if needed */
  if (ring->nextfree >= ring->last->size)
    {
      panSwsegRingSection *newsect;
      int newsectsize = 2 * ring->last->size;
      
      newsect = panMemory_alloc(sizeof(panSwsegRingSection)
				+ newsectsize*sizeof(panSwsegRingEntry));
      if (!newsect)
	{ panPrint_error("panSwsegRing_add: couldn't malloc!\n"); return; }
      newsect->size = newsectsize;
      newsect->next = NULL;
      newsect->entries = (panSwsegRingEntry*)(newsect+1);
      memset(newsect->entries, 0, newsectsize*sizeof(panSwsegRingEntry));

      ring->last->next = newsect;
      ring->last = newsect;
      ring->nextfree = 0;
    }

  ring->last->entries[ring->nextfree].type = panSwsegRingEntry_TYPE_SWSEG;
  ring->last->entries[ring->nextfree++].e.sws = sws;

}


/* 
 * Clears all of the elements in a software segment ring
 * (unreferencing the swsegs and setting the ring's size to 0).
 * Eliminates all sections except for the first one.
 */
void
panSwsegRing_clearAll(panSwsegRing *swsring)
{
  panSwsegRingSection *sect, *nextsect;
  int i, j;

  for (j=0, sect = swsring->first; sect; sect = sect->next, j++)
    {
      for (i=0; i < sect->size; i++)
	{
	  if (sect->entries[i].type == panSwsegRingEntry_TYPE_SWSEG)
	    {
	      panPrint_debug("Freed swseg %p from ring %p (sect=%d,pos=%d)\n", 
			     sect->entries[i].e.sws, swsring, j, i);
	      panSwseg_unref(sect->entries[i].e.sws);	 
	      sect->entries[i].type = panSwsegRingEntry_TYPE_UNUSED;
	    }
	  else if (sect->entries[i].type == panSwsegRingEntry_TYPE_UNUSED)
	    goto no_more_entries_to_clear;
	}
    }
no_more_entries_to_clear:

  for (sect = swsring->first->next; sect; sect = nextsect)
    {
      nextsect = sect->next;
      panMemory_free(sect);
    }
  swsring->first->next = NULL;
  swsring->last = swsring->first;
  swsring->nextfree = 0;
  swsring->freelist = NULL;
}


/* Destroys a swseg ring and unreferences all of the swsegs in it */
void
panSwsegRing_destroy(panSwsegRing *swsring)
{
  panSwsegRing_clearAll(swsring);
  panMemory_free(swsring->first);
  panMemory_free(swsring);  
}


/* Removes a swseg from a ring and unreferences it.  If "all" is true,
 * removes all occurances of the swseg, otherwise only removes the
 * first one.  Puts the entry/entries onto the free list. 
 * Returns the number of swseg references removed. */
int
panSwsegRing_remove(panSwsegRing *ring, panSwseg *sws, int all)
{
  panSwsegRingSection *sect;
  int i, j, findcount=0;

  for (j=0, sect = ring->first; sect; sect = sect->next, j++)
    {
      for (i=0; i < sect->size; i++)
	{
	  if ((sect->entries[i].type == panSwsegRingEntry_TYPE_SWSEG)
	      && (sect->entries[i].e.sws == sws))
	    {
	      panPrint_debug("Removed swseg %p (type %d) from ring %p "
			     "(sect=%d,pos=%d)\n", 
			     sect->entries[i].e.sws, 
			     sect->entries[i].e.sws->type,
			     ring, j, i);
	      panSwseg_unref(sect->entries[i].e.sws);	
	      sect->entries[i].type = panSwsegRingEntry_TYPE_FREELIST;
	      sect->entries[i].e.flnext = ring->freelist;
	      ring->freelist = &(sect->entries[i]);
	      findcount++;
	      if (!all) goto no_more_entries_to_remove;
	    }
	  else if (sect->entries[i].type == panSwsegRingEntry_TYPE_UNUSED)
	    goto no_more_entries_to_remove;
	}
    }
no_more_entries_to_remove:
  return findcount;
}


/* 
 * Returns the number of swsegs in a software segment ring.
 */
int
panSwsegRing_numEntries(panSwsegRing *swsring)
{
  panSwsegRingSection *sect;
  int i, numentries=0;

  for (sect = swsring->first; sect; sect = sect->next)
    {
      for (i=0; i < sect->size; i++)
	{
	  if (sect->entries[i].type == panSwsegRingEntry_TYPE_SWSEG)
	    {
	      numentries++;
	    }
	  else if (sect->entries[i].type == panSwsegRingEntry_TYPE_UNUSED)
	    goto no_more_entries_to_count;
	}
    }
no_more_entries_to_count:

  return numentries;
}
