/* $Header: /home/dl2/nygren/cvs/pan/src/common/swsstream.c,v 1.4 1997/12/10 06:58:24 nygren Exp $
 *
 * FILE:    swsstream.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for software segment stream data structures and routines.
 * See PAN/doc/software-segments.txt for a detailed description
 * of software segments.  Be sure to read that file before trying to
 * use software segment streams.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>

#undef  panStream_WORDIO_MAKEINLINE
#undef  panStream_WORDIO_MAKEPROTO
#undef  pniStream_WORDIO_MAKEPROTO
#define panStream_WORDIO_MAKEBODY
#include <common/swsstream_wordio.h>

#define MIN(x,y) ((x)>(y)?(y):(x))

/*
 * Initializes the panStream object ps to start pos bytes into sws.
 * (if pos==0, starts at start os sws)
 */
void
panStream_init(panStream *ps, panSwseg *sws, unsigned pos)
{
  while ((pos > sws->length) && sws)
    {
      pos -= sws->length;
      sws = sws->next;
    }
  assert(ps&&sws);
  ps->pos = sws->data + pos;
  ps->rem = sws->length - pos;
  ps->cursws = sws;     
}


/*
 * Seeks the current position forward by offset bytes and returns
 * the number of bytes that were successfully seeked.
 */
unsigned
panStream_seekFwd(panStream *ps, unsigned offset)
{
  unsigned numseeked=0;
  while (1 /*offset > 0*/) 
    {
      if (ps->rem < offset)
	{
	  if (ps->cursws->next)	/* Skip to next sws */
	    {
	      numseeked += ps->rem;
	      offset    -= ps->rem;
	      ps->cursws = ps->cursws->next;
	      ps->rem = ps->cursws->length;	      
	      ps->pos = ps->cursws->data;
	    }
	  else			/* There is no next segment.  Seek to 
				 * end and return.  */
	    {
	      numseeked += ps->rem;
	      ps->rem = 0;
	      ps->pos += ps->rem;
	      return numseeked;
	    }
	}
      else			/* Seek to somewhere in current swseg */
	{
	  numseeked += offset;
	  ps->rem -= offset;
	  ps->pos += offset;
	  /* offset = 0; */
	  return numseeked;
	}
    }
  return 0;
}


/*
 * Returns the current software segment and stores the position
 * of the stream into pos.
 */
panSwseg *
panStream_getState(panStream *ps, unsigned *curpos)
{
  if (!ps) return NULL;
  *curpos = ps->cursws->length - ps->rem;
  return ps->cursws;
}

/*
 * Returns how many bytes remain between the current position in
 * swseg stream and the end of the stream.  Also advances to the
 * position of ps to the end of the stream.
 */
unsigned
panStream_getBytesRemaining(panStream *ps)
{
  unsigned count=0;

  if (!ps) return 0;
  while (!((ps->cursws->next == NULL) && (ps->rem == 0)))
    {
      if (ps->cursws->next == NULL)
	{
	  count   += ps->rem;
	  ps->rem  = 0;
	}
      else
	{
	  count     += ps->rem;
	  ps->cursws = ps->cursws->next;
	  ps->rem    = ps->cursws->length;
	}
    }
  ps->pos = ps->cursws->data + ps->cursws->length;
  return count;
}

/*
 * Copies length bytes from pssrc into psdest and returns
 * the number of bytes copied (which may be less than length
 * if either is shorter than length).  Seeks forward both
 * streams by length.
 */
unsigned 
panStream_copy(panStream *psdest, panStream *pssrc, unsigned length)
{
  unsigned numcopied = 0;
  unsigned tocopy;		/* bytes to copy */

  while (length > 0)
    {
      if ((pssrc->rem == 0 && !pssrc->cursws->next) 
	  || (psdest->rem == 0 && !psdest->cursws->next))
	return numcopied;

      if (pssrc->rem == 0)
	{
	  pssrc->cursws = pssrc->cursws->next;
	  pssrc->rem    = pssrc->cursws->length;
	      pssrc->pos    = pssrc->cursws->data;
	}
      if (psdest->rem == 0)
	{
	  psdest->cursws = psdest->cursws->next;
	  psdest->rem    = psdest->cursws->length;
	  psdest->pos    = psdest->cursws->data;
	}

      tocopy = MIN(pssrc->rem, psdest->rem);
      tocopy = MIN(tocopy, length);
      
      memcpy(psdest->pos, pssrc->pos, tocopy);
      pssrc->rem  -= tocopy;
      psdest->rem -= tocopy;
      pssrc->pos  += tocopy;
      psdest->pos += tocopy;
      numcopied   += tocopy;
      length      -= tocopy;	  
    }
  return numcopied;
}

/* Copies length bytes from ps into the memory region starting at
 * dest.  Returns the number of bytes copied (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  
 */
panStream_GEN_SCAN_FN(
      panStream_read(u_int8_t *dest, panStream *ps, unsigned length),
      { if (!dest || !ps) return 0; },
      { memcpy(dest, ps->pos, numscan);
        dest += numscan;
      })

/* Copies length bytes from the memory region starting at
 * src into ps.  Returns the number of bytes copied (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  
 */
panStream_GEN_SCAN_FN(
      panStream_write(panStream *ps, u_int8_t *src, unsigned length),
      { if (!src || !ps) return 0; },
      {	memcpy(ps->pos, src, numscan);
        src += numscan;
      })

/* Runs sha1_update over the memory regions in a swsstream ps and advances
 * the stream.  Passed an initialized sha1_ctx structure to store 
 * context info in.
 */
panStream_GEN_SCAN_FN(
      panStream_sha1_update(sha1_ctx *ctx, panStream *ps, unsigned length),
      { if (!ctx || !ps) return 0; },
      { sha1_update(ctx, ps->pos, numscan); })

/* Copies length bytes from pssrc into the memory region starting at
 * dest.  Returns the number of bytes copied (which may be less than
 * length if the stream is shorter than length).  Seeks forward the pssrc
 * stream by length.  
 */
unsigned 
panStream_read_old(u_int8_t *dest, panStream *pssrc, unsigned length)
{
  unsigned numcopied = 0;
  unsigned numscan;

  if (!dest || !pssrc) return 0;

  while (length > 0)
    {
      if (pssrc->rem == 0)
	{
	  if (!pssrc->cursws->next) 
	    return numcopied;
	  else
	    {
	      pssrc->cursws = pssrc->cursws->next;
	      pssrc->rem    = pssrc->cursws->length;
	      pssrc->pos    = pssrc->cursws->data;
	    }
	}
      numscan = (length <= pssrc->rem)?(length):(pssrc->rem);
      memcpy(dest, pssrc->pos, numscan);
      dest        += numscan;
      numcopied   += numscan;
      length      -= numscan;
      pssrc->rem  -= numscan;
      pssrc->pos  += numscan;	  
    }
  return numcopied;
}

#if 0

/*
 * Returns whether a panStream is at the end of the software segment chain.
 * (Returns 1 iff at end, else returns 0).
 */
int
panStream_eofP(panStream *ps)
{
  if (!ps || (ps->rem == 0) || !(ps->cursws))
    return 1;
  else
    return 0;
}

u_int32_t 
panStream_read_msbfirst_uint32(panStream *ps)
{
  if (ps->rem >= 4)
    {
      ps->rem -= 4;
      return ntohl(*(((u_int32_t*)(ps->pos))++));
    }
  else if (ps->rem != 0)
    {
      panPrint_error("panStream_read_msbfirst_uint32: unaligned read\n");
      return 0;
    }
  else if (ps->cursws->next)
    {
      ps->cursws = ps->cursws->next;
      ps->rem = ps->cursws->length - 4;
      ps->pos = ps->cursws->data;
      return ntohl(*(((u_int32_t*)(ps->pos))++));
    }
  else
    {
      panPrint_error("panStream_read_msbfirst_uint32: read of empty stream\n");
      return 0;
    }
}

#endif
