/* $Header: /home/dl2/nygren/cvs/pan/src/common/swsstream.h,v 1.4 1997/12/13 05:12:26 nygren Exp $
 *
 * FILE:    swsstream.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Header file for software segment data streams.  These provide an
 * interface for sequentially reading data out of a software segment
 * in an efficient manner.  See PAN/doc/software-segments.txt for a
 * detailed description of software segments.  Be sure to read that
 * file before trying to use software segments.
 *
 */

#ifndef PAN_COMMON_SWSSTREAM_H_INC
#define PAN_COMMON_SWSSTREAM_H_INC

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/sha1.h>


/* The structure to hold a swseg stream.  This structure should
 * be lightweight as possible and should essentially represent the
 * location in a software segment. */
typedef struct _panStream {
  u_int8_t      *pos;		/* A pointer to the current position 
				 * (where to next read from)
				 * in the software segment */
  unsigned       rem;		/* The number of bytes remaining before 
				 * the end of the swseg */
  panSwseg      *cursws;	/* The header for the current software segment
				 * (into which pos points) */
} panStream;


extern void panStream_init(panStream *ps, panSwseg *sws, unsigned pos);
extern unsigned panStream_seekFwd(panStream *ps, unsigned offset);
extern unsigned panStream_getBytesRemaining(panStream *ps);
extern panSwseg *panStream_getState(panStream *ps, unsigned *curpos);
extern unsigned panStream_copy(panStream *psdest, panStream *pssrc, unsigned length);
extern int panStream_read(u_int8_t *dest, panStream *ps, unsigned length);
extern int panStream_write(panStream *ps, u_int8_t *src, unsigned length);
extern int panStream_sha1_update(sha1_ctx *ctx, panStream *ps, unsigned length);


/* Duplicates the state of an swsstream */
#define panStream_dup(psdst,pssrc) memcpy((psdst),(pssrc),(sizeof(panStream)))

#if !defined(PAN_CODEOBJ)
#undef  panStream_WORDIO_MAKEBODY
#undef  panStream_WORDIO_MAKEPROTO
#undef  pniStream_WORDIO_MAKEPROTO
#define panStream_WORDIO_MAKEINLINE
#include <common/swsstream_wordio.h>
#endif /* !defined(PAN_CODEOBJ) */

/* The above includes the following routines:

   int panStream_eofP(panStream *ps);

   [u_]int{8,16,32}_t panStream_read_[u]int{8,16,32}(panStream *ps);
   void panStream_write_[u]int{8,16,32}(panStream *ps, [u_]int{8,16,32}_t x);

 */


/* Generates code to iterate over a panStream
 * and perform an operation such as a swseg chain -> memory copy
 * or a sha1 checksum.
 *
 * The macro can be provided with:
 *    PROTO    -- The function prototype's (name and args)
 *                The function MUST at least take two arguments:
 *                      panStream *ps  -- The swseg stream to scan through
 *                      unsigned   length -- The number of bytes to scan
 *    SCANBLK  -- The procedure (or block) to run on all the scanned memory.
 *                Available to it are:
 *                      ps->pos     The address of the start of a region
 *                      numscan     The number of bytes available at sws->pos
 *                      any other variables in the prototype
 *    INITBLK  -- A block at the beginning to set up initial state
 */
#define panStream_GEN_SCAN_FN(PROTO, INITBLK, SCANBLK) \
int                                                    \
PROTO                                                  \
{                                                      \
  int numcopied = 0;                                   \
  int numscan;                                         \
                                                       \
  INITBLK			                       \
                                                       \
  while (length > 0)                                   \
    {                                                  \
      if (ps->rem == 0)                                \
	{                                              \
	  if (!ps->cursws->next)                       \
	    return numcopied;                          \
	  else                                         \
	    {                                          \
	      ps->cursws = ps->cursws->next;           \
	      ps->rem    = ps->cursws->length;         \
	      ps->pos    = ps->cursws->data;           \
	    }                                          \
	}                                              \
      numscan = (length <= ps->rem)?(length):(ps->rem);\
      SCANBLK                                          \
      numcopied   += numscan;                          \
      length      -= numscan;                          \
      ps->rem     -= numscan;                          \
      ps->pos     += numscan;	                       \
    }                                                  \
  return numcopied;                                    \
}                         



#endif /* PAN_COMMON_SWSSTREAM_H_INC */
