/* $Header: /home/dl2/nygren/cvs/pan/src/common/swsstream_wordio.h,v 1.3 1997/12/01 22:16:01 nygren Exp $
 *
 * FILE:    swsstream_wordio.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for reading and writing words to and from software segment streams.
 * Generates different code depending on the values of the
 * following preprocessor macros:
 *
 *      panStream_WORDIO_MAKEINLINE    Makes inline functions
 *      panStream_WORDIO_MAKEPROTO     Makes prototypes
 *      pniStream_WORDIO_MAKEPROTO     Makes prototypes for PAN Node Interface
 *      panStream_WORDIO_MAKEBODY      Makes C code
 *
 * Only one may be per inclusion of this file.
 * All three are #undef'ed at the end.
 * 
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swsstream.h>


#undef panStream_WORDIO_INLINE 
#ifdef panStream_WORDIO_MAKEINLINE 
#define panStream_WORDIO_INLINE extern __inline__
#else
#define panStream_WORDIO_INLINE 
#endif

/*********** MAKE NOOP MACROS ***************/
#if defined(pniStream_WORDIO_MADEPROTOS_ALREADY)

#define __panStream_make_read(FNNAME, BYTES, TYPE, CONVERTFN)
#define __panStream_make_write(FNNAME, BYTES, TYPE, CONVERTFN) 

/*********** PAN PROTOTYPE ONLY ***************/
#elif defined(panStream_WORDIO_MAKEPROTO)

#define __panStream_make_read(FNNAME, BYTES, TYPE, CONVERTFN) \
    extern TYPE panStream_##FNNAME(panStream *ps);
#define __panStream_make_write(FNNAME, BYTES, TYPE, CONVERTFN) \
    extern void panStream_##FNNAME(panStream *ps, TYPE x);

extern int panStream_eofP(panStream *ps);

/*********** PNI PROTOTYPE ONLY ***************/
#elif defined(pniStream_WORDIO_MAKEPROTO) && defined(PAN_CODEOBJ)

#define pniStream_WORDIO_MADEPROTOS_ALREADY
#define __panStream_make_read(FNNAME, BYTES, TYPE, CONVERTFN) \
    extern TYPE pniStream_##FNNAME(pniStream *ps);
#define __panStream_make_write(FNNAME, BYTES, TYPE, CONVERTFN) \
    extern void pniStream_##FNNAME(pniStream *ps, TYPE x);

extern int pniStream_eofP(pniStream *ps);

/************** BODY OR INLINED BODY *****************/
#elif defined(panStream_WORDIO_MAKEBODY) || defined(panStream_WORDIO_MAKEINLINE)

/*
 * Returns whether a panStream is at the end of the software segment chain.
 * (Returns 1 iff at end, else returns 0).
 */
panStream_WORDIO_INLINE int
panStream_eofP(panStream *ps)
{
  if (!ps || ((ps->rem == 0) && (!(ps->cursws) || !(ps->cursws->next))))
    return 1;
  else
    return 0;
}


/* Makes a routine that takes a pointer to a panStream and
 * reads a value of type TYPE (which is BYTES long) from
 * it and converts the value using CONVERTFN then
 * advances the position in the panStream.
 */
#define __panStream_make_read(FNNAME, BYTES, TYPE, CONVERTFN) \
    panStream_WORDIO_INLINE TYPE panStream_##FNNAME(panStream *ps) { \
        TYPE retval; \
	if (ps->rem >= BYTES) { \
             ps->rem -= BYTES; \
             retval = *((TYPE*)(ps->pos)); \
             ps->pos += BYTES; \
             return CONVERTFN(retval); \
        } else if (ps->rem != 0) { \
             panPrint_error( "panStream_" #FNNAME ": unaligned read\n"); return 0; \
        } else if (ps->cursws->next) { \
             ps->cursws = ps->cursws->next; \
             ps->rem = ps->cursws->length - BYTES; \
             ps->pos = ps->cursws->data; \
             retval = *((TYPE*)(ps->pos)); \
             ps->pos += BYTES; \
             return CONVERTFN(retval); \
        } else { \
             panPrint_error( "panStream_" #FNNAME ": read of empty stream\n"); \
             return 0; \
        }  \
     }

/* Makes a routine that takes a pointer to a panStream and
 * writes the value x of type TYPE (which is BYTES long) to
 * it after converting the value using CONVERTFN. Then
 * advances the position in the panStream.
 */
#define __panStream_make_write(FNNAME, BYTES, TYPE, CONVERTFN) \
    panStream_WORDIO_INLINE void panStream_##FNNAME(panStream *ps, TYPE x) { \
	if (ps->rem >= BYTES) { \
             ps->rem -= BYTES; \
             *((TYPE*)(ps->pos)) = CONVERTFN(x); \
             ps->pos += BYTES; \
        } else if (ps->rem != 0) { \
             panPrint_error( "panStream_" #FNNAME ": unaligned write\n");  \
        } else if (ps->cursws->next) { \
             ps->cursws = ps->cursws->next; \
             ps->rem = ps->cursws->length - BYTES; \
             ps->pos = ps->cursws->data; \
             *((TYPE*)(ps->pos)) = CONVERTFN(x); \
             ps->pos += BYTES; \
        } else { \
             panPrint_error( "panStream_" #FNNAME ": write to empty stream\n"); \
        }  \
     }


/******** Neither BODY or INLINE or PROTO specified ******/
#else
#error "No wordio macro specified"
#endif


/* Now actually generate all of the appropriate functions */

__panStream_make_read(read_uint8, 1, u_int8_t, )
__panStream_make_read(read_int8,  1, int8_t, )

__panStream_make_read(read_uint16, 2, u_int16_t, )
__panStream_make_read(read_uint32, 4, u_int32_t, )
__panStream_make_read(read_int16,  2, int16_t,   )
__panStream_make_read(read_int32,  4, int32_t,   )

__panStream_make_read(read_lsbfirst_uint16, 2, u_int16_t,FROM_LSBF16)
__panStream_make_read(read_lsbfirst_uint32, 4, u_int32_t,FROM_LSBF32)
__panStream_make_read(read_lsbfirst_int16,  2, int16_t,  FROM_LSBF16)
__panStream_make_read(read_lsbfirst_int32,  4, int32_t,  FROM_LSBF32)

__panStream_make_read(read_msbfirst_uint16, 2, u_int16_t,FROM_MSBF16)
__panStream_make_read(read_msbfirst_uint32, 4, u_int32_t,FROM_MSBF32)
__panStream_make_read(read_msbfirst_int16,  2, int16_t,  FROM_MSBF16)
__panStream_make_read(read_msbfirst_int32,  4, int32_t,  FROM_MSBF32)

#if !defined(pniStream_WORDIO_MAKEPROTO)
#define panStream_read_net_uint8 panStream_read_uint8
#define panStream_read_net_uint16 panStream_read_msbfirst_uint16
#define panStream_read_net_uint32 panStream_read_msbfirst_uint32
#define panStream_read_net_int8   panStream_read_int8
#define panStream_read_net_int16  panStream_read_msbfirst_int16
#define panStream_read_net_int32  panStream_read_msbfirst_int32
#else
__panStream_make_read(read_net_uint8,  1, u_int8_t, )
__panStream_make_read(read_net_uint16, 2, u_int16_t,FROM_MSBF16)
__panStream_make_read(read_net_uint32, 4, u_int32_t,FROM_MSBF32)
__panStream_make_read(read_net_int8,   1, int8_t,   )
__panStream_make_read(read_net_int16,  2, int16_t,  FROM_MSBF16)
__panStream_make_read(read_net_int32,  4, int32_t,  FROM_MSBF32)
#endif

__panStream_make_write(write_uint8, 1, u_int8_t, )
__panStream_make_write(write_int8,  1, int8_t, )

__panStream_make_write(write_uint16, 2, u_int16_t, )
__panStream_make_write(write_uint32, 4, u_int32_t, )
__panStream_make_write(write_int16,  2, int16_t,   )
__panStream_make_write(write_int32,  4, int32_t,   )

__panStream_make_write(write_lsbfirst_uint16, 2, u_int16_t,TO_LSBF16)
__panStream_make_write(write_lsbfirst_uint32, 4, u_int32_t,TO_LSBF32)
__panStream_make_write(write_lsbfirst_int16,  2, int16_t,  TO_LSBF16)
__panStream_make_write(write_lsbfirst_int32,  4, int32_t,  TO_LSBF32)

__panStream_make_write(write_msbfirst_uint16, 2, u_int16_t,TO_MSBF16)
__panStream_make_write(write_msbfirst_uint32, 4, u_int32_t,TO_MSBF32)
__panStream_make_write(write_msbfirst_int16,  2, int16_t,  TO_MSBF16)
__panStream_make_write(write_msbfirst_int32,  4, int32_t,  TO_MSBF32)

#if !defined(pniStream_WORDIO_MAKEPROTO)
#define panStream_write_net_uint8  panStream_write_uint8
#define panStream_write_net_uint16 panStream_write_msbfirst_uint16
#define panStream_write_net_uint32 panStream_write_msbfirst_uint32
#define panStream_write_net_int8   panStream_write_int8
#define panStream_write_net_int16  panStream_write_msbfirst_int16
#define panStream_write_net_int32  panStream_write_msbfirst_int32
#else
__panStream_make_write(write_net_uint8,  1, u_int8_t, )
__panStream_make_write(write_net_uint16, 2, u_int16_t,TO_MSBF16)
__panStream_make_write(write_net_uint32, 4, u_int32_t,TO_MSBF32)
__panStream_make_write(write_net_int8,   1, int8_t,   )
__panStream_make_write(write_net_int16,  2, int16_t,  TO_MSBF16)
__panStream_make_write(write_net_int32,  4, int32_t,  TO_MSBF32)
#endif

#undef  panStream_WORDIO_MAKEINLINE
#undef  panStream_WORDIO_MAKEPROTO
#undef  pniStream_WORDIO_MAKEPROTO
#undef  panStream_WORDIO_MAKEBODY
