# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgField.pm,v 1.1 1997/11/14 07:24:03 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgField.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObject derivative containing a description of a structure
# field.  The field has a type, a list of access specifiers, 
# an array size (for arrays or "scalar" for non-arrays, and a 
# name for the field.

require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgAccessSpec;
use psgType;
use psgStructDef;
use psgError;

package psgField;
@ISA = qw( psgObject );


# Creates a new field with a specified name (in an identifier)
sub new {
    my ($type, $fieldname) = @_;
    my $self = psgObject->new($fieldname->getLocation());
    $self->{f_name}      = $fieldname->getTermVal;
    $self->{f_arraysize} = "scalar";    
    $self->{f_type}      = "undefined";    
    $self->{f_pos}       = "UnspecifiedPos";
    $self->{f_aspecs}    = {};
    return bless $self, $type;
}

sub getName {
    my ($self) = @_;
    return $self->{f_name};
}

sub getArraySize {
    my ($self) = @_;
    return $self->{f_arraysize};
}

sub getType {
    my ($self) = @_;
    return $self->{f_type};
}

sub getPos {
    my ($self) = @_;
    return $self->{f_pos};
}

sub getSize {
    my ($self) = @_;
    my $asize = $self->getArraySize;
    if ($asize eq "scalar") {
	return $self->getType->getSize;
    } else {
	my $curpos = 0;
	foreach (0..($asize-1)) {
	    $curpos =  $self->getType->getStartPos($curpos);
	    $curpos += $self->getType->getSize;
	}
	return $curpos;
    }
}

sub getAlignment {
    my ($self) = @_;
    return $self->getType->getAlignment;
}

sub setPos {
    my ($self, $newpos) = @_;
    $self->{f_pos} = $newpos;
}

sub getAccessSpecs {
    my ($self) = @_;
    return $self->{f_aspecs};
}


# Takes a psgNumVal or number and sets the array size to it
sub setArraySize {
    my ($self, $newsize) = @_;    
    if (ref($newsize)) {
	return $newsize if ($newsize->isError);
	$self->{f_arraysize} = $newsize->getValue;
    } else {
	$self->{f_arraysize} = $newsize;
    }
    if ($self->{f_arraysize} < 1) {
	my $errmsg = "invalid array size";
	$self->printError($errmsg);
	return psgError($self->getLocation, $errmsg);
    }
    return $self;
}

# Passed either identifier terminal for a type name and a symbol table
# and returns $self modified or psgError
sub setType {
    my ($self, $newtype, $symtab) = @_;

    return $newtype if ($newtype->isError);
    return $self if ($self->isError);

    my $symres = $symtab->getSym($newtype);
    if ($symres) {
	$self->{f_type} = $symres;
	return $self;
    } else {
	my $errmsg = "couldn't resolve type <".$newtype->getTermVal.">";
	$newtype->printError($errmsg);
	return psgError->new($self->getLocation, $errmsg);
    }
}


# prints into a generator a field in a structure
sub makeStructContents {
    my ($self, $gen) = @_;    
    if ($self->getArraySize eq "scalar") {
	$gen->putLine(sprintf "%-20s %-20s /* size=%-4s offset=%-4s (%1s%1s) */",
		      $self->getType->getPrefixedName($gen),
		      $self->getName.";",
		      $self->getSize, $self->getPos,
		      $self->isReadable($gen->getSpecifier)?"r":" ",
		      $self->isWritable($gen->getSpecifier)?"w":" ",
		      ); 
    } else {
	$gen->putLine(sprintf "%-20s %-20s /* size=%-4s offset=%-4s (%1s%1s) */",
		      $self->getType->getPrefixedName($gen),
		      $self->getName."[".$self->getArraySize."];",
		      $self->getSize, $self->getPos,
		      $self->isReadable($gen->getSpecifier)?"r":" ",
		      $self->isWritable($gen->getSpecifier)?"w":" ",
		      );
    }
}


# Passed an ObjList of access specs and returns a modified self
sub setAccessSpecs {
    my ($self, $newaspecs) = @_;
    return $newaspecs if ($newaspecs->isError);
    return $self if ($self->isError);

    foreach $as ($newaspecs->getElements) {
	if ($self->{f_aspecs}->{$as->getSpecifier}) {
	    $as->printWarn("duplicate access specifier");
	}
	$self->{f_aspecs}->{$as->getSpecifier} = $as;
    }    
    return $self;
}

sub isReadable {
    my ($self, $specname) = @_;
    my $aspec = $self->{f_aspecs}->{$specname};
    if (!$aspec) { # default to rw
	return 1;
    } else {
	return $aspec->isReadable;
    }
}

sub isWritable {
    my ($self, $specname) = @_;
    my $aspec = $self->{f_aspecs}->{$specname};
    if (!$aspec) { # default to rw
	return 1;
    } else {
	return $aspec->isWritable;
    }
}

# Generates code to read the field into the structure pointed to by $dstname
# and incriments count.
sub putReader {
    my ($self, $gen, $dstname) = @_;
    my $asize = $self->getArraySize;
    if ($asize eq "scalar") {
	$self->getType->makeReadInline($gen, $dstname."->".$self->getName);
    } else {
	$gen->putLine( "{" );
	$gen->pushIndent;
	$gen->putLine( "int arrayctr;" );
	$gen->putLine( "for (arrayctr=0; arrayctr<".$asize."; arrayctr++) {");
	$gen->pushIndent;
	$self->getType->makeReadInline($gen, $dstname."->".$self->getName);
	$gen->popIndent;
	$gen->putLine( "}" );
	$gen->popIndent;
	$gen->putLine( "}" );
    }
}

# Generates code to write the field into the structure pointed to by $srcname
# and incriments count.
sub putWriter {
    my ($self, $gen, $srcname) = @_;
    my $asize = $self->getArraySize;
    if ($asize eq "scalar") {
	$self->getType->makeWriteInline($gen, $srcname."->".$self->getName);
    } else {
	$gen->putLine( "{" );
	$gen->pushIndent;
	$gen->putLine( "int arrayctr;" );
	$gen->putLine( "for (arrayctr=0; arrayctr<".$asize."; arrayctr++) {");
	$gen->pushIndent;
	$self->getType->makeWriteInline($gen, $srcname."->".$self->getName);
	$gen->popIndent;
	$gen->putLine( "}" );
	$gen->popIndent;
	$gen->putLine( "}" );
    }
}


sub getDebug {
    my ($self) = @_;
    return "(field:".$self->getLineNumber." ".$self->getName." ".(($self->getType ne "undefined")?$self->getType->getDebug:"undefined").")";
} 

