# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgGenerator.pm,v 1.3 1998/01/20 23:44:41 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgGenerator.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# An object that can take a list of datatypes and
# constants and generate serializer source code from it.
# Also contains information about the type of file
# to generate.
#
# Generators have a number of properties:
#   gen_outfilename  -- The filename of the file to generate
#   gen_type     -- The type of file to generate.  This may be one
#                   of capsule, appheader, or appcode.  These specify
#                   defaults for the other properties and control
#                   the prefixes appended to the various datatypes
#   gen_makelist -- A list of things to put into the generated
#                   file.  These may include struct, const, 
#                   readcode, writecode, readproto, and writeproto
#   gen_spec     -- A "specifier" string that corresponds to access specifiers
#                   for fields.  For example, might be "a" for apps and
#                   a field in a structure have the access specifier "a:ro"
#                   meaning that that field should be read-only for any
#                   code generated with "a" as the specifier.
#   gen_prefix   -- A prefix string to be appended to all of the types
#                   and constants generated by this generator.
#               


require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgError;

package psgGenerator;
@ISA = qw( psgObject );

sub new {
    my ($type, $location, $gen_outfilename, $gen_type, $gen_makelist, 
	$gen_spec, $gen_prefix) = @_;
    my $self = psgObject->new($location);
    $self->{gen_outfilename}  = $gen_outfilename;
    $self->{gen_type}     = $gen_type;
    $self->{gen_makelist} = $gen_makelist;
    $self->{gen_spec}     = $gen_spec;
    $self->{gen_prefix}   = $gen_prefix;
    $self->{gen_outfh}    = \*STDOUT;
    $self->{gen_iprefix}  =  ($gen_type eq "capsule")?"pni":"pan";

    return bless $self, $type;
}

# Creates a generator from a outfilename and a gen_type Identifier terminal
# and sets the rest to defaults based on the value of gen_type.
sub newWithDefaults {
    my ($type, $gen_outfilename_term, $gen_type_term) = @_;
    return $gen_outfilename_term if $gen_outfilename_term->isError;
    return $gen_type_term    if $gen_type_term->isError;
    my ($gen_outfilename, $gen_type, $gen_makelist, 
	$gen_spec, $gen_prefix);

    $gen_outfilename = $gen_outfilename_term->getTermVal;
    $gen_type    = $gen_type_term->getTermVal;
    if ($gen_type eq "capsule") {
	$gen_makelist = [ "struct", "const", "readcode", "writecode", 
			  "readproto", "writeproto" ];
	$gen_spec   = "c";
	$gen_prefix = "pco";
    } elsif ($gen_type eq "appheader") {
	$gen_makelist = [ "struct", "const", "readproto", "writeproto" ];
	$gen_spec   = "a";
	$gen_prefix = "pal";	
    } elsif ($gen_type eq "appcode") {
	$gen_makelist = [ "readcode", "writecode" ];
	$gen_spec   = "a";
	$gen_prefix = "pal";	
    } else {
	my $errmsg = "unknown generator type ".$gen_type;
	$gen_outfilename_term->printError($errmsg);
	return psgError->new($gen_outfilename_term->getLocation, $errmsg);
    }
    return $type->new($gen_outfilename_term->getLocation,
		      $gen_outfilename, $gen_type, $gen_makelist, 
		      $gen_spec, $gen_prefix);
}

# Creates a generator from an outfilename and a gen_mod_list.
# Uses defaults from the gen_type.
sub newWithModList {
    my ($type, $gen_outfilename_term, $gen_modlist) = @_;

    return $gen_outfilename_term if ($gen_outfilename_term->isError);
    if ($gen_modlist->containsError) { 
	return (psgError->new($gen_modlist->getLocation, 
			      "gen_mod_list contains error"));
    }

    my @gen_type_list = $gen_modlist->getWithTag("gentype");
    if (@gen_type_list == 0) {
	my $errmsg = "no generator type specified";
	$gen_modlist->printError($errmsg);
	return psgError->new($gen_modlist->getLocation, $errmsg);
    } elsif (@gen_type_list > 1) {
	my $errmsg = "multiple generator types specified.  using first.";
	$gen_modlist->printWarn($errmsg);	
    }
    my $self = $type->newWithDefaults($gen_outfilename_term, $gen_type_list[0]);

    # Now apply any modifications listed in the modlist
    my @gen_spec_list = $gen_modlist->getWithTag("spec");
    if (@gen_spec_list > 1) {
	my $errmsg = "multiple generator specs specified.  using first.";
	$gen_modlist->printWarn($errmsg);	
    } elsif (@gen_spec_list == 1) {
	$self->{gen_spec} = $gen_spec_list[0]->getTermVal;
    }
    
    my @gen_prefix_list = $gen_modlist->getWithTag("prefix");
    if (@gen_prefix_list > 1) {
	my $errmsg = "multiple generator prefixes specified.  using first.";
	$gen_modlist->printWarn($errmsg);	
    } elsif (@gen_prefix_list == 1) {
	$self->{gen_prefix} = $gen_prefix_list[0]->getTermVal;
    }
    
    my @gen_make_list = $gen_modlist->getWithTag("make");
    if (@gen_make_list > 0) {  # override and replace
	$self->{gen_makelist} = [];
	foreach $i (@gen_make_list) {
	    foreach $j ($i->getElements) {
		push @{$self->{gen_makelist}}, $j->getTermVal;
	    }
	}
    }
 
    return $self;
}

sub getOutFilename {
    my ($self) = @_;
    return $self->{gen_outfilename};
}

sub getGenType {
    my ($self) = @_;
    return $self->{gen_type};
}

sub isCapsule {
    my ($self) = @_;
    return ($self->{gen_type} eq "capsule");
}

sub isHeader {
    my ($self) = @_;
    return (($self->{gen_type} eq "capsule")
	    || ($self->{gen_type} eq "appheader"))
}

sub nameToDefun {
    my ($self) = @_;
    my $s = uc("INC_".$self->getOutFilename);
    $s =~ s/\./_/;
    return $s;

}

sub getMakelist {
    my ($self) = @_;
    return $self->{gen_makelist};
}

# Returns whether the argument passed should be made 
# (ie, appears in the makelist)
sub queryMake {
    my ($self, $make) = @_;
    foreach $m (@{$self->{gen_makelist}}) {
	if ($m eq $make) { return 1; }
    }
    return 0;
}

sub getSpecifier {
    my ($self) = @_;
    return $self->{gen_spec};
}

sub getPrefix {
    my ($self) = @_;
    return $self->{gen_prefix};
}

sub getBuiltInPrefix {
    my ($self) = @_;
    return $self->{gen_iprefix};
}


# Incriment the indentation level by one
sub pushIndent {
    my ($self) = @_;
    $self->{gen_indentlevel} += 1;
}

# Decriment the indentation level by one
sub popIndent {
    my ($self) = @_;
    $self->{gen_indentlevel} -= 1;
}

# Writes a line out to the file being generated
sub putLine {
    my ($self, $line) = @_;
    my $fh = $self->{gen_outfh};
    print $fh (" "x(2*$self->{gen_indentlevel})).$line."\n";
}

# Puts a seek command into the output stream
sub putSeek {
    my ($self, $offset) = @_;
    $self->putLine("count += ".
		   $self->{gen_iprefix}."Stream_seekFwd(ps, $offset);");
}

# Takes a TVList of const's and struct's and generates a file
# containing the requested features.  Returns 1 on success, 0 on failure.
sub makeOutputFile {
    my ($self, $srctvlist) = @_;
    if ($srctvlist->containsError) {
	$self->printError("an error occured in the input file.  ".
			  "not generating.");
    }	
    #print ">>> Generating ".$self->getOutFilename."\n";
    if (!(open OUTFILE, ">".$self->getOutFilename)) {
	$self->printError("couldn't open ".$self->getOutFilename.
			  " for writing");
	return 0;
    }
    $self->{gen_outfh} = \*OUTFILE;
    
    my @structs = $srctvlist->getWithTag("struct");
    my @consts = $srctvlist->getWithTag("const");    
   
    $self->putLine("/*************** THIS FILE WAS AUTOGENERATED -- DO NOT EDIT ***************");
    $self->putLine(" * Generated by the PAN Serializer Generator");
    $self->putLine(" * File:       ".$self->getOutFilename);
    $self->putLine(" * Source:     ".$self->getLocation->getFilename);
    $self->putLine(" * Created on: ".(scalar localtime));
    $self->putLine(" * Contains:   ".(join " ", @{$self->getMakelist}));
    $self->putLine(" * GenType:    ".$self->getGenType);
    $self->putLine(" * Specifier:  ".$self->getSpecifier);		  
    $self->putLine(" **************************************************************************/");
    $self->putLine;
    if ($self->isHeader) {
	$self->putLine("#ifndef ".$self->nameToDefun);
	$self->putLine("#define ".$self->nameToDefun);
	$self->putLine;
    }
    if ($self->isCapsule) {
	$self->putLine("#include <pni/pniswseg.h>");
	$self->putLine("#include <pni/pniswsstream.h>");
	$self->putLine("#include <pni/pninetwork.h>");
	$self->putLine("#include <pni/pnicodeobj.h>");
	$self->putLine("#include <pni/pnitimestamp.h>");
    } else {
	$self->putLine("#include <applib/panapplib.h>");
	if (!$self->isHeader) {
	    # include the header
	    my $header = $self->getOutFilename;
	    $header =~ s/.c$/.h/;
	    $self->putLine("#include \"".$header."\"");
	}
    }
    $self->putLine;
        
    if ($self->queryMake("const")) {
	$self->putLine("/* constant defintions */");
	foreach $c (@consts) {
	    $c->makeConstDef($self);
	}
	$self->putLine;
    }

    if ($self->queryMake("struct")) {
	$self->putLine("/* type defintions */");
	foreach $s (@structs) {
	    $self->putLine;
	    $s->makeTypedef($self);
	}
	$self->putLine;
    }

    if ($self->queryMake("readproto")) {
	$self->putLine("/* reader prototypes */");
	foreach $s (@structs) {
	    $self->putLine;
	    $s->makeReadProto($self);
	}
	$self->putLine;
    }

    if ($self->queryMake("writeproto")) {
	$self->putLine("/* writer prototypes */");
	foreach $s (@structs) {
	    $self->putLine;
	    $s->makeWriteProto($self);
	}
	$self->putLine;
    }

    if ($self->queryMake("readcode")) {
	$self->putLine("/* reader functions */");
	foreach $s (@structs) {
	    $self->putLine;
	    $s->makeReadBody($self);
	}
	$self->putLine;
    }

    if ($self->queryMake("writecode")) {
	$self->putLine("/* writer functions */");
	foreach $s (@structs) {
	    $self->putLine;
	    $s->makeWriteBody($self);
	}
	$self->putLine;
    }
    
    if ($self->isHeader) {
	$self->putLine("#endif /* ".$self->nameToDefun." */");
    }
    $self->putLine; 
    $self->putLine("/************************* END OF GENERATED CODE **************************/");

    close OUTFILE;    
    return 1;
}


sub getDebug {
    my ($self) = @_;
    return "(generator:".$self->getLineNumber." (gentype ".$self->{gen_type}.") (outfile \"".$self->{gen_outfile}."\") (spec ".$self->{gen_spec}.") (prefix \"".$self->{gen_prefix}."\") (makelist ".(join " ", @{$self->{gen_makelist}})."))";
} 

