# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgStructDef.pm,v 1.1 1997/11/14 07:24:07 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgStructDef.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgType derivative containing a description of a datatype
# as specified by a structure.
# Also capable of generating code for accessing the structure.
#


require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgGenerator;
use psgError;

package psgStructDef;
@ISA = qw( psgType );


# Creates a new structdef from a identifier terminal naming the
# structure and from a psgFieldList.
sub new {
    my ($type, $name_term, $fieldlist) = @_;

    return $name_term if ($name_term->isError);
    if ($fieldlist->containsError) { 
	return (psgError->new($fieldlist->getLocation, 
			      "fieldlist contains error"));
    }

    my ($t_align, $t_size, $t_name);
    $t_align = 4;   # all structures have an alignment of 4
    $t_name = $name_term->getTermVal;

    my $pasize = $fieldlist->getTotalSizeAndSetFieldPos; # pre-aligned size
    if (($pasize % $t_align) == 0) { $t_size = $pasize; }
    else { $t_size = $pasize + $t_align - ($pasize % $t_align); }
   
    my $self = psgType->new($name_term->getLocation, 
			    $t_name, $t_size, $t_align);
    $self->{t_fieldlist} = $fieldlist;
    return bless $self, $type;
}

# sub getSize
# sub getAlignment 
# sub getStartPos 
# sub getName 

sub getFieldList {
    my ($self) = @_;
    return $self->{t_fieldlist};
}

sub getPrefixedName {
    my ($self, $gen) = @_;
    my $pname = $gen->getPrefix.$self->{t_name};
    return $pname;
}


# Takes a generator and
# containing a typedef for the type.
sub makeTypedef {
    my ($self, $gen) = @_;            
    $gen->putLine( "typedef struct _".$self->getPrefixedName($gen)." {" );
    $gen->pushIndent;
    $self->getFieldList->makeStructContents($gen);
    $gen->popIndent;
    $gen->putLine( "} ".$self->getPrefixedName($gen).";" );
    $gen->putLine( "#define ".$self->getPrefixedName($gen)."_SIZE ".$self->getSize);
}

# Takes a generator and prints lines 
# containing a function that reads in the type.
sub makeReadBody {
    my ($self, $gen) = @_;    
    my $nextpos = 0;
    my $curpos = 0;
    my @fields = $self->{t_fieldlist}->getElements;

    $gen->putLine( "int" );
    $gen->putLine( $self->getPrefixedName($gen)."_read(".
		  $self->getPrefixedName($gen)." *dst, panStream *ps)" );
    $gen->putLine( "{" );
    $gen->pushIndent;
    $gen->putLine( "int count=0;" );

    foreach $f (@fields) {
	$nextpos = $f->getPos;
	if ($f->isReadable($gen->getSpecifier)) {
	    if ($curpos != $nextpos) {
		$gen->putSeek($nextpos-$curpos);
	    }
	    $gen->putLine( "/* at position $nextpos */" );
	    $f->putReader($gen, "dst");
	    $curpos = $nextpos + $f->getSize;
	} else {
	    $gen->putLine( "/* skipping ".$f->getName." */" );
	}
    }
    if ($curpos != $self->getSize) {
	$gen->putSeek($self->getSize-$curpos);
	$curpos = $self->getSize;
    }
    $gen->putLine( "/* final size is $curpos */" );
    $gen->putLine( "return count;" );
    $gen->popIndent;
    $gen->putLine( "}" );    
}

# Takes a generator and prints lines 
# containing a function that writes out the type.
sub makeWriteBody {
    my ($self, $gen) = @_;    
    my $nextpos = 0;
    my $curpos = 0;
    my @fields = $self->{t_fieldlist}->getElements;

    $gen->putLine( "int" );
    $gen->putLine( $self->getPrefixedName($gen)."_write(panStream *ps, ".
		  $self->getPrefixedName($gen)." *src)" );
    $gen->putLine( "{" );
    $gen->pushIndent;
    $gen->putLine( "int count=0;" );

    foreach $f (@fields) {
	$nextpos = $f->getPos;
	if ($f->isWritable($gen->getSpecifier)) {
	    if ($curpos != $nextpos) {
		$gen->putSeek($nextpos-$curpos);
	    }
	    $gen->putLine( "/* at position $nextpos */" );
	    $f->putWriter($gen, "src");
	    $curpos = $nextpos + $f->getSize;
	} else {
	    $gen->putLine( "/* skipping ".$f->getName." */" );
	}
    }
    if ($curpos != $self->getSize) {
	$gen->putSeek($self->getSize-$curpos);
	$curpos = $self->getSize;
    }
    $gen->putLine( "/* final size is $curpos */" );
    $gen->putLine( "return count;" );
    $gen->popIndent;
    $gen->putLine( "}" );    
}


# Takes a generator and the name of a pointer to the object to read into 
# and prints lines containing code to read in the type from ps.
# Incriments the variable count by the number of bytes read.
sub makeReadInline {
    my ($self, $gen, $dst) = @_;    
    $gen->putLine( "count += ".$self->getPrefixedName($gen).
	"_read(&(".$dst."), ps);" );
}

# Takes a generator and the name of a pointer to the object to write from 
# and prints lines containing code to write out the type to ps.
# Incriments the variable count by the number of bytes written.
sub makeWriteInline {
    my ($self, $gen, $src) = @_;    
    $gen->putLine( "count += ".$self->getPrefixedName($gen).
	"_write(ps, &(".$src."));" );
}



sub getDebug {
    my ($self) = @_;
    return "(structdef:".$self->getLineNumber." ".$self->getName..")";
} 

