# $Header: /home/dl2/nygren/cvs/pan/src/etc/pansergen/psgType.pm,v 1.1 1997/11/14 07:24:09 nygren Exp $
#
###########################################################################
#
# Copyright (C) 1997 by the
#   Laboratory for Computer Science, Massachusetts Institute of Technology
#   545 Technology Square
#   Cambridge, MA 02139
#
# Permission to use, copy, modify, distribute, and sell this software
# and its documentation for any purpose is hereby granted without fee,
# provided that the above copyright notice appear in all copies and
# that both that copyright notice and this permission notice appear in
# supporting documentation, and that the name of the authors and
# copyright holders not be used in advertising or publicity pertaining
# to distribution of the software without specific, written prior
# permission.  The authors and copyright holders make no
# representations about the suitability of this software for any
# purpose.  It is provided "as is" without express or implied
# warranty.
#
# THE AUTHORS AND COPYRIGHT HOLDERS DISCLAIM ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT OR
# CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
# OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
# NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
# CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#
###########################################################################
#
# FILE:    psgType.pm
# AUTHORS: Erik Nygren (nygren@mit.edu)
#
# A psgObject derivative containing a description of a datatype.
# Also capable of generating code for accessing the datatype.
#


require 5.002;

BEGIN {		
  push(@INC,  ("../parselex"));
}

use Parse::CLex;
use psgObject;
use psgGenerator;
use psgError;

package psgType;
@ISA = qw( psgObject );


# Creates a new type from a name, size, and alignment
sub new {
    my ($type, $location, $t_name, $t_size, $t_align) = @_;
    my $self = psgObject->new($location);
    $self->{t_name}   = $t_name;
    $self->{t_size}   = $t_size;
    $self->{t_align}  = $t_align;
    return bless $self, $type;
}

# Methods:
#
# getSize = # of bytes
# getAlignment = byte boundary that must start on
# getName = returns the unprefixed name
# getPrefixedName = returns the name with a prefix added
# makeTypedef = returns an array that typedefs the type
# makeReadProto
# makeWriteProto
# makeReadBody
# makeWriteBody
# makeReadInline
# makeWriteInline
#
 
sub getSize {
    my ($self) = @_;
    return $self->{t_size};
}

sub getAlignment {
    my ($self) = @_;
    return $self->{t_align};
}

# Passed the current position.  Uses alignment and size
# requirements to figure out the starting position of the 
# data structure.
sub getStartPos {
    my ($self, $curpos) = @_;
    my $align = $self->getAlignment;
    if (($curpos % $align) == 0) {
	return $curpos;
    } else {
	return $curpos + ($align - ($curpos % $align));
    }
}

sub getName {
    my ($self) = @_;    
    return $self->{t_name};
}

sub getPrefixedName {
    my ($self, $generator) = @_;    
    return $self->{t_name};
}

# Takes a generator and prints lines 
# containing a typedef for the type.
sub makeTypedef {
    my ($self, $gen) = @_;    
    $self->printError("makeTypedef not implemented for ".ref($self));
    $gen->putLine( "/* Can't make typedef for ".
	$self->getPrefixedName($gen)." */" );
}

# Takes a generator and prints lines 
# containing a prototype for a reader function for the type.
# The generated function takes a pointer to the object named dst
# and a pointer to a panStream named src and returns the number of bytes read.
sub makeReadProto {
    my ($self, $gen) = @_;    
    $gen->putLine( "extern int" );
    $gen->putLine( $self->getPrefixedName($gen)."_read(".
	$self->getPrefixedName($gen)." *dst, panStream *ps);" );
}

# Takes a generator and prints lines 
# containing a prototype for a writer function for the type.
# The generated function takes a a pointer to a panStream ps
# and apointer to the object named src
# and returns the number of bytes written.
sub makeWriteProto {
    my ($self, $gen) = @_;    
    $gen->putLine( "extern int" );
    $gen->putLine( $self->getPrefixedName($gen)."_write(panStream *ps, ".
	$self->getPrefixedName($gen)." *src);" );
}

# Takes a generator and prints lines 
# containing a function that writes out the type.
sub makeReadBody {
    my ($self, $gen) = @_;    
    $self->printError("makeReadBody not implemented for ".ref($self));
    $gen->putLine( "/* Can't make body for ".
	$self->getPrefixedName($gen)."_read */" );
}

# Takes a generator and prints lines 
# containing a function that writes out the type.
sub makeWriteBody {
    my ($self, $gen) = @_;    
    $self->printError("makeWriteBody not implemented for ".ref($self));
    $gen->putLine( "/* Can't make body for ".
	$self->getPrefixedName($gen)."_write */" );
}

# Takes a generator and the name of a pointer to the object to read into 
# and prints lines containing code to read in the type from ps.
# Incriments the variable count by the number of bytes read.
sub makeReadInline {
    my ($self, $gen, $dst) = @_;    
    $gen->putLine( "count += ".$self->getPrefixedName($gen).
	"_read(".$dst.", ps);" );
}

# Takes a generator and the name of a pointer to the object to write from 
# and prints lines containing code to write out the type to ps.
# Incriments the variable count by the number of bytes written.
sub makeWriteInline {
    my ($self, $gen, $src) = @_;    
    $gen->putLine( "count += ".$self->getPrefixedName($gen).
	"_write(ps, ".$src.");" );
}


sub getDebug {
    my ($self) = @_;
    return "(type:".$self->getLineNumber." ".$self->getName..")";
} 

