/* $Header: /home/dl2/nygren/cvs/pan/src/etc/udpforward/udpforward.c,v 1.1 1998/02/18 20:01:43 nygren Exp $
 *
 * FILE:    udpforward.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Simple user-space packet forwarder.  Listens on two ports and forwards
 * any received packets to specified hosts.  (Thus, acting as a gateway).
 *
 * Usage:  udpfoward listenport desthost destport
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <signal.h>
#include <assert.h>
#include <sys/time.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>

#undef DEBUG
#ifdef DEBUG
#define printdebug(fmt, arg...) printf(fmt, ##arg)
#else
#define printdebug(fmt, arg...)
#endif

int gotexitsig=0;

#define MAX_PACKET_SIZE 16384
static char *packetbuf[MAX_PACKET_SIZE];

struct gwroute {
  int gwrnum;
  int netsock;
  struct sockaddr_in listenaddr;
  int listenportnum;
  int destportnum;
  struct sockaddr_in destaddr;
} gwr1, gwr2;


void 
breakhandler(int signum)
{
  printdebug("breakhandler: got signal %d\n", signum);
  gotexitsig = 1;
}

int
init_gwroute(int gwrnum, struct gwroute *gwr, 
	     char *listenport, char *desthost, char *destport)
{
  struct hostent *hentry;
  
  if ((gwr->netsock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    { perror("socket"); return -1; }
  gwr->gwrnum = gwrnum;
  gwr->listenportnum = atoi(listenport);
  gwr->destportnum   = atoi(destport);

  memset((char*)&gwr->listenaddr, 0, sizeof(gwr->listenaddr));
  gwr->listenaddr.sin_family      = AF_INET;
  gwr->listenaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  gwr->listenaddr.sin_port        = htons(gwr->listenportnum);

  if (bind(gwr->netsock, (struct sockaddr *)&gwr->listenaddr, 
	   sizeof(gwr->listenaddr)) < 0)
    { perror("socket"); return -1; }
  
  printdebug("listening on %d for gwr%d\n", gwr->listenportnum, gwrnum);

  hentry = gethostbyname(desthost);
  if (!hentry)
    { perror("gethostbyname"); return -1; }  
  assert((hentry->h_length == 4) && (hentry->h_addrtype == AF_INET));

  memset((char*)&gwr->destaddr, 0, sizeof(gwr->destaddr));
  gwr->destaddr.sin_family      = AF_INET;
  memcpy((char*)&(gwr->destaddr.sin_addr.s_addr), 
	 hentry->h_addr, hentry->h_length);
  gwr->destaddr.sin_port        = htons(gwr->destportnum);

  printdebug("got destination name for gwr%d\n", gwrnum);
  
  return 0;
}

int 
packet_forward(struct gwroute *gwr)
{
    struct msghdr  msghdr;	/* Msghdr structure for recvmsg */
    struct iovec   rcvvecs[1];	/* Receive vectors */ 
    int bread;			/* bytes read */
    int retval;

    printdebug("entered packet_forward for gwr%d\n", gwr->gwrnum);
    
    memset(&msghdr, 0, sizeof(msghdr));
    msghdr.msg_iov    = rcvvecs;
    msghdr.msg_iovlen = 1;
    rcvvecs[0].iov_base = packetbuf;
    rcvvecs[0].iov_len  = MAX_PACKET_SIZE;

    bread = recvmsg(gwr->netsock, &msghdr, 0);
    if (bread < 0)
      { perror("recvmsg"); return -1; }
    
    printdebug("got %d bytes on gwr%d\n", bread, gwr->gwrnum);

    memset(&msghdr, 0, sizeof(msghdr));
    msghdr.msg_name    = &gwr->destaddr;
    msghdr.msg_namelen = sizeof(gwr->destaddr);
    msghdr.msg_iov    = rcvvecs;
    msghdr.msg_iovlen = 1;
    rcvvecs[0].iov_base = packetbuf;
    rcvvecs[0].iov_len  = bread;
    
    retval = sendmsg(gwr->netsock, &msghdr, 0);
    if (retval < 0)
      { perror("sendmsg"); return -1; }
    return 0;
}


void
main(int argc, char *argv[])
{
  fd_set _readfds, *readfds=&_readfds;
  
  if (argc != 7) {
    fprintf(stderr, "Usage:  udpfoward listenport1 desthost1 destport1\n");
    fprintf(stderr, "                  listenport2 desthost2 destport2\n");
    exit(-1);
  }

  signal(SIGHUP, breakhandler);
  signal(SIGTERM, breakhandler);

  if (init_gwroute(1, &gwr1, argv[1], argv[2], argv[3])) 
    exit(-1);
  if (init_gwroute(2, &gwr2, argv[4], argv[5], argv[6])) 
    exit(-1);

  while (!gotexitsig) 
    {
      FD_ZERO(readfds);
      FD_SET(gwr1.netsock, readfds);
      FD_SET(gwr2.netsock, readfds);
      printdebug("about to select\n");
      if (select(gwr2.netsock+1, readfds, NULL, NULL, NULL) < 0)
	{ perror("select"); exit(-1); }
      printdebug("select returned\n");
      if (FD_ISSET(gwr1.netsock, readfds))
	if (packet_forward(&gwr1)) exit(-1);
      if (FD_ISSET(gwr2.netsock, readfds))
	if (packet_forward(&gwr2)) exit(-1);
    }
  close(gwr1.netsock);
  close(gwr2.netsock);
}

