/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/app_if.c,v 1.2 1997/12/18 02:33:28 nygren Exp $
 *
 * FILE:    app_if.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Based on linux/net/unix/af_unix.c code from Linux 2.0.32,
 * written by Alan Cox, <alan@cymru.net>
 *
 *
 * Provides an application interface to a pan KNode.  Contains routines
 * for connecting a PAN node to an application and for
 * sending/receiving capsules.
 *
 * We're currently using AF_PAN socket streams where
 * a 4-byte integer capsule length is sent, followed by the capsule
 * itself (this repeats).  The first thing sent over new connections
 * is the local node address.
 *
 */


#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <common/nodeaddr.h>
#include <node/environment.h>

#include "app_if.h"
#include "net_if.h"
#include "knode_stubs.h"
#include "skbuffswseg.h"

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/stat.h>
#include <linux/socket.h>
#include <linux/un.h>
#include <linux/fcntl.h>
#include <linux/termios.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/ip.h>
#include <linux/fs.h>
#include <linux/malloc.h>
#include <asm/segment.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <net/sock.h>
#include <net/tcp.h>
#include <linux/proc_fs.h>


static panNodeAppLinkSocket *panNodeAppLink_SocketList=NULL;

#define min(a,b)	(((a)<(b))?(a):(b))

/*
 *	Note: Sockets may not be removed _during_ an interrupt or net_bh
 *	handler using this technique. They can be added although we do not
 *	use this facility.
 */
 
static void 
panNodeAppLink_remove_socket(panNodeAppLinkSocket *sk)
{
  panNodeAppLinkSocket **s;
	
  cli();
  s=&panNodeAppLink_SocketList;

  while(*s!=NULL)
    {
      if(*s==sk)
	{
	  *s=sk->next;
	  sti();
	  return;
	}
      s=&((*s)->next);
    }
  sti();
}

static void 
panNodeAppLink_insert_socket(panNodeAppLinkSocket *sk)
{
  cli();
  sk->next=panNodeAppLink_SocketList;
  panNodeAppLink_SocketList=sk;
  sti();
}



/* Finalizes a panSwseg_TYPE_APPLINK swseg 
 */
static void
panSwsegAppLink_finalize(panSwseg *sws)
{
  if (!sws) return;
  if (sws->type != panSwseg_TYPE_APPLINK) 
    { panPrint_error("panNodeAppLink_finalize called with segment of type %d\n", sws->type); return; }
  panMemory_free(sws);
}


/*
 *	Delete a unix socket. We have to allow for deferring this on a timer.
 */

static void 
panNodeAppLink_destroy_timer(unsigned long data)
{
  panNodeAppLinkSocket *sk=(panNodeAppLinkSocket *)data;
  if(sk->protinfo.af_unix.locks==0 && sk->wmem_alloc==0)
    {
      if(sk->protinfo.af_unix.name)
	kfree(sk->protinfo.af_unix.name);
      sk_free(sk);
      MOD_DEC_USE_COUNT;
      return;
    }
	
  /*
   *	Retry;
   */
	 
  sk->timer.expires=jiffies+10*HZ;	/* No real hurry try it every 10 seconds or so */
  add_timer(&sk->timer);
}
	 
	 
static void 
panNodeAppLink_delayed_delete(panNodeAppLinkSocket *sk)
{
  sk->timer.data=(unsigned long)sk;
  sk->timer.expires=jiffies+HZ;		/* Normally 1 second after will clean up. After that we try every 10 */
  sk->timer.function=panNodeAppLink_destroy_timer;
  add_timer(&sk->timer);
}
	
static void panNodeAppLink_destroy_socket(panNodeAppLinkSocket *sk)
{
  struct sk_buff *skb;

  panNodeAppLink_remove_socket(sk);
	
  while((skb=skb_dequeue(&sk->receive_queue))!=NULL)
    {
      if(sk->state==TCP_LISTEN && skb->sk)
	{
	  panNodeAppLinkSocket *osk=skb->sk;
	  osk->state=TCP_CLOSE;
	  kfree_skb(skb, FREE_WRITE);	/* Now surplus - free the skb first before the socket */
	  osk->state_change(osk);		/* So the connect wakes and cleans up (if any) */
	  /* osk will be destroyed when it gets to close or the timer fires */			
	}
      else
	{
	  /* passed fds are erased in the kfree_skb hook */
	  kfree_skb(skb,FREE_WRITE);
	}
    }

  if(--sk->protinfo.af_unix.locks==0 && sk->wmem_alloc==0)
    {
      if(sk->protinfo.af_unix.name)
	kfree(sk->protinfo.af_unix.name);
      sk_free(sk);
      MOD_DEC_USE_COUNT;
    }
  else
    {
      sk->dead=1;
      panNodeAppLink_delayed_delete(sk);	/* Try every so often until buffers are all freed */
    }
}



 
static int 
panNodeAppLink_fcntl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
  return -EINVAL;
}

/*
 *	Yes socket options work with the new unix domain socketry!!!!!!!
 */
 
static int 
panNodeAppLink_setsockopt(struct socket *sock, int level, 
			  int optname, char *optval, int optlen)
{
  panNodeAppLinkSocket *sk=sock->data;
  if(level!=SOL_SOCKET)
    return -EOPNOTSUPP;
  return sock_setsockopt(sk,level,optname,optval,optlen);	
}

static int panNodeAppLink_getsockopt(struct socket *sock, int level, int optname, char *optval, int *optlen)
{
  panNodeAppLinkSocket *sk=sock->data;
  if(level!=SOL_SOCKET)
    return -EOPNOTSUPP;
  return sock_getsockopt(sk,level,optname,optval,optlen);
}

static int 
panNodeAppLink_listen(struct socket *sock, int backlog)
{
  return -EOPNOTSUPP;		
#if 0
  panNodeAppLinkSocket *sk=sock->data;
  if(sk->type!=SOCK_STREAM)
    return -EOPNOTSUPP;		/* Only stream sockets accept */
  if(sk->protinfo.af_unix.name==NULL)
    return -EINVAL;			/* No listens on an unbound socket */
  sk->max_ack_backlog=backlog;
  sk->state=TCP_LISTEN;
  return 0;
#endif
}

static void 
def_callback1(struct sock *sk)
{
  if(!sk->dead)
    wake_up_interruptible(sk->sleep);
}

static void 
def_callback2(struct sock *sk, int len)
{
  if(!sk->dead)
    {
      wake_up_interruptible(sk->sleep);
      sock_wake_async(sk->socket, 1);
    }
}

static void 
def_callback3(struct sock *sk)
{
  if(!sk->dead)
    {
      wake_up_interruptible(sk->sleep);
      sock_wake_async(sk->socket, 2);
    }
}

static int 
panNodeAppLink_create(struct socket *sock, int protocol)
{
  panNodeAppLinkSocket *sk;
  panNodeAppLink       *al;

  MOD_INC_USE_COUNT;

  panPrint_debug("panNodeAppLink: creating a new socket\n");

  if(protocol && protocol != PF_PAN) {
    MOD_DEC_USE_COUNT;
    return -EPROTONOSUPPORT;
  }
  sk=(panNodeAppLinkSocket *)sk_alloc(GFP_KERNEL);
  al=panMemory_allocStruct(panNodeAppLink);
  if((sk==NULL) && (al==NULL)){
    MOD_DEC_USE_COUNT;
    return -ENOMEM;
  }
  switch(sock->type)
    {
    case SOCK_STREAM:
      break;
      /*
       *	Believe it or not BSD has AF_UNIX, SOCK_RAW though
       *	nothing uses it.
       */
    case SOCK_RAW:
      sock->type=SOCK_DGRAM;
    case SOCK_DGRAM:
      break;
    default:
      sk_free(sk);
      panMemory_free(al);
      MOD_DEC_USE_COUNT;
      return -ESOCKTNOSUPPORT;
    }
  sk->type=sock->type;
  init_timer(&sk->timer);
  skb_queue_head_init(&sk->write_queue);
  skb_queue_head_init(&sk->receive_queue);
  skb_queue_head_init(&sk->back_log);
  sk->protinfo.af_unix.family=AF_PAN;
  sk->protinfo.af_unix.locks=1;		/* Us */
  sk->protinfo.af_unix.readsem=MUTEX;	/* single task reading lock */
  sk->rcvbuf=SK_RMEM_MAX;
  sk->sndbuf=SK_WMEM_MAX;
  sk->allocation=GFP_ATOMIC;
  sk->state=TCP_CLOSE;
  sk->priority=SOPRI_NORMAL;
  sk->state_change=def_callback1;
  sk->data_ready=def_callback2;
  sk->write_space=def_callback3;
  sk->error_report=def_callback1;
  sk->mtu=SK_WMEM_MAX;       /* TODO: what should this be???? */
  sk->socket=sock;
  sock->data=(void *)sk;
  sk->sleep=sock->wait;

  /* Create a panNodeAppLink and stick it into the inode ptr.
   * (memory allocated above). */
  al->sk           = sk;
  al->sws.data     = NULL;
  al->sws.length   = 0;
  al->sws.type     = panSwseg_TYPE_APPLINK;
  al->sws.numref   = 1;		/* The socket ref's the swseg */
  al->sws.next     = NULL;
  al->sws.finalize = panSwsegAppLink_finalize;
  panNodeAppLinkSocket_setAppLink(sk, al);

  panNodeAppLink_insert_socket(sk);
  
  panPrint_debug("panNodeAppLink: socket successfully created!\n");
  return 0;
}

static int 
panNodeAppLink_dup(struct socket *newsock, struct socket *oldsock)
{
  return panNodeAppLink_create(newsock,0);
}

static int 
panNodeAppLink_release(struct socket *sock, struct socket *peer)
{
  panNodeAppLinkSocket *sk=sock->data;
	
  /* May not have data attached */
	
  if(sk==NULL)
    return 0;

  /* Nuke this first so it can't try and send things through
   * here anymore. */
  panLOCK;  /* TODO: Not sure what sort of locking we might need here */
  panNodeAppLinkSocket_getAppLink(sk)->sk = NULL;
  panSwseg_unref((panSwseg*)panNodeAppLinkSocket_getAppLink(sk));
  panUNLOCK;
		
  sk->state_change(sk);
  sk->dead=1;

  panNodeAppLink_destroy_socket(sk);   /* Try to flush out this socket. Throw out buffers at least */


  /*
   *	FIXME: BSD difference: In BSD all sockets connected to use get ECONNRESET and we die on the spot. In
   *	Linux we behave like files and pipes do and wait for the last dereference.
   */
	 
  sock->data = NULL;
  sk->socket = NULL;

  return 0;
}


/*
 *   Insert a capsule into the node.
 */
		
static int 
panNodeAppLink_sendmsg(struct socket *sock, struct msghdr *msg, 
		       int len, int nonblock, int flags)
{
  panNodeAppLinkSocket *sk=sock->data;
  void *saddr=msg->msg_name;
  panSwseg *sws;
  panStream capstart;


  if(sk->err)
    return sock_error(sk);

  if(flags&MSG_OOB)
    return -EOPNOTSUPP;
			
  if(flags)	/* For now */ {
    return -EINVAL;
  }
  
  /* Addresses don't mean anything */
  if(saddr!=NULL)
    {
      return -EINVAL;
    }


  /*
   *	A control message has been attached.
   */
  if(msg->msg_control) 
    {
      return -EOPNOTSUPP;
    }

  if (len > sk->mtu) 
    {
      return -EMSGSIZE;
    }

  /*** Create a swseg, copy the data into it, then evaluate the capsule ***/

  /* We can add headroom and tailroom here later to increase performance */
  sws = panSwsegSkbuff_create(len, 
			      (((ETH_HLEN+15)&~15)+0+sizeof(struct iphdr)),
			      4);
  if (!sws)
    {
      return -ENOMEM;
    }

  memcpy_fromiovec(sws->data, msg->msg_iov, len);
  
  panStream_init(&capstart, sws, 0);

  panLOCK;  /* Only one capsule can be executing at once! */
  panNode_acceptCapsule(&capstart, len, 
			(panSwseg*)panNodeAppLinkSocket_getAppLink(sk));
  panUNLOCK;

  return len;
  



#if 0 /*************************************************/
  while(sent < len)
    {
		 
      size=len-sent;

      if(size>(sk->sndbuf-sizeof(struct sk_buff))/2)	/* Keep two messages in the pipe so it schedules better */
	{
	  if(sock->type==SOCK_DGRAM)
	    {
	      panNodeAppLink_fd_free(sk,fp,fpnum);
	      return -EMSGSIZE;
	    }
	  size=(sk->sndbuf-sizeof(struct sk_buff))/2;
	}
      /*
       *	Keep to page sized kmalloc()'s as various people
       *	have suggested. Big mallocs stress the vm too
       *	much.
       */
#define MAX_ALLOC (PAGE_SIZE*7/8)
      if(size > MAX_ALLOC && sock->type!=SOCK_DGRAM)
	limit = MAX_ALLOC;	/* Fall back to 4K if we can't grab a big buffer this instant */
      else
	limit = 0;	/* Otherwise just grab and wait */

      /*
       *	Grab a buffer
       */
		 
      skb=sock_alloc_send_skb(sk,size,limit,nonblock, &err);
		
      if(skb==NULL)
	{
	  panNodeAppLink_fd_free(sk,fp,fpnum);
	  if(sent)
	    {
	      sk->err=-err;
	      return sent;
	    }
	  return err;
	}
      size=skb_tailroom(skb);		/* If we dropped back on a limit then our skb is smaller */

      skb->sk=sk;
      skb->free=1;
		
      if(fpnum)
	{
	  panNodeAppLink_attach_fds(fpnum,fp,skb);
	  fpnum=0;
	}
      else
	skb->h.filp=NULL;

      memcpy_fromiovec(skb_put(skb,size),msg->msg_iov, size);

      cli();
      if(saddr==NULL)
	{
	  other=sk->protinfo.af_unix.other;
	  if(sock->type==SOCK_DGRAM && other->dead)
	    {
	      other->protinfo.af_unix.locks--;
	      sk->protinfo.af_unix.other=NULL;
	      sock->state=SS_UNCONNECTED;
	      sti();
	      kfree_skb(skb, FREE_WRITE);
	      /*
	       *	Check with 1003.1g - what should
	       *	datagram error
	       */
	      if (!sent)
		sent = -ECONNRESET;
	      return sent;
	    }
	  /*
	   *	Stream sockets SIGPIPE
	   */
	  if(sock->type==SOCK_STREAM && other->dead)
	    {
	      kfree_skb(skb, FREE_WRITE);
	      sti();
	      if(!sent)
		{
		  send_sig(SIGPIPE,current,0);
		  sent = -EPIPE;
		}
	      return sent;
	    }
	}
      else
	{
	  panNodeAppLink_mkname(saddr, msg->msg_namelen);
	  other=panNodeAppLink_find_other(saddr->sun_path, &err);
	  if(other==NULL)
	    {
	      sti();
	      kfree_skb(skb, FREE_WRITE);
	      if(sent)
		return sent;
	      else
		return err;
	    }
	}
      skb_queue_tail(&other->receive_queue, skb);
      sti();
      /* if we sent an fd, only do it once */
      other->data_ready(other,size);
      sent+=size;
    }
  return sent;
#endif /*************************************************/

}


/* Delivers data to an application.  
 * TODO:  We need to deal with both "send and release write rights" and
 * "copy before send" semantics.  Right now we're always doing
 * "copy before send" because there's no way for the 
 * application to express the other possibility.
 * This can result in more copies than necessary, however.
 *
 * Returns the number of bytes sent.  TODO: Maybe we should also 
 * return an indication of whether the app really needs to 
 * release all write privledges.  (Ie, even in "send and release write
 * rights" there may be cases where we need to do the
 * copy anyways, so the code object should be allowed to then
 * write if it really wants to).
 *
 * TODO: We really need to do some flow control here as well!
 *
 */
int
panNodeAppLink_deliverData(panNodeAppLink *alink, panStream *datastart, 
			    int length)
{
  panNodeAppLinkSocket *sk;
  struct sk_buff *skb;
  int sent=0;

  if (!alink || !alink->sk || alink->sk->dead)
    return -1;
  sk = alink->sk;

  skb = alloc_skb(length, GFP_ATOMIC);
  if (!skb) 
    return -1;
  skb->free = 1;
  sent = panStream_read(skb_put(skb, length), datastart, length);
  if (sent<length)
    skb_trim(skb, sent);
  skb_queue_tail(&sk->receive_queue, skb);
  sk->data_ready(sk, sent);
  return sent;
}


static int 
panNodeAppLink_getname(struct socket *sock, struct sockaddr *saddr,
		       int *panaddr_len, int peer)
{
  /*panNodeAppLinkSocket *sk=sock->data;*/
  struct sockaddr_pan *panaddr=(struct sockaddr_pan*)saddr;

  if(peer)
    {
      return -EINVAL;
    }

  panaddr->span_family=AF_PAN;
  if(panNodeAddr_LocalAddr==NULL)
    {
      return -ENOTCONN;
    }
  panNodeAddr_dup(&(panaddr->span_addr), panNodeAddr_LocalAddr);
  *panaddr_len = sizeof(struct sockaddr_pan);
  return 0;
}


/*
 *	Sleep until data has arrive. But check for races..
 */
 
static void 
panNodeAppLink_data_wait(panNodeAppLinkSocket * sk)
{
  cli();
  if (!skb_peek(&sk->receive_queue)) {
    sk->socket->flags |= SO_WAITDATA;
    interruptible_sleep_on(sk->sleep);
    sk->socket->flags &= ~SO_WAITDATA;
  }
  sti();
}

static int 
panNodeAppLink_recvmsg(struct socket *sock, struct msghdr *msg, 
		       int size, int noblock, int flags, int *addr_len)
{
  panNodeAppLinkSocket *sk=sock->data;
  struct sk_buff *skb;
  int copied=0;
  unsigned char *sp;
  int len;
  int num;
  struct iovec *iov=msg->msg_iov;
  int ct=msg->msg_iovlen;

  if(flags&MSG_OOB)
    return -EOPNOTSUPP;
		
  if(addr_len)
    *addr_len=0;
		
  if(sk->err)
    return sock_error(sk);

  if((msg->msg_control) || (msg->msg_name!=NULL))
    {
      return -EOPNOTSUPP;
    }
	
  down(&sk->protinfo.af_unix.readsem);		/* Lock the socket */
  while(ct--)
    {
      int done=0;
      sp=iov->iov_base;
      len=iov->iov_len;
      iov++;
		
      while(done<len)
	{
	  if (copied && (flags & MSG_PEEK))
	    goto out;
	  if (copied == size)
	    goto out;
	  skb=skb_dequeue(&sk->receive_queue);
	  if(skb==NULL)
	    {
	      up(&sk->protinfo.af_unix.readsem);
	      if(sk->shutdown & RCV_SHUTDOWN)
		return copied;
	      if(copied)
		return copied;
	      if(noblock)
		return -EAGAIN;
	      panNodeAppLink_data_wait(sk);
	      if(current->signal & ~current->blocked)
		return -ERESTARTSYS;
	      down(&sk->protinfo.af_unix.readsem);
	      continue;
	    }

	  num=min(skb->len,len-done);
	  memcpy_tofs(sp, skb->data, num);

	  copied+=num;
	  done+=num;
	  sp+=num;
	  if (!(flags & MSG_PEEK))
	    skb_pull(skb, num);
	  /* put the skb back if we didn't use it up.. */
	  if (skb->len) {
	    skb_queue_head(&sk->receive_queue, skb);
	    continue;
	  }
	  kfree_skb(skb, FREE_WRITE);
	  if(sock->type==SOCK_DGRAM)
	    goto out;
	}
    }
out:
  up(&sk->protinfo.af_unix.readsem);
  return copied;
}



	
static int 
panNodeAppLink_select(struct socket *sock,  int sel_type, select_table *wait)
{
  return datagram_select(sock->data,sel_type,wait);
}


static int 
panNodeAppLink_ioctl(struct socket *sock, unsigned int cmd, unsigned long arg)
{
  panNodeAppLinkSocket *sk=sock->data;
  int err;
  long amount=0;
			
  switch(cmd)
    {
	
    case TIOCOUTQ:
      err=verify_area(VERIFY_WRITE,(void *)arg,sizeof(unsigned long));
      if(err)
	return err;
      amount=sk->sndbuf-sk->wmem_alloc;
      if(amount<0)
	amount=0;
      put_fs_long(amount,(unsigned long *)arg);
      return 0;
    case TIOCINQ:
      {
	struct sk_buff *skb;
	/* These two are safe on a single CPU system as only user
	 * tasks fiddle here */
	/* Return only the amount of data available in the first sk_buff */
	if((skb=skb_peek(&sk->receive_queue))!=NULL)
	  amount=skb->len;
	err=verify_area(VERIFY_WRITE,(void *)arg,sizeof(unsigned long));
	if(err)
	  return err;
	put_fs_long(amount,(unsigned long *)arg);
	return 0;
      }

    default:
      return -EINVAL;
    }
  /*NOTREACHED*/
  return(0);
}


int 
panNodeAppLink_isConnected(panNodeAppLink *applink)
{
  if (applink->sk && !applink->sk->dead) return 1;
  else                                   return 0;
}


#ifdef CONFIG_PROC_FS
static int 
panNodeAppLink_get_info(char *buffer, char **start, off_t offset, 
			int length, int dummy)
{
  off_t pos=0;
  off_t begin=0;
  int len=0;
  panNodeAppLinkSocket *s=panNodeAppLink_SocketList;
	
  len+= sprintf(buffer,"Num       RefCount Protocol Flags    Type St "
		"AppLink Path\n");
	
  while(s!=NULL)
    {
      len+=sprintf(buffer+len,"%p: %08X %08X %08lX %04X %02X %p",
		   s,
		   s->protinfo.af_unix.locks,
		   0,
		   s->socket->flags,
		   s->socket->type,
		   s->socket->state,
		   s->socket->inode);
      if(s->protinfo.af_unix.name!=NULL)
	len+=sprintf(buffer+len, " %s\n", s->protinfo.af_unix.name);
      else
	buffer[len++]='\n';
		
      pos=begin+len;
      if(pos<offset)
	{
	  len=0;
	  begin=pos;
	}
      if(pos>offset+length)
	break;
      s=s->next;
    }
  *start=buffer+(offset-begin);
  len-=(offset-begin);
  if(len>length)
    len=length;
  return len;
}

static int panNodeAppLink_ProcInode=-1;

#endif


struct proto_ops panNodeAppLink_proto_ops = {
  AF_PAN,
	
  panNodeAppLink_create,
  panNodeAppLink_dup,
  panNodeAppLink_release,
  NULL, /* panNodeAppLink_bind, */
  NULL, /* panNodeAppLink_connect, */
  NULL, /* panNodeAppLink_socketpair, */
  NULL, /* panNodeAppLink_accept, */
  panNodeAppLink_getname,
  panNodeAppLink_select,
  panNodeAppLink_ioctl,
  NULL, /* panNodeAppLink_listen, */
  NULL, /* panNodeAppLink_shutdown, */
  panNodeAppLink_setsockopt,
  panNodeAppLink_getsockopt,
  panNodeAppLink_fcntl,
  panNodeAppLink_sendmsg,
  panNodeAppLink_recvmsg
};


int 
panNodeAppLink_proto_init(struct net_proto *pro)
{
  if (sock_register(panNodeAppLink_proto_ops.family,
		    &panNodeAppLink_proto_ops) < 0) 
    {
      panPrint_error("panNodeAppLink_proto_init: "
		     "couldn't register protocol\n"); 
      return -1;
    }
#ifdef CONFIG_PROC_FS
  {
    static struct proc_dir_entry pde = {
      -1,  11, "panapplinks",
      S_IFREG | S_IRUGO, 1, 0, 0,
      0, &proc_net_inode_operations,
      panNodeAppLink_get_info
    };
    proc_register_dynamic(&proc_net, &pde); 
    panNodeAppLink_ProcInode = pde.low_ino;
  }
#endif

  return 0;
}

void 
panNodeAppLink_proto_remove(void)
{
  if (panNodeAppLink_SocketList)
    {
      panPrint_error("panNodeAppLink_proto_remove: "
		     "there are still sockets hanging around!\n"); 
    }
#ifdef CONFIG_PROC_FS
    if (panNodeAppLink_ProcInode > 0)
      proc_unregister(&proc_net, panNodeAppLink_ProcInode); 
#endif
  sock_unregister(panNodeAppLink_proto_ops.family);
}


/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/************ EVERYTHING AFTER HERE HAS BEEN REMOVED ***************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/
/*******************************************************************/








#if 0 /********************************************************/	

static int panNodeAppLink_bind(struct socket *sock, struct sockaddr *uaddr, int addr_len)
{
  struct sockaddr_un *sunaddr=(struct sockaddr_un *)uaddr;
  panNodeAppLinkSocket *sk=sock->data;
  int old_fs;
  int err;
	
  if(sk->protinfo.af_unix.name)
    return -EINVAL;		/* Already bound */
	
  if(addr_len>sizeof(struct sockaddr_un) || addr_len<3 || sunaddr->sun_family!=AF_PAN)
    return -EINVAL;
  panNodeAppLink_mkname(sunaddr, addr_len);
  /*
   *	Put ourselves in the filesystem
   */
  if(sk->protinfo.af_unix.inode!=NULL)
    return -EINVAL;
	
  sk->protinfo.af_unix.name=kmalloc(addr_len+1, GFP_KERNEL);
  if(sk->protinfo.af_unix.name==NULL)
    return -ENOMEM;
  memcpy(sk->protinfo.af_unix.name, sunaddr->sun_path, addr_len+1);
	
  old_fs=get_fs();
  set_fs(get_ds());
	
  err=do_mknod(sk->protinfo.af_unix.name,S_IFSOCK|S_IRWXUGO,0);
  if(err==0)
    err=open_namei(sk->protinfo.af_unix.name, 2, S_IFSOCK, &sk->protinfo.af_unix.inode, NULL);
	
  set_fs(old_fs);
	
  if(err<0)
    {
      kfree_s(sk->protinfo.af_unix.name,addr_len+1);
      sk->protinfo.af_unix.name=NULL;
      if(err==-EEXIST)
	return -EADDRINUSE;
      else
	return err;
    }
	
  return 0;
	
}

#endif /********************************************************/	



#if 0 /********************************************************/	

static int panNodeAppLink_connect(struct socket *sock, struct sockaddr *uaddr, int addr_len, int flags)
{
  panNodeAppLinkSocket *sk=sock->data;
  struct sockaddr_un *sunaddr=(struct sockaddr_un *)uaddr;
  panNodeAppLinkSocket *other;
  struct sk_buff *skb;
  int err;

  if(sk->type==SOCK_STREAM && sk->protinfo.af_unix.other)
    {
      if(sock->state==SS_CONNECTING && sk->state==TCP_ESTABLISHED)
	{
	  sock->state=SS_CONNECTED;
	  return 0;
	}
      if(sock->state==SS_CONNECTING && sk->state == TCP_CLOSE)
	{
	  sock->state=SS_UNCONNECTED;
	  return -ECONNREFUSED;
	}
      if(sock->state!=SS_CONNECTING)
	return -EISCONN;
      if(flags&O_NONBLOCK)
	return -EALREADY;
      /*
       *	Drop through the connect up logic to the wait.
       */
    }
	
  if(addr_len < sizeof(sunaddr->sun_family)+1 || sunaddr->sun_family!=AF_PAN)
    return -EINVAL;
		
  panNodeAppLink_mkname(sunaddr, addr_len);
		
  if(sk->type==SOCK_DGRAM)
    {
      if(sk->protinfo.af_unix.other)
	{
	  sk->protinfo.af_unix.other->protinfo.af_unix.locks--;
	  sk->protinfo.af_unix.other=NULL;
	  sock->state=SS_UNCONNECTED;
	}
      other=panNodeAppLink_find_other(sunaddr->sun_path, &err);
      if(other==NULL)
	return err;
      if(other->type!=sk->type)
	return -EPROTOTYPE;
      other->protinfo.af_unix.locks++;
      sk->protinfo.af_unix.other=other;
      sock->state=SS_CONNECTED;
      sk->state=TCP_ESTABLISHED;
      return 0;			/* Done */
    }
	

  if(sock->state==SS_UNCONNECTED)
    {
      /*
       *	Now ready to connect
       */
	 
      skb=sock_alloc_send_skb(sk, 0, 0, 0, &err); /* Marker object */
      if(skb==NULL)
	return err;
      skb->sk=sk;				/* So they know it is us */
      skb->free=1;
      skb->h.filp=NULL;
      sk->state=TCP_CLOSE;
      panNodeAppLink_mkname(sunaddr, addr_len);
      other=panNodeAppLink_find_other(sunaddr->sun_path, &err);
      if(other==NULL)
	{
	  kfree_skb(skb, FREE_WRITE);
	  return err;
	}
      if(other->type!=sk->type)
	{
	  kfree_skb(skb, FREE_WRITE);
	  return -EPROTOTYPE;
	}
      other->protinfo.af_unix.locks++;		/* Lock the other socket so it doesn't run off for a moment */
      other->ack_backlog++;
      sk->protinfo.af_unix.other=other;
      skb_queue_tail(&other->receive_queue,skb);
      sk->state=TCP_SYN_SENT;
      sock->state=SS_CONNECTING;
      sti();
      other->data_ready(other,0);		/* Wake up ! */		
    }
			
	
  /* Wait for an accept */
	
  cli();
  while(sk->state==TCP_SYN_SENT)
    {
      if(flags&O_NONBLOCK)
	{
	  sti();
	  return -EINPROGRESS;
	}
      interruptible_sleep_on(sk->sleep);
      if(current->signal & ~current->blocked)
	{
	  sti();
	  return -ERESTARTSYS;
	}
    }
	
  /*
   *	Has the other end closed on us ?
   */
	 
  if(sk->state==TCP_CLOSE)
    {
      sk->protinfo.af_unix.other->protinfo.af_unix.locks--;
      sk->protinfo.af_unix.other=NULL;
      sock->state=SS_UNCONNECTED;
      sti();
      return -ECONNREFUSED;
    }
	
  /*
   *	Amazingly it has worked
   */
	 
  sock->state=SS_CONNECTED;
  sti();
  return 0;
}

#endif  /********************************************************/	



#if 0  /********************************************************/	

static int panNodeAppLink_socketpair(struct socket *a, struct socket *b)
{
  panNodeAppLinkSocket *ska,*skb;	
	
  ska=a->data;
  skb=b->data;

  /* Join our sockets back to back */
  ska->protinfo.af_unix.locks++;
  skb->protinfo.af_unix.locks++;
  ska->protinfo.af_unix.other=skb;
  skb->protinfo.af_unix.other=ska;
  ska->state=TCP_ESTABLISHED;
  skb->state=TCP_ESTABLISHED;
  return 0;
}

#endif  /********************************************************/	


#if 0  /********************************************************/	

static int panNodeAppLink_accept(struct socket *sock, struct socket *newsock, int flags)
{
  panNodeAppLinkSocket *sk=sock->data;
  panNodeAppLinkSocket *newsk, *tsk;
  struct sk_buff *skb;
	
  if(sk->type!=SOCK_STREAM)
    {
      return -EOPNOTSUPP;
    }
  if(sk->state!=TCP_LISTEN)
    {
      return -EINVAL;
    }
		
  newsk=newsock->data;
  if(sk->protinfo.af_unix.name!=NULL)
    {
      newsk->protinfo.af_unix.name=kmalloc(strlen(sk->protinfo.af_unix.name)+1, GFP_KERNEL);
      if(newsk->protinfo.af_unix.name==NULL)
	return -ENOMEM;
      strcpy(newsk->protinfo.af_unix.name, sk->protinfo.af_unix.name);
    }
		
  do
    {
      cli();
      skb=skb_dequeue(&sk->receive_queue);
      if(skb==NULL)
	{
	  if(flags&O_NONBLOCK)
	    {
	      sti();
	      return -EAGAIN;
	    }
	  interruptible_sleep_on(sk->sleep);
	  if(current->signal & ~current->blocked)
	    {
	      sti();
	      return -ERESTARTSYS;
	    }
	  sti();
	}
    }
  while(skb==NULL);
  tsk=skb->sk;
  kfree_skb(skb, FREE_WRITE);	/* The buffer is just used as a tag */
  sk->ack_backlog--;
  newsk->protinfo.af_unix.other=tsk;
  tsk->protinfo.af_unix.other=newsk;
  tsk->state=TCP_ESTABLISHED;
  newsk->state=TCP_ESTABLISHED;
  newsk->protinfo.af_unix.locks++;	/* Swap lock over */
  sk->protinfo.af_unix.locks--;	/* Locked to child socket not master */
  tsk->protinfo.af_unix.locks++;	/* Back lock */
  sti();
  tsk->state_change(tsk);		/* Wake up any sleeping connect */
  sock_wake_async(tsk->socket, 0);
  return 0;
}

#endif  /********************************************************/



#if 0  /********************************************************/	

static int panNodeAppLink_shutdown(struct socket *sock, int mode)
{
  panNodeAppLinkSocket *sk=(panNodeAppLinkSocket *)sock->data;
  panNodeAppLinkSocket *other=sk->protinfo.af_unix.other;
  if(mode&SEND_SHUTDOWN)
    {
      sk->shutdown|=SEND_SHUTDOWN;
      sk->state_change(sk);
      if(other)
	{
	  other->shutdown|=RCV_SHUTDOWN;
	  other->state_change(other);
	}
    }
  other=sk->protinfo.af_unix.other;
  if(mode&RCV_SHUTDOWN)
    {
      sk->shutdown|=RCV_SHUTDOWN;
      sk->state_change(sk);
      if(other)
	{
	  other->shutdown|=SEND_SHUTDOWN;
	  other->state_change(other);
	}
    }
  return 0;
}

#endif /********************************************************/	
