/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/kpand.c,v 1.1 1997/12/18 02:33:30 nygren Exp $
 *
 * FILE:    kpand.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Runs as a thread within a PAN node and deals with relatively 
 * long-running tasks like loading in new code objects.
 * May also at some point be used to periodically do 
 * soft-state cache cleanup and other gc-like things.
 *
 */


#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <common/nodeaddr.h>
#include <node/environment.h>
#include <common/conventions.h>
#include <node/node.h>

#include <linux/socket.h>
#include <linux/ip.h>
#include <linux/sched.h>
#include <net/ip.h>		/* As in /usr/src/linux/include/net/ip.h */
#include <asm/atomic.h>

extern int errno;  /* This is a complete hack.  The kernel needs a better
		    * interface to this... */
#define __KERNEL_SYSCALLS__
#include <asm/unistd.h>

#include "net_if.h"
#include "kpand.h"
#include "skbuffswseg.h"


static panCodeobjLoadReq *panCodeobjLoadReq_QueueHead = NULL;
static panCodeobjLoadReq *panCodeobjLoadReq_QueueTail = NULL;

struct wait_queue *panKPanD_SchedWaitQueue = NULL;


/* Adds a request to the tail of the panCodeobj Load Request Queue */
static void
panCodeobjLoadReq_enqueue(panCodeobjLoadReq *clr) 
{
  panCritRegion_ENTER();
  if (panCodeobjLoadReq_QueueTail)
    {
      panCodeobjLoadReq_QueueTail->next = clr; 
      panCodeobjLoadReq_QueueTail       = clr; 
    }
  else
    {
      panCodeobjLoadReq_QueueHead = panCodeobjLoadReq_QueueTail = clr;
    }
  panCritRegion_LEAVE();
}


/* Returns the head of the panCodeobj Load Request Queue
 * or returns NULL. */
static panCodeobjLoadReq *
panCodeobjLoadReq_dequeue(void) 
{
  panCodeobjLoadReq *clr;
  panCritRegion_ENTER();
  clr = panCodeobjLoadReq_QueueHead;
  if (clr)
    {
      panCodeobjLoadReq_QueueHead = clr->next;
      if (panCodeobjLoadReq_QueueTail == clr) 
	panCodeobjLoadReq_QueueTail = NULL;
    }  
  panCritRegion_LEAVE();
  return clr;
}


/* Loads the code object starting at startps into the code cache.
 * Also takes the address of the node to go to to request dependencies from.
 * Should we advance startps???
 * Returns 1 on success, 0 on failure.
 *
 * Actually just queues a request with kpand to load the code object
 * and then wakes kpand up...
 *
 */
int
pniCodeobj_load_enqueue(panStream *startps, panNodeAddr *homeaddr)
{
  panCodeobjLoadReq *clr;

  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniCodeobj_load: "
		     "called with no running capsule!\n");
      return 0;
    }
  
  panPrint_debug("pniCodeobj_load: called from <%s> with homeaddr ",
		 panNode_CurrentCapsule->codeobj->symname);
  panNodeAddr_print(homeaddr);
  panPrint_debug("\n");

  clr = panMemory_allocStruct(panCodeobjLoadReq);
  if (!clr) {
    panPrint_error("pniCodeobj_load: couldn't allocate code load request\n");
    return 0;
  }

  if (startps->cursws)
    panSwseg_ref(startps->cursws);
  panStream_dup(&(clr->startps), startps);
  panNodeAddr_dup(&(clr->homeaddr), homeaddr);

  panCodeobjLoadReq_enqueue(clr);

  /* Wake up the kernel daemon.  Whee!  :) */
  if (panKPanD_SchedWaitQueue)
    {
      wake_up_interruptible(&panKPanD_SchedWaitQueue);
      /* schedule();  --  don't do this here for now unless we need to. */
    }

  return 1;
}


/* The body of the kpand daemon.  Based on start-up code 
 * for nfsiod from fs/nfs/nfsiod.c and fs/nfs/inode.c in the Linux kernel.
 *
 * TODO: There may be better ways of doing this, though.  From
 * the nfs code:
 *
 *    This is an awful hack, because when running
 *    as a module, we will keep insmod's memory.  Besides, the current->comm
 *    hack won't work in this case
 *    The best would be to have a syscall for nfs client control that (among
 *    other things) forks biod's.
 *    Alternatively, we might want to have the idle task spawn
 *    biod's on demand.
 */
static int 
panKPanD_threadBody(void *dummy)
{
  panCodeobjLoadReq *clr;
  
  MOD_INC_USE_COUNT;

#ifdef __SMP__
  lock_kernel();
  syscall_count++;
#endif

  exit_mm(current);
  current->session = 1;
  current->pgrp = 1;
  sprintf(current->comm, "kpand");

  panPrint_debug("panKPanD_threadBody: started and done initializing\n");


  /* Now that we're done with start-up code, start the main loop. */

  while (1) 
    {
      panPrint_debug("panKPanD_threadBody: going to sleep...  ZZZzzz...\n");

      interruptible_sleep_on(&panKPanD_SchedWaitQueue);

      panPrint_debug("panKPanD_threadBody: awake again!\n");
      
      if (current->signal & ~current->blocked)
	{
	  panPrint_debug("panKPanD_threadBody: exiting on signal %ld\n",
			 current->signal);
	  break;
	}

      while (clr = panCodeobjLoadReq_dequeue())
	{
	  panCodeobj_load(&clr->startps, panNode_CodeCache, &clr->homeaddr);
	  if (clr->startps.cursws)
	    panSwseg_unref(clr->startps.cursws);
	  panMemory_free(clr);
	}
      panCapsuleEnvList_evalAll(panNode_PendingCapsules);
    }


  MOD_DEC_USE_COUNT;	
  return 0;
}


int 
panKPanD_startup(void)
{
  panPrint_debug("panKPanD_startup: entered.\n");
  init_waitqueue(&panKPanD_SchedWaitQueue);
  kernel_thread(panKPanD_threadBody, NULL, 0);  
  panPrint_debug("panKPanD_startup: done.\n");
  return 0;
}
