/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/linux2/net_if.c,v 1.3 1998/02/18 20:03:05 nygren Exp $
 *
 * FILE:    net_if.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides a network interface to a pan KNode.
 * Contains routines for sending and receiving capsules across the network.
 *
 */


#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <common/nodeaddr.h>
#include <node/environment.h>
#include <common/conventions.h>
#include <node/node.h>

#include <linux/socket.h>
#include <linux/ip.h>
#include <net/ip.h>		/* As in /usr/src/linux/include/net/ip.h */
#include <asm/atomic.h>

#include "net_if.h"
#include "skbuffswseg.h"



int localaddr=0;		/* Symbol that may be set when loading the
				 * kernel module.  Must match an existing 
				 * interface.  Used to determine 
				 * panNodeAddr_LocalAddr. */
int rewritedestaddr=1;		/* Another symbol that may be set when
				 * loading the kernel module.  If non-zero,
				 * the node will rewrite destination
				 * addresses to the gateway for the 
				 * route that the packet would be sent
				 * on, if there is such a gateway. 
				 * If set to 2, does the lookup of what
				 * it would rewrite to, but doesn't
				 * change anything.  (Useful for benchmarks).
				 */
int copyonsend=0;		/* Another symbol that may be set when
				 * loading the kernel module.  If non-zero,
				 * the node will always copy capsules 
				 * when sending out IP packets. */

/* The address of this node */
static panNodeAddr _panNodeAddr_LocalAddr;
panNodeAddr *panNodeAddr_LocalAddr = &_panNodeAddr_LocalAddr;


/* Return a device that packets are coming from (the loopback device).
 * We can use the loopback address if all else fails. */
inline struct device *
panNodeNetIF_getLocalDev(void) 
{
  struct device *dev;
  
  for (dev = dev_base; dev != NULL; dev = dev->next) 
    {
      /* Find our loopback address */
      if (dev->flags & IFF_LOOPBACK)
	{
	  return dev;
	}
    }  
  panPrint_error("panNodeNetIF_getLocalDev: no loopback device!\n");
  return NULL;
}


/*
 * Set the node address to a local address.
 * The address can be specified by passing the "localaddr" symbol
 * as the IP address in hex when loading the kernel module.
 */
int
panNodeAddr_setLocalAddr(void) 
{
  struct device *dev;
  int found=0;
  
  memset(panNodeAddr_LocalAddr, 0, sizeof(panNodeAddr));
  for (dev = dev_base; dev != NULL; dev = dev->next) 
    {
      /* Find our non-loopback address */
      if (!(dev->flags & IFF_LOOPBACK))
	{
	  if (localaddr 
	      && (htonl(localaddr) != dev->pa_addr))
	    continue;
	  found = 1;
	  *((u_int32_t*)panNodeAddr_LocalAddr) = dev->pa_addr;
	}
    }
  if (!found)
    {
      panPrint_warn("panNodeAddr_setLocalAddr: couldn't find an interface\n");
      return -1;
    }
  panPrint_debug("panNodeAddr_setLocalAddr: node address set to:\n");
  panNodeAddr_print(panNodeAddr_LocalAddr);
  return 0;
}


/* Sends out a capsule across the network to remote node
 * for evaluation.  Requires that a capsule be running
 * with a valid environment, as some information is
 * obtained from the environment.
 * TODO: Add an argument to specify if packet was consumed, and
 * indicate this in the return value as well.
 */
extern int
panNodeNetIF_sendCapsule(panNodeAddr *destaddr, panStream *capstart, 
			  int length)
{
  int max_headroom;		/* Extra space needed for IP and 
				 * link-level header */
  int pan_plus_ip_hlen;		/* The length of a PAN header plus an
				 * IP header... */
  struct iphdr *iph;
  struct sk_buff *skb = NULL;
  panSwsegSkbuff *skbsws=NULL;
  u_int32_t target, savetarget;	/* IP addr of where we're going */
  int bread;			/* Amount the stream was advanced */
  struct rtable *rt=NULL;	/* Route to the destination */

  panPrint_debug("panNodeNetIF_sendCapsule: entered (length %d).\n", length);

  if (!destaddr || !capstart || !capstart->cursws) {
    return -1;
  }
  
  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING) {
    panPrint_error("panNodeNetIF_sendCapsule: "
		   "called with no running capsule!\n");
    return -1;
  }

  target = *((u_int32_t*)destaddr);
  
  /* We only want to duplicate the skbuff if we know now
   * that we need to.  Otherwise let a lower level deal, 
   * especially since it may later turn out that 
   * the packet needs to be fragmented.  Ideally, we don't
   * want to have to look up the routing entry since 
   * ip_forward also does that.
   *
   * In order to not make any copies, a few things need to hold true:
   *    * The swseg needs to be of type SkbuffSwseg
   *    * length must be shorter than the length of the skbuff
   *      contained within the swseg
   *    * The skbuff must have enough headroom for at least an IP header
   *      (later on, if it doesn't also have enough headroom for 
   *      the hardware header a copy will happen).
   *    * The caller must have relinquished control of the swseg to us,
   *      giving up all rights to send it again or write to it later.
   *      We should probably have a way of enforcing this, but how?
   *
   */
  if (capstart->cursws 
      && (capstart->cursws->type == panSwseg_TYPE_SKBUFF)
      && (capstart->rem >= length)) 
    {
      skbsws = (panSwsegSkbuff*)(capstart->cursws);
      skb = skbsws->skb;
      /* Warning: the following breaks the abstraction */
      skb->data  = capstart->pos;
      skb->len   = length;
      skb->tail  = skbsws->skb->data + length;
      if (skb->tail > skb->end) {
	panPrint_error("panNodeNetIF_sendCapsule: skb weirdness!\n");
	return -1;
      }
    }
  pan_plus_ip_hlen = (sizeof(struct iphdr)) + 0;
  if (!copyonsend && skbsws 
      /* && skb  -- not needed since set if skbsws is set */
      && (skb_headroom(skb) >= pan_plus_ip_hlen)
      && skb->free) 
    {      
      /* We don't really need to do much else here... */
      atomic_inc(&skb->count);
      bread = panStream_seekFwd(capstart, length); /* We really shouldn't need
						    * to do this here?  (esp 
						    * if the capsule will never
						    * be touched again...) */
    } 
  else		  /* Just use the panStream abstraction */
    {
      struct sk_buff *new_skb;
      
      /* We use ETH_HLEN here so that we don't need to 
       * do a lookup in the routing table.
       * We may want to do more of an over-estimate
       * at some point.  We do want the IP header
       * to be on a long-word boundary, though. 
       * If we're wrong, it will just result in an additional copy. */
      max_headroom = (((ETH_HLEN+15)&~15)+pan_plus_ip_hlen);
      
      if ( !(new_skb = alloc_skb(length+max_headroom, GFP_ATOMIC)) ) 
	{
	  panPrint_error("panNodeNetIF_sendCapsule: out of memory, "
			 "dropped packet\n");
	  return -1;
	}
      new_skb->free = 1;
      
      /*
       * Reserve space for our header and the lower device header
       */
      skb_reserve(new_skb, max_headroom);
      
      skb_put(new_skb, length);
      bread = panStream_read(new_skb->data, capstart, length);
      if (bread < length) {	/* Add filler if needed... */
	memset(new_skb->data+bread, 0, length-bread);
      }
      memset(new_skb->proto_priv, 0, sizeof(skb->proto_priv));
      
      skb = new_skb;
    }

  panPrint_debug("panNodeNetIF_sendCapsule: starts with 0x%04x 0x%04x\n", 
		 *((u_int32_t*)(skb->data)), *((u_int32_t*)(skb->data+4)));
  
  /* We would push the PAN header here if there was one... */

  /*
   *	Push down and install the IP header.
   */

  skb->h.iph = (struct iphdr *) skb_push(skb, sizeof(struct iphdr));
  skb->ip_hdr = skb->h.iph; 
  skb->mac.raw = (void*)skb->ip_hdr; /* TODO: is this the right thing
				      * for mac.raw to get set to? */

  /* Look up a route if we might need to rewrite the destination address */
  if (rewritedestaddr) 
    {
      savetarget = target;
      /* TODO: We might want to set the third argument to something
       * other than just NULL */
      if ((rt = ip_rt_route(target, 0, NULL)) == NULL)
        {
	  panPrint_warn("panNodeNetIF_sendCapsule: no route to host!\n");
	  kfree_skb(skb, FREE_WRITE);
	  return -1;
	}

      /* Rewrite the destination address! */
      if (rt->rt_flags & RTF_GATEWAY)
	{
	  panPrint_debug("panNodeNetIF_sendCapsule: rewriting destaddr\n");
	  target = rt->rt_gateway;      
	}
      ip_rt_put(rt);
      if (rewritedestaddr == 2)
	target = savetarget;
    }
	 
  skb->dev = panNodeNetIF_getLocalDev();
  iph 			=	skb->h.iph;
  iph->version		= 	4;
  iph->tos		=	0; /* TODO: have a way for capsules to set */
  iph->ttl		=	128; /* TODO: have this propagated */
  iph->frag_off		=	0;
  iph->daddr		=	target;
  iph->saddr		=	*((u_int32_t*)panNodeAddr_LocalAddr);
  iph->protocol		=	IPPROTO_PAN;
  iph->ihl		=	5;
  iph->tot_len		=	htons(length+sizeof(struct iphdr));
  iph->id		=	htons(ip_id_count++);/* Race condition here? */
  ip_send_check(iph);
  skb->protocol		=	htons(ETH_P_IP);

  /*
   *	Send the packet on its way!
   *	Note that dev_queue_xmit() will eventually free the skb.
   *	If ip_forward() made a copy, it will return 1 so we can free.
   */

  if (!sysctl_ip_forward) {
    panPrint_warn("panNodeNetIF_sendCapsule: sysctl_ip_forward is not set\n");
  }

  if (ip_forward(skb, skb->dev, IPFWD_NOTTLDEC, target))
    {      
      kfree_skb(skb, FREE_WRITE);
    }
  return bread;
}



/* Receives a defragmented packet from the IP layer.
 * We now pull the IP header off and process the capsule.
 */
int 
panNodeNetIF_rcv(struct sk_buff *skb, struct device *dev, 
		 struct options *opt, __u32 daddr, 
		 unsigned short len, __u32 saddr,
		 int redo, struct inet_protocol *protocol)
{
  panStream      capstart;
  panSwseg      *capsws;

  /* Don't unlink in the middle of a turnaround */
  MOD_INC_USE_COUNT;
  panPrint_debug("panNodeNetIF_rcv: got a packet from %d.%d.%d.%d to "
		 "%d.%d.%d.%d (length=%d)\n", NIPQUAD(saddr), NIPQUAD(daddr),
		 len-(((struct iphdr *)skb->data)->ihl<<2));
  
  /*
   *	Discard the original IP header
   */
  
  skb->mac.raw = skb->data;
  skb_pull(skb, ((struct iphdr *)skb->data)->ihl<<2);

  /* TODO: at some point we should pull another header here
   * for handling different types and versions of active networks
   * and for containing information like the ttl and length 
   * of the capsule. */

  panPrint_debug("panNodeNetIF_rcv: starts with 0x%04x 0x%04x\n", 
		 *(u_int32_t*)(skb->data), *(u_int32_t*)(skb->data+4));

  capsws = panSwsegSkbuff_create_fromSkbuff(skb);
  if (!capsws) 
    {
      kfree_skb(skb, FREE_READ);
      MOD_DEC_USE_COUNT;
      return -ENOMEM;
    }
  
  /* TODO: do we need to incriment the count on the skb here? */
  
  panStream_init(&capstart, capsws, 0);
  
  /* TODO: At some point we need to also pass information about
   * the source of the capsule to acceptCapsule (ie, the ttl, saddr, etc).
   */

  panNode_acceptCapsule(&capstart, skb->len, NULL);

  MOD_DEC_USE_COUNT;
  return(0);
}


static struct inet_protocol panNodeNetIF_protocol = {
  panNodeNetIF_rcv,     /* receive handler       */
#if 0
  NULL,                 /* Will be UDP fraglist handler */
#endif
  NULL,                 /* error control handler */
  0,                    /* next                  */
  IPPROTO_PAN,          /* protocol ID           */
  0,                    /* copy -- TODO: make this 0 or 1? */
  NULL,                 /* data                  */
  "PAN"                 /* name                  */
};


extern int
panNodeNetIF_init(void)
{
  /* Set the local node address */
  if (panNodeAddr_setLocalAddr() != 0)
    return -1;
  panPrint_msg("rewritedestaddr = %d %s\n", rewritedestaddr,
	       rewritedestaddr==2?"(doing lookup but no rewrite)":"");
  panPrint_msg("copyonsend      = %d\n", copyonsend);

  inet_add_protocol(&panNodeNetIF_protocol);

  return 0;
}


extern void
panNodeNetIF_cleanup(void)
{
  /* Set the local node address */
  panNodeAddr_setLocalAddr();

  if ( inet_del_protocol(&panNodeNetIF_protocol) < 0 )
    panPrint_error("panNodeNetIF_cleanup: can't remove protocol!\n");
  

}

/* Stores the node's local nodeaddr into na */
void
panNodeAddr_getLocalAddr(panNodeAddr *na)
{
  if (na&&panNodeAddr_getLocalAddr)
    panNodeAddr_dup(na, panNodeAddr_LocalAddr);
}
