/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/app_if.c,v 1.8 1998/02/18 20:04:00 nygren Exp $
 *
 * FILE:    app_if.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides an application interface to a pan UNode.  Contains routines
 * for connecting a PAN node to an application and for
 * sending/receiving capsules.
 *
 * We're currently using UNIX domain socket streams where
 * a 4-byte integer capsule length is sent, followed by the capsule
 * itself (this repeats).  The first thing sent over new connections
 * is the local node address.
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <sys/uio.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <common/nodeaddr.h>
#include <node/environment.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/un.h>

#include "net_if.h"
#include "app_if.h"

/* For filling in writes */
#define filler_data_size 1024
static u_int8_t filler_data[1024];


/* For constructing an iovec for sending data to
 * a node.  We use a static structure to avoid
 * having to allocate one each time, especially since this
 * is much larger than we need in order to be safe. */
#define DATA_SEND_IOVEC_SIZE 128
static struct iovec data_send_iovec[DATA_SEND_IOVEC_SIZE];
static int data_send_iovec_pos; /* The next free slot in data_send_iovec */


/* Writes length bytes from a swseg stream into an iovec by making
 * slots in the iovec point to regions of memory.  Incriments
 * iovpos, which contains the next free slot location.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns 0 on error.
 */
panStream_GEN_SCAN_FN(
      static panStream_write_toIOVec(struct iovec *iov, int *iovpos,
				     int iovsize,
				     panStream *ps, unsigned length),
      { if (!iov || !iovpos || !ps) return 0; },
      {
	if (*iovpos>=iovsize) 
	  { panPrint_error("panStream_write_toIOVec: overflow of iovec\n");
	  return 0; }
	iov[*iovpos].iov_base = ps->pos;
	iov[*iovpos].iov_len  = numscan;
	*iovpos += 1;
      })

#if 0
/* Writes length bytes from a swseg stream into a socket using write.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns 0 on error.
 */
panStream_GEN_SCAN_FN(
      static panStream_write_toSock(int sock, panStream *ps, unsigned length),
      { if ((sock<0) || !ps) return 0; },
      {	if (numscan != write(sock, ps->pos, numscan))
	{ perror("write"); return 0; }
      })
#endif


/* Finalizes a panSwseg_TYPE_APPLINK swseg by 
 */
static void
panSwsegAppLink_finalize(panSwseg *sws)
{
  panNodeAppLink *al;
  if (sws)
    {
      if (sws->type != panSwseg_TYPE_APPLINK)
	{ panPrint_error("panNodeAppLink_finalize called with segment of type %d\n", sws->type); return; }
      al = (panNodeAppLink*)sws;
      if (al->status == panNodeAppLink_STATUS_DISCONNECTED)
	{
	  panNodeAppLink_destroy(al);
	}
    }
}


/* Creates a panNodeAppIF that listens on a socket 
 *        /tmp/.PAN-unix/node.$portnum
 * and deletes it first if there's already one there.
 * Portnum is the portnumber this UNode is listening at for UDP packets
 * (in network byte order).
 */
panNodeAppIF *
panNodeAppIF_create(u_int16_t portnum)
{
  char sockfname[UNIX_PATH_MAX]; /* The name of the socket to listen on */
  int  noblock;
  panNodeAppIF *appif;

  /* Make sure that the /tmp/.PAN-unix directory exists 
   * and that the socket doesn't exist (unlinking it if it's there). 
   * We should do some tests to make sure that a node isn't already
   * running on this port, though... */
  mkdir(panNodeAppIF_SOCKDIR_NAME, panNodeAppIF_SOCKDIR_MODE);

  snprintf(sockfname, UNIX_PATH_MAX-1,
	   "%s/node.%d", panNodeAppIF_SOCKDIR_NAME,
	   ntohs(portnum));
  unlink(sockfname);

  panPrint_debug("panNodeAppIF_create: creating app interface listening at <%s>\n", sockfname);

  appif = panMemory_alloc(sizeof(panNodeAppIF));
  if (!appif)
    { panPrint_error("panNodeAppIF_create: couldn't alloc memory\n");
    return NULL; }

  appif->listensockaddr.sun_family = AF_UNIX;
  strcpy(appif->listensockaddr.sun_path, sockfname);

  if ((appif->listensock = socket(AF_UNIX, SOCK_STREAM, 0)) < 0)
    { perror("socket");
      panPrint_error("panNodeAppIF_create: couldn't create socket\n");
      return NULL; }
  
  if (bind(appif->listensock, (struct sockaddr *)&appif->listensockaddr, sizeof(struct sockaddr_un)) < 0)
    { perror("bind");
      panPrint_error("panNodeAppIF_create: couldn't bind to socket %s\n",
		     sockfname);
      return NULL; }
    
  if (listen(appif->listensock, 5) < 0)
    { perror("listen");
      panPrint_error("panNodeAppIF_create: couldn't listen to socket\n");
      return NULL; }

  /* Make the socket non-blocking */
  noblock = 1;
  ioctl(appif->listensock, FIONBIO, &noblock);

  appif->applinks = NULL;

  return appif;
}


/* Accepts a new app link from a app interface and returns it.
 * Returns NULL if the accept would block or if an error occurs. 
 */
panNodeAppLink *
panNodeAppIF_accept(panNodeAppIF *appif)
{
  panNodeAppLink *alink;	
  int              sockaddrlen, bwrote;

  alink = panMemory_calloc(1,sizeof(panNodeAppLink));
  if (!alink)
    { panPrint_error("panNodeAppLink_accept: couldn't alloc memory\n");
    return NULL; }

  sockaddrlen = sizeof(alink->appsockaddr);
  alink->sock = accept(appif->listensock, (struct sockaddr *)&alink->appsockaddr, &sockaddrlen);
  if (!alink->sock)
    {
      panMemory_free(alink);
      if (errno == EWOULDBLOCK)
	{ panPrint_warn("panNodeAppLink_accept: accept would block\n"); }
      else 
	{ perror("accept");
	panPrint_error("panNodeAppLink_accept: accept failed\n"); }
      return NULL;
    }
  
  panPrint_debug("panNodeAppLink_accept: accepted connection from <%s>\n",
		alink->appsockaddr.sun_path); 

  bwrote = write(alink->sock, panNode_NetIF->localaddr, sizeof(panNodeAddr));
  if (bwrote != sizeof(panNodeAddr))
    { perror("write");
    panPrint_error("panNodeAppLink_accept: couldn't write nodeaddr\n"); 
    close(alink->sock); panMemory_free(alink);
    return NULL; }

  alink->sws.numref   = 0;
  alink->sws.type     = panSwseg_TYPE_APPLINK;
  alink->sws.next     = NULL;
  alink->sws.finalize = panSwsegAppLink_finalize;

  alink->capsize    = -1;
  alink->status     = panNodeAppLink_STATUS_CONNECTED;
  alink->appif      = appif;

  /* Add this to the chain of applinks */
  alink->prev = NULL;
  if (appif->applinks)
    {
      appif->applinks->prev = alink;
      alink->next = appif->applinks;
      appif->applinks = alink;
    }
  else
    {
      alink->next = NULL;
      appif->applinks = alink;
    }

  return alink;
}


/* Receives a capsule over an applink and evaluates it. 
 * If no full capsule is available, returns immediately. */
void
panNodeAppLink_recvAndEval(panNodeAppLink *alink)
{
  int navail;			/* Bytes available to be read */
  int nread;			/* Number of bytes read */
  panSwseg *capdata;
  panStream      capstart;

  if (!alink || (alink->status == panNodeAppLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeAppLink_recvAndEval: passed invalid applink\n");
    return; }

  if (ioctl (alink->sock, FIONREAD, (void *) &navail) == -1) 
    { perror ("ioctl");
    panPrint_error("panNodeAppLink_recvAndEval: failed to FIONREAD\n");
    return; }

  if (alink->capsize == -1)
    {
      if (read(alink->sock, &alink->capsize, 4) != 4) 
	{ 
	  if (navail==0) 
	    {
	      panPrint_debug("panNodeAppLink_recvAndEval: "
			     "application disconnected\n");
	    }
	  else
	    { 
	      perror ("read");
	      panPrint_error("panNodeAppLink_recvAndEval: "
			     "error reading capsize\n");
	    }
	panNodeAppLink_disconnect(alink);
	return; }
      navail -= 4;
    }

  if ((alink->capsize < 0) || (navail < alink->capsize))
    { /* panPrint_warn("panNodeAppLink_recvAndEval: got short capsule (navail=%d, alink->capsize=%d)\n", navail, alink->capsize); */
      return; } 

  capdata = panSwseg_create_byMalloc(alink->capsize);
  if (!capdata)
    { panPrint_error("panNodeAppLink_recvAndEval: couldn't create swseg\n");
    return; }

  nread = read(alink->sock, capdata->data, alink->capsize);

  if ((nread == -1) || (nread != alink->capsize))
    { perror ("read");
    panPrint_error("panNodeAppLink_recvAndEval: error reading capdata."
		   "  Expected %d, got %d.\n", alink->capsize, nread);
    panSwseg_unref(capdata);
    panNodeAppLink_disconnect(alink);
    return; }

  panPrint_debug("panNodeAppLink_recvAndEval: accepting capsule of size %d\n",
		 nread);

  alink->capsize = -1;
  panStream_init(&capstart, capdata, 0);
  panNode_acceptCapsule(&capstart, nread, (panSwseg*)alink);

}


/* Delivers length bytes of data starting at datastart to the 
 * application at the other end of alink.  Advances datastart
 * and returns the number of bytes delivered or -1 if an error occurs.
 */
int
panNodeAppLink_deliverData(panNodeAppLink *alink, panStream *datastart, 
			    int length)
{
  int nsent, tosend = length, tosendthistime, nstreamadv, totalsending=0;
  u_int32_t datlen=length;
  
  
  if (length == 0) return 0;

  if (!alink || (alink->status == panNodeAppLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeAppLink_deliverData: passed invalid applink\n");
    return -1; }

  nsent = write(alink->sock, &datlen, 4);
  if (nsent != 4)
    { perror ("write");
    panPrint_error("panNodeAppLink_deliverData: error sending datlen \n");
    panNodeAppLink_disconnect(alink);
    return -1; }
  
  data_send_iovec_pos = 0;
  
  nsent = panStream_write_toIOVec(data_send_iovec, &data_send_iovec_pos,
				  DATA_SEND_IOVEC_SIZE, datastart, length);
  totalsending += nsent;
  nstreamadv = nsent;
  tosend = length-nsent;

  /* If the stream was shorter than length, fill the rest with garbage
   * to avoid desyncronizing. */
  while (tosend>0)
    {
      tosendthistime = tosend>filler_data_size?filler_data_size:tosend;

      if (data_send_iovec_pos >= DATA_SEND_IOVEC_SIZE)
	{ panPrint_error("panNodeAppLink_deliverData: iovec overflow\n");
	return -1; }
      data_send_iovec[data_send_iovec_pos].iov_base = filler_data;
      data_send_iovec[data_send_iovec_pos].iov_len  = tosendthistime;
      totalsending += tosendthistime;
      data_send_iovec_pos++;
    }

  panPrint_debug("panNodeAppLink_deliverData: (datlen=%d) sending %d bytes in %d vecs\n",
		 datlen, totalsending, data_send_iovec_pos);
  nsent = writev(alink->sock, data_send_iovec, data_send_iovec_pos);
  if (nsent != totalsending)
    { perror ("write");
    panPrint_error("panNodeAppLink_deliverData: error sending data body\n");
    panNodeAppLink_disconnect(alink);
    alink->status = panNodeAppLink_STATUS_DISCONNECTED;
    return -1; }
 
  return nstreamadv;
}


/* Disconnects an applink (and frees it if there are no references
 * left to it). */
void
panNodeAppLink_disconnect(panNodeAppLink *alink)
{
  if (!alink || (alink->status == panNodeAppLink_STATUS_DISCONNECTED))
    { panPrint_error("panNodeAppLink_disconnect: passed invalid applink\n");
    return; }

  close(alink->sock);
  alink->status = panNodeAppLink_STATUS_DISCONNECTED;
  
  if (alink->sws.numref <= 0)
    {
      panSwsegAppLink_finalize((panSwseg*)alink);
    }
}


/* Destroys an applink (and frees assocated memory).  It must
 * have been disconnected first. */
void
panNodeAppLink_destroy(panNodeAppLink *alink)
{
  if (!alink)
    { panPrint_error("panNodeAppLink_destroy: passed invalid applink\n");
    return; }
  if ((alink->status == panNodeAppLink_STATUS_CONNECTED)
      || (alink->sws.numref > 0))
    { panPrint_error("panNodeAppLink_destroy: passed a connected or"
		     " referenced (%d) applink.\n", alink->sws.numref);
    return; }
  panPrint_debug("panNodeAppLink_destroy: destroying link connected to <%s>\n",
		 alink->appsockaddr.sun_path);
  if (alink->sws.next) panSwseg_unref(alink->sws.next);
  if (alink->appif->applinks == alink)
    alink->appif->applinks = alink->next;
  if (alink->next) alink->next->prev = alink->prev;
  if (alink->prev) alink->prev->next = alink->next;
  
  panMemory_free(alink);
}


/* Returns 1 if a pniAppLink is still connected.
 * Returns 0 if it is no longer connected. */
int 
panNodeAppLink_isConnected(panNodeAppLink *applink)
{
  return applink && (applink->status == panNodeAppLink_STATUS_CONNECTED);
}

