/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/net_if.c,v 1.8 1998/02/18 20:04:01 nygren Exp $
 *
 * FILE:    net_if.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Provides a network interface to a pan UNode.  Contains routines
 * for creating network sockets and sending/receiving capsules.
 *
 */


#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <common/nodeaddr.h>
#include <node/environment.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>


#include "routing.h"
#include "net_if.h"


panUNodeNetIF *panNode_NetIF = NULL;


#define panUNodeNetIF_REFILL_BUF_SLOT(netif, slotnum, procname, onfail) \
{ \
  (netif)->rcvbufs[(slotnum)] = panSwseg_create_byMalloc(panUNodeNetIF_INBUF_SIZE); \
  if (NULL == (netif)->rcvbufs[(slotnum)]) \
    { panPrint_error(procname ": couldn't alloc swseg\n"); \
      onfail; } \
  (netif)->rcvvecs[(slotnum)].iov_base = (netif)->rcvbufs[(slotnum)]->data; \
  (netif)->rcvvecs[(slotnum)].iov_len  = panUNodeNetIF_INBUF_SIZE; \
}

/* Creates a new panUNodeNetIF (and associated sockets) 
 * with the specified local address and routing table. */
panUNodeNetIF *
panUNodeNetIF_create(panRouteTable *rtab, 
		     panNodeAddr   *localaddr)
{
  panUNodeNetIF        *netif;
  struct sockaddr_in   bindaddr;
  int                  i, noblock;
  u_int16_t            *pLocalport;

  netif = panMemory_alloc(sizeof(panUNodeNetIF));
  if (!netif) 
    { panPrint_error("panUNodeNetIF_create: couldn't alloc memory\n");
    return NULL; }

  netif->rtab      = rtab;
  netif->localaddr = localaddr;

  if ((netif->netsock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
    { perror("socket");
      panPrint_error("panUNodeNetIF_create: couldn't create socket\n");
      panMemory_free(netif);
      return NULL; }

  memset((char*)&bindaddr, 0, sizeof(bindaddr));
  bindaddr.sin_family      = AF_INET;
  bindaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  pLocalport               = (u_int16_t*)&(localaddr->data[4]);
  bindaddr.sin_port        = *pLocalport;

  panPrint_debug("panUNodeNetIF_create: using port %hu\n",
		 ntohs(*pLocalport));
  panPrint_debug("panUNodeNetIF_create: local node addr: ");
  panNodeAddr_print(panNodeAddr_LocalAddr);
  panPrint_debug("\n");
  

  if (bind(netif->netsock, (struct sockaddr *)&bindaddr, sizeof(bindaddr)) < 0)
    { perror("bind");
      panPrint_error("panUNodeNetIF_create: couldn't bind to socket\n");
      close(netif->netsock); panMemory_free(netif);
      return NULL; }

#if 0
  /* Make the socket non-blocking */
  noblock = 1;
  ioctl(netif->netsock, FIONBIO, &noblock);
#endif

  for (i=0; i<panUNodeNetIF_INBUF_COUNT; i++)
    {
      panUNodeNetIF_REFILL_BUF_SLOT(netif, i, "panUNodeNetIF_create", 
				  return NULL);
    }

  return netif;
}



/* Receives a capsule from the network and evaluates it.
 * Refills any buffers in the network interface that may have 
 * been used during reception of the capsule. */
void
panUNodeNetIF_recvAndEval(panUNodeNetIF *netif)
{
  int            i;
  int            bytesrcvd, caplength;	/* Number of bytes received */
  struct msghdr  msghdr;	/* Msghdr structure for recvmsg */
  panSwseg      *firstsws, *lastsws, *cursws; 
  panStream      capstart;

  memset(&msghdr, 0, sizeof(msghdr));
  msghdr.msg_iov    = netif->rcvvecs;
  msghdr.msg_iovlen = panUNodeNetIF_INBUF_COUNT;

  caplength = bytesrcvd = recvmsg(netif->netsock, &msghdr, 0);

  if (bytesrcvd < 0)
    { perror("recvmsg");
      panPrint_error("panUNodeNetIF_recvAndEval: error receiving capsule\n");
      return; }
  else if (bytesrcvd < panNodeAddr_SIZE+panCodeobjName_SIZE)
    { panPrint_error("panUNodeNetIF_recvAndEval: got short capsule\n");
      return; }

#ifdef MSG_TRUNC
  if (msghdr.msg_flags & MSG_TRUNC)
    { panPrint_warn("panUNodeNetIF_recvAndEval: long capsule truncated."
		    "  Dropping.\n");
    return; }
#endif

  panPrint_debug("panUNodeNetIF_recvAndEval: rcvd capsule of length %d\n",
		 bytesrcvd);

  firstsws = netif->rcvbufs[0];

  panUNodeNetIF_REFILL_BUF_SLOT(netif, 0, "panUNodeNetIF_recvAndEval", return);
  bytesrcvd -= panUNodeNetIF_INBUF_SIZE;

  /* Build up the chain of software segments while also refilling
   * buffer slots. */
  for (i=1, lastsws=firstsws; bytesrcvd > 0; 
       i++, bytesrcvd -= panUNodeNetIF_INBUF_SIZE)
    {
      cursws = netif->rcvbufs[i];
      panSwseg_chain(lastsws, cursws);
      lastsws = cursws;
      panUNodeNetIF_REFILL_BUF_SLOT(netif, i, 
				    "panUNodeNetIF_recvAndEval", return);
    }

  panStream_init(&capstart, firstsws, 0);
  
  panNode_acceptCapsule(&capstart, caplength, NULL);

}



/* Sends a capsule of length, which starts at capstart to the address
 * specified at destaddr.   Returns number of bytes sent on success,
 * otherwise, returns -1 on failure.  */
int
panUNodeNetIF_sendCapsule(panUNodeNetIF *netif,
			  panNodeAddr   *destaddr, panStream *capstart, 
			  int length)
{
  panStream       _ps, *ps=&_ps;
  panRouteEntry  *rentry;
  struct msghdr  msghdr;	/* Msghdr structure for recvmsg */
  struct iovec   sendvec[panUNodeNetIF_MAX_SEND_CAPSULE_SWSEGS];
  int            segcnt=0;	/* Number of swsegs */
  int            bytecount=0;	/* Number of bytes being sent */
  int            numscan, retval;

                
  if (!netif || !destaddr || !capstart)
    {
      panPrint_warn("panUNodeNetIF_sendCapsule: invalid args\n");
      return -1;
    }

  if (length > panUNodeNetIF_MAX_CAPSULE_SIZE)
    {
      panPrint_warn("panUNodeNetIF_sendCapsule: "
		    "attempt to send capsule larger than max capsule size "
		    "(%d bytes, max=%d bytes)!\n", length, 
		    panUNodeNetIF_MAX_CAPSULE_SIZE);
      return -1;
    }


  rentry = panRouteTable_lookup(netif->rtab, destaddr);
  if (!rentry) 
    {
      panPrint_warn("panUNodeNetIF_sendCapsule: no route to ");
      panNodeAddr_print(destaddr);   panPrint_warn("\n");
      return -1;
    }
  
  /* Build up sendvec from the capsule data starting at capstart */
  panStream_dup(ps, capstart);
  while (length > 0)
    {
      if (ps->rem == 0)
	{
	  if (!ps->cursws->next)
	    break;
	  else
	    {
	      ps->cursws = ps->cursws->next;
	      ps->rem    = ps->cursws->length;
	      ps->pos    = ps->cursws->data;
	    }
	}
      numscan = (length <= ps->rem)?(length):(ps->rem);
      sendvec[segcnt].iov_base = ps->pos;
      sendvec[segcnt].iov_len  = numscan;
      segcnt++;
      bytecount += numscan;
      length    -= numscan;
      ps->rem   -= numscan;
      ps->pos   += numscan;
    }

  /* Build up the message header, then send it off */
  memset(&msghdr, 0, sizeof(msghdr));
  msghdr.msg_name    = &(rentry->nexthop);
  msghdr.msg_namelen = sizeof(rentry->nexthop);
  msghdr.msg_iov     = sendvec;
  msghdr.msg_iovlen  = segcnt;
  
  retval = sendmsg(netif->netsock, &msghdr, 0);

  if ((retval < 0) || (retval != bytecount))
    {
      perror("sendmsg");
      panPrint_warn("panUNodeNetIF_sendCapsule: tried to send %d bytes and "
		    "sent %d instead\n", bytecount, retval);
    }

  return retval;
}
	

/* Returns the port number of a panUNode (used to distinguish the
 * node from other nodes running on the same machine). 
 * The address returned is in network byte order.  */
u_int16_t
panUNodeNetIF_getPortNum(panUNodeNetIF *netif)
{
  u_int16_t retval;
  
  if (!netif && !netif->localaddr)
    {
      panPrint_error("panUNodeNetIF_getPortNum: passed bad netif\n");
      return 0;
    }
  /* Weird.  This works, but otherwise the type casting causes
   * part to get truncated.  I must need sleep or something.  */
  retval = *(u_int16_t*)(&(netif->localaddr->data[4]));  
  return retval;
}

