/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/unode.c,v 1.6 1997/12/13 05:14:33 nygren Exp $
 *
 * FILE:    unode.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Mainline for a user-space PAN node.
 * The PAN node starts up, taking a scheme script as its configuration
 * file.  The scheme script handles the initialization of the
 * node's routing table and loads initial code objects into the
 * code cache.
 *
 * The user-space PAN node support is meant primarily for
 * experimentation with new protocols on small networks (or even on
 * single machines).  Each node has an address consisting of a 16-bit
 * port number and a 32-bit IP address.  Routing tables must be
 * specified for each node in its config file.
 *
 * We use SIOD as the scheme interpreter (the library is about 200k).
 * By using a scheme interpreter rather than a home-grown config file,
 * we don't need to write a parser/scanner, we get rich functionality,
 * and we allow really silly things to be done.  For example, 
 * we can implement an algorithm to generate the global routing within
 * the config file and then load the appropriate routing table
 * into the node.
 *
 * The node gets started with:
 *    panUNode [-v] [-l <nodelabel>] -f <configfile>
 *
 * The config file <configfile> is then loaded with pan-node-label bound
 * to the string specified by -l.  The -v flag enables verbosity.
 *
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/time.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include "routing.h"
#include "net_if.h"
#include "app_if.h"
#include "unodecfg.h"

#include <siod/siod.h>

extern void panUNode_sendTestCapsule(void);

void panUNode_receiveLoop(void);

static panNodeAppIF *panNode_AppIF; /* Node's interface to applications */

/* Prints an error message then a usage message and then exits. */
void 
print_usage_and_exit(char *progname, char *msg)
{
  if (msg)
    panPrint_warn("%s: %s\n\n", progname, msg);
  panPrint_warn("Usage:  %s [-v] [-l <nodelabel>] -f <configfile.scm>\n%s\n\n",
		progname,
"     -v                Enables verbose output\n"
"     -l <nodelabel>    Sets pan-node-label to the string <nodelabel> when\n"
"                       executing the config file\n"
"     -f <configfile>   Evaluates the configuration file in Scheme before\n"
"                       starting the node up\n");	     
  exit(-1);
}

void
main(int argc, char *argv[])
{
  int actr;
  LISP l_1;

  int   opt_verbose = 0;
  char *opt_configfile = NULL;
  char *opt_nodelabel = NULL;
  int   opt_misc1 = 0;
  

  /* First, initialize the scheme interpreter. 
   * Next, run SIOD on each of the config files specified
   * on the command line prefixed by "-f". */
  
  panPrint_msg("Starting up user-space PAN node, Version %s\n", "0.0");
  panPrint_msg("\t(C) Copyright 1997 MIT Laboratory for Computer Science\n");
  panPrint_msg("Contains SIOD Version %s\n"
	       "\t(C) Copyright 1988-1994 Paradigm Associates Inc.)\n\n",
	       siod_version());
  print_hs_1();
  init_storage();
  init_subrs();
  l_1 = flocons((double) 1.0); gc_protect(&l_1);    
  
  /* Add in the routines we export to siod */
  init_subr_0("pan-install-buildin-codeobjs", 
	      panUNodeCfg_installBuiltInCodeobjs);
  init_subr_1("pan-set!-node-addr", panUNodeCfg_setNodeAddr);
  init_subr_1("pan-set!-routing-table", panUNodeCfg_setRouteTable);
  init_subr_1("pan-load-codeobj", panUNodeCfg_loadCodeobj);

  /*   siod_verbose(l_1); */

  for (actr=1; actr<argc; actr++)
    {
      if (argv[actr][0] == '-' && argv[actr][1])
	{
	  switch (argv[actr][1])
	    {
	    case 'v':
	      opt_verbose = 1;
	      break;
	    case 'm':
	      opt_misc1 = 1;
	      break;
	    case 'f':
	      if (argv[actr][2])
		opt_configfile = &(argv[actr][2]);
	      else if (++actr<argc)
		opt_configfile = argv[actr];
	      else
		print_usage_and_exit(argv[0], "no config file name given");
	      break;
	    case 'l':
	      if (argv[actr][2])
		opt_nodelabel = &(argv[actr][2]);
	      else if (++actr<argc)
		opt_nodelabel = argv[actr];
	      else
		print_usage_and_exit(argv[0], "no node label given");
	      break;	      
	    default:
	      print_usage_and_exit(argv[0], "unknown option");
	    }
	}
      else
	{
	  print_usage_and_exit(argv[0], "invalid argument");
	}
    }

  /* Initialize the node's data structures */
  if (!panNode_init())
    { panPrint_error("main: node initialization failed!\n");
    exit(-1); }

  if (opt_nodelabel)
    setvar(cintern("pan-node-label"), strcons(-1, opt_nodelabel), NIL);
  else
    setvar(cintern("pan-node-label"), strcons(-1, ""), NIL);

  /* Evaluate the config file which specifies things
   * like the routing table. */
  if (!opt_configfile)
    print_usage_and_exit(argv[0], "no config file specified");
  panPrint_msg(">>> Loading config file: %s\n", opt_configfile);
  (void)vload(opt_configfile, 0, 1);

  panNode_NetIF = panUNodeNetIF_create(&panRouteTable_RTab,
				       panNodeAddr_LocalAddr);
  if (!panNode_NetIF)
    { panPrint_error("main: net_if initialization failed!\n");
    exit(-1); }

  panNode_AppIF = panNodeAppIF_create(panUNodeNetIF_getPortNum(panNode_NetIF));
  if (!panNode_AppIF)
    { panPrint_error("main: app_if initialization failed!\n");
    exit(-1); }

  if (opt_misc1)
    {
      int i;
      for (i=0; i<1; i++)
	{
	  panUNode_sendTestCapsule();
	  sleep(1);
	}
    }

  /* Now loop and receive capsules and process them */
  panUNode_receiveLoop();

}

#define FINDMAX(max, n) (max) = ((n)>(max))?(n):(max)

/* Waits for capsules to arrive and processes them.
 * (The main loop of the node).
 */
void
panUNode_receiveLoop(void)
{
  fd_set readfds, writefds, exceptfds;
  int retval, maxfd=0;
  panNodeAppLink *nal, *nextnal;

  /* build up the fd_set */
  FD_ZERO(&writefds); 
  FD_ZERO(&exceptfds); 

  while (1)
    {
      /* build up the fd_set -- should do this only when need to!  *TO*DO* */
      FD_ZERO(&readfds); 
      FD_SET(panNode_NetIF->netsock, &readfds);
      FINDMAX(maxfd, panNode_NetIF->netsock);
      FD_SET(panNode_AppIF->listensock, &readfds);
      FINDMAX(maxfd, panNode_AppIF->listensock);
      for (nal=panNode_AppIF->applinks; nal; nal=nal->next)
	if (nal->status == panNodeAppLink_STATUS_CONNECTED)
	  {
	    FD_SET(nal->sock, &readfds);  
	    FINDMAX(maxfd, nal->sock);
	  }

      /* select on reads */
      retval = select(maxfd+1, &readfds, 
		      &writefds, &exceptfds, NULL);
      if (retval < 0)
	{
	  perror("select");
	  panPrint_warn("panUNode_receiveLoop: select returned an error\n");
	}     
      else if (retval > 0)
	{
	  /* Receive and evaluate and pending capsules from nodes and apps
	   * and connection requests from apps */
	  if (FD_ISSET(panNode_NetIF->netsock, &readfds))
	    {
	      panUNodeNetIF_recvAndEval(panNode_NetIF);
	    }
	  if (FD_ISSET(panNode_AppIF->listensock, &readfds))
	    {
	      panPrint_debug("panUNode_receiveLoop: about to accept new appif\n");
	      panNodeAppIF_accept(panNode_AppIF);
	    }

	  for (nal=panNode_AppIF->applinks; nal; nal=nextnal)
	    { 
	      nextnal = nal->next;  /* The applink may get destroyed here,
				     * so don't rely on it still being around.
				     */
	      if ((nal->status == panNodeAppLink_STATUS_CONNECTED) 
		  && FD_ISSET(nal->sock, &readfds))
		{
		  panNodeAppLink_recvAndEval(nal);
		}
	    }
	}
      
      panPrint_debug("panUNode_receiveLoop: loop complete\n");
      /* process any other waiting capsules */
      if (!panCapsuleEnvList_EMPTY_P(panNode_PendingCapsules))
	{
	  panPrint_debug("panUNode_receiveLoop: processing pending\n");
	  panCapsuleEnvList_evalAll(panNode_PendingCapsules);
	  panPrint_debug("panUNode_receiveLoop: done processing pending\n");
	}
    }
}
