/* $Header: /home/dl2/nygren/cvs/pan/src/node/arch/unixuser/unodecfg.c,v 1.4 1997/12/18 02:33:35 nygren Exp $
 *
 * FILE:    unode.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Routines that are accessable from a panUNode's config file.
 * 
 * We use SIOD as the scheme interpreter (the library is about 200k).
 * By using a scheme interpreter rather than a home-grown config file,
 * we don't need to write a parser/scanner, we get rich functionality,
 * and we allow really silly things to be done.  For example, 
 * we can implement an algorithm to generate the global routing within
 * the config file and then load the appropriate routing table
 * into the node.
 *
 */

/*
 * Config file routines:
 *
 * (pan-set!-node-addr     '("x.x.x.x" port))
 * (pan-set!-routing-table '((("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))))
 * pan-node-label => "string specified as nodelabel on commandline"
 * (pan-load-codeobj "filename.po")
 * (pan-install-buildin-codeobjs)
 * 
 */

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include "routing.h"
#include "unodecfg.h"


#include <siod/siod.h>


extern panCodeobj *panCodeobj_TestCO; /* from sendtest.c */


/* Built-in code objects */
extern const unsigned char _panCodeobjBinary_dynamicload_po[];
extern const unsigned      _panCodeobjBinary_dynamicload_po_size;
extern const unsigned char _panCodeobjBinary_insertco_po[];
extern const unsigned      _panCodeobjBinary_insertco_po_size;


/* Installs all of the built-in code objects into the code cache. */
LISP
panUNodeCfg_installBuiltInCodeobjs(void)
{
  panPrint_debug("panUNodeCfg_installBuiltInCodeobjs: "
		 "loading panCodeobjBinary_dynamicload_po\n");
  panCodeobj_DynamicLoadCodeobj = 
    panCodeobj_loadFromMemory((char*)_panCodeobjBinary_dynamicload_po,
			      _panCodeobjBinary_dynamicload_po_size,
			      panNode_CodeCache, panNodeAddr_LocalAddr);
  if (panCodeobj_DynamicLoadCodeobj) {
    panPrint_debug("panUNodeCfg_installBuiltInCodeobjs: "
		   "dynamic loader code object installed\n");
  }
  panPrint_debug("panUNodeCfg_installBuiltInCodeobjs: "
		 "loading panCodeobjBinary_insertco_po\n");
  panCodeobj_loadFromMemory((char*)_panCodeobjBinary_insertco_po,
			    _panCodeobjBinary_insertco_po_size,
			    panNode_CodeCache, panNodeAddr_LocalAddr);
  
  return NIL;
}


/* Loads a code object into the code cache from
 * a specified file. */
LISP
panUNodeCfg_loadCodeobj(LISP l_fname)
{
  int       infd;		/* Input file descriptor */
  struct stat infstatbuf;	/* Stat info about input file */
  unsigned totalsize;
  panSwseg  *cosws;		/* Software segment for code object */
  panCodeobj *co;
  panStream   _ps, *ps=&_ps;
  char *infilename;

  if (NULLP(l_fname) || !TYPEP(l_fname, tc_string))
    { panPrint_warn("pan-load-codeobj: passed invalid argument\n");
    return NIL; }

  infilename = get_c_string(l_fname);

  panPrint_debug("\n\n********* panUNodeCfg_loadCodeobj: %s ***********\n", 
	       infilename);

  /* Get the size of the input file */
  if (0 != stat(infilename, &infstatbuf))
    { perror("stat");
      panPrint_warn("pan-load-codeobj: couldn't stat input file %s\n", 
		     infilename);  return NIL;  }
  totalsize = infstatbuf.st_size;
  if (ADDR_NOT_ALIGNED_P(totalsize))
    { panPrint_warn("pan-load-codeobj: invalid code object (size of is not 4-byte word aligned).\n"); return NIL; }

  /* Now open and create a swseg by mmapping in the file */
  infd = open(infilename, O_RDONLY);
  if (-1 == infd)
    { perror("open"); panPrint_warn("pan-load-codeobj: couldn't open file for reading\n"); return NIL; }

  cosws = panSwseg_create_byMmap(totalsize,
				  PROT_READ, MAP_PRIVATE, infd, 0);
  close(infd);			/* doesn't need to stay open to 
				 * keep the mapping */
  if (!cosws) 
    { panPrint_error("panUNodeCfg_loadCodeobj: couldn't mmap %s\n", 
		     infilename); 
    return NIL; }
  
  panStream_init(ps, cosws, 0);
  co = panCodeobj_load(ps, panNode_CodeCache, panNodeAddr_LocalAddr);
  if (!co)
    panPrint_warn("pan-load-codeobj: load of codeobj %s failed!\n", 
		  infilename);

#if 0 /* A total crock from early testing */
  if (0 == strcmp(co->symname, "dynamicload.o"))
      {
	panPrint_debug("FOUND dynamicload.o\n");
	panCodeobj_DynamicLoadCodeobj = co;
      }
  else
    panCodeobj_TestCO = co;
#endif

  panPrint_debug("Done with %s\n", infilename);

  panSwseg_unref(cosws);

  return NIL;
}



/* Generates an IP address and port number from a ("hostname" port) list.
 * Both numbers returned are in network byte order.
 * Returns 1 on success, 0 on failure. */
static int
panUNodeCfg_getNodeAddr(LISP una, u_int32_t *addr, u_int16_t *port)
{
  char           *hostname;
  struct hostent *hentry;

  if (!(NNULLP(una) 
	&& NNULLP(cdr(una)) 
	&& NNULLP(car(una))
	&& TYPEP(cdr(una), tc_cons) 
	&& NNULLP(cadr(una))
	&& TYPEP(car(una), tc_string) 
	&& TYPEP(cadr(una), tc_flonum)))
    { panPrint_warn("panUNodeCfg_getNodeAddr: passed invalid (\"addr\" port) list\n"); return 0;
    }

  hostname = get_c_string(car(una));
  *port    = htons(get_c_long(cadr(una)));
  hentry = gethostbyname(hostname);

  if (!hentry)
    { panPrint_warn("panUNodeCfg_getNodeAddr: couldn't resolve <%s>\n", hostname); return 0;
    }

  assert((hentry->h_length == 4) && (hentry->h_addrtype == AF_INET));
  memcpy((char*)addr, hentry->h_addr, hentry->h_length);

  panPrint_debug(">>> Got hostname <%s>=%08lx and port <%d>\n",
		 hostname, ntohl(*addr), ntohs(*port));
  return 1;
}


/* Sets the local node address from a ("hostname" port) list.
 * Called by the scheme function "pan-set!-node-addr".
 */
LISP 
panUNodeCfg_setNodeAddr(LISP nodeaddr)
{
  u_int32_t addr;
  u_int16_t port;

  if (!panNodeAddr_LocalAddr)
    {
      panNodeAddr_LocalAddr = panMemory_alloc(sizeof(panNodeAddr));
      if (!panNodeAddr_LocalAddr)
	{ panPrint_error("panUNodeCfg_setNodeAddr: couldn't alloc memory!\n");
	return NIL; }      
    }

  if (!panUNodeCfg_getNodeAddr(nodeaddr, &addr, &port))
    { panPrint_warn("pan-set!-node-addr: passed invalid argument\n");
    return NIL; }
  
  panNodeAddr_init(panNodeAddr_LocalAddr, addr, port);
  panPrint_msg(">> set node addr to: ");
  panNodeAddr_print(panNodeAddr_LocalAddr);
  panPrint_msg("\n");

  return nodeaddr;
}



/* Sets the routing table for the local node.
 *
 * Called with:
 * (pan-set!-routing-table '((("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))
 *                           (("dest" port) ("nexthop-addr" port))))
 *
 */
LISP 
panUNodeCfg_setRouteTable(LISP rtlist)
{
  int rctr;			/* for iterating through routing table */
  int recount;			/* Number of entries in routing table */
  LISP lctr;			/* For iterating through lists */
  panRouteEntry *re;
  u_int32_t dest_addr, nexthop_addr;
  u_int16_t dest_port, nexthop_port;


  /* Count the number of entries in the routing table we've been passed.  */
  if (NULLP(rtlist))
    { panPrint_warn("pan-set!-routing-table: passed invalid argument\n");
    return NIL; }
  recount = nlength(rtlist);

  if (panRouteTable_RTab.rtab)
    panMemory_free(panRouteTable_RTab.rtab);
  panRouteTable_RTab.rtab = panMemory_alloc(sizeof(panRouteEntry)*recount);
  if (!panRouteTable_RTab.rtab)
    { panPrint_error("panUNodeCfg_setRouteTable: couldn't alloc mem!\n");
    return NIL; }
  panRouteTable_RTab.numentries = recount;

  for (rctr=0, lctr=rtlist; rctr<recount; rctr++, lctr=cdr(lctr))
    {
      if (NNULLP(car(lctr)) && (2 == nlength(car(lctr))))
	{
	  if (!panUNodeCfg_getNodeAddr(caar(lctr), &dest_addr, &dest_port))
	    {
	      panPrint_error("pan-set!-routing-table: entry %d had invalid"
			     " destination\n", rctr); 
	      panMemory_free(panRouteTable_RTab.rtab);
	      panRouteTable_RTab.numentries = 0;
	      panRouteTable_RTab.rtab = NULL;
	      return NIL; 
	    }
	  if (!panUNodeCfg_getNodeAddr(cadar(lctr), &nexthop_addr, &nexthop_port))
	    { 
	      panPrint_error("pan-set!-routing-table: entry %d had invalid"
			     " next hop address\n", rctr); 
	      panMemory_free(panRouteTable_RTab.rtab); 
	      panRouteTable_RTab.numentries = 0;
	      panRouteTable_RTab.rtab = NULL;
	      return NIL; 
	    }
	  re = &(panRouteTable_RTab.rtab[rctr]);
	  panNodeAddr_init(&(re->dest), dest_addr, dest_port);
#ifdef HAVE_SIN_LEN
	  re->nexthop.sin_len         = sizeof(struct sockaddr_in);
#endif
	  re->nexthop.sin_family      = AF_INET;
	  re->nexthop.sin_port        = nexthop_port;
	  re->nexthop.sin_addr.s_addr = nexthop_addr;
	}
      else
	{ 
	  panPrint_error("pan-set!-routing-table: entry %d is invalid\n",rctr);
	  panPrint_error("\treason: nullp %d length %ld\n", NULLP(car(lctr)),
			 nlength(car(lctr)));
	  panMemory_free(panRouteTable_RTab.rtab); 
	  panRouteTable_RTab.numentries = 0;
	  panRouteTable_RTab.rtab = NULL;
	  return NIL;
	}
    }
  
  panPrint_debug(">>> LOADED ROUTING TABLE:\n");
  panRouteTable_print(&panRouteTable_RTab);

  return rtlist;
}

