/* $Header: /home/dl2/nygren/cvs/pan/src/node/environment.c,v 1.8 1998/01/23 07:42:06 nygren Exp $
 *
 * FILE:    environment.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for capsule execution environments.
 * (An environment contains information about the 
 * capsule currently being evaluated).
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>


int panCapsuleRT_Lock=0;	/* A lock for making sure that only
				 * one capsule is running at once.
				 * When set, we should assume that
				 * capsules is running or is about
				 * to run.  Whenever we unset it,
				 * we should run the pending queue
				 * right afterwards. 
				 */


/* Creates a new capsule execution environment and returns it (or NULL
 * if an error occurs.  */
panCapsuleEnv *
panCapsuleEnv_create(void)
{
  panCapsuleEnv *cenv;

  cenv = panMemory_calloc(1, sizeof(panCapsuleEnv));
  if (!cenv) 
    { panPrint_error("panCapsuleEnv_create: couldn't alloc memory\n"); 
    return NULL; }
  cenv->sring = panSwsegRing_create(32);
  if (!cenv->sring) 
    { panPrint_error("panCapsuleEnv_create: couldn't create sring\n"); 
    panMemory_free(cenv);
    return NULL; }  
  cenv->status = panCapsuleEnv_STATUS_DONE;
  return cenv;
}

/* Destroys a capsule execution enviroment, freeing memory
 * and unreferencing swsegments */
void
panCapsuleEnv_destroy(panCapsuleEnv *cenv)
{
  if (cenv)
    {
      if (cenv->sring) panSwsegRing_destroy(cenv->sring);
      panMemory_free(cenv);
    }
}


/* Aborts the execution of a capsule (by calling a longjmp) and logs
 * the appropriate error message (if non-null) */
void
panCapsuleEnv_abort(panCapsuleEnv *cenv, char *msg)
{
  assert(cenv && (cenv->status == panCapsuleEnv_STATUS_RUNNING));
  
  if (msg)
    panPrint_warn("ABORT of <%s>: %s\n", cenv->codeobj->symname, msg);
    
  panJump_longjmp(cenv->exitjump, -1);
}



/* Clears a capsule execution enviroment, preparing it for reuse
 * and unreferencing swsegments.  Sets status to panCapsuleEnv_STATUS_DONE */
void
panCapsuleEnv_clearAll(panCapsuleEnv *cenv)
{
  if (!cenv) return;
  if (cenv->sring) 
    panSwsegRing_clearAll(cenv->sring);
  cenv->waitnext = NULL;
  cenv->codeobj = NULL;
  {
    panCritRegion_ENTER();
    cenv->status = panCapsuleEnv_STATUS_DONE;
    panCritRegion_LEAVE();
  }
}


/* Creates an execution environment for a capsule, then
 * evaluates the contents of the capsule to completion.
 * If a capsule is already being executed, places the capsule
 * on the list of pending capsules to be evaluated ther next
 * time the node is idle.
 *
 * Passed a panStream pointing to the start of the capsule's
 * data area.  Assumes that the chain of software segments starting
 * with capdata's swseg haven't been referenced yet (other than
 * at creation) and places them in the environment's swseg ring.
 * May also be passed a reference-only Swseg as a handle to the
 * applink that inserted the capsule.
 *
 */
void
panNode_acceptCapsule(panStream *capdata, int caplength, panSwseg *applink)
{
  panCapsuleEnv  *cenv;		/* The capsule env we're evaluating within. */
  panCodeobjName  coname;
  panSwseg       *sws;
  panStream       ps;
  unsigned        upos;

  cenv = panCapsuleEnv_create();
  if (!cenv) 
    {
      panPrint_error("panNode_evalCapsule: couldn't create capsule env\n");
      return;
    }

  cenv->method = panCapsuleEnv_METHOD_ACCEPT;
  cenv->caplength = caplength;
  panStream_dup(&cenv->capstart, capdata);
  panStream_dup(&ps, capdata);
  
  /* Add the capsule contents to the environment's ring */
  if (capdata)
    {
      for (sws=panStream_getState(capdata, &upos); sws; sws = sws->next)
	panSwsegRing_add(cenv->sring, sws);
    }
  
  /* Add the applink to the capsule's env ring (and add a reference to it) */
  if (applink)
    {
      cenv->applink = applink;
      panSwseg_ref(applink);
      panSwsegRing_add(cenv->sring, applink);      
    }

  /* Read in the code object name */
  if (panCodeobjName_SIZE != panCodeobjName_read(&coname, &ps, 1))
    {
      panPrint_error("panNode_evalCapsule: error reading codeobj name\n");
      panCapsuleEnv_destroy(cenv);
      return;
    }

  /* Read in the code object home address */
  if (panNodeAddr_SIZE != panNodeAddr_read(&cenv->homeaddr, &ps))
    {
      panPrint_error("panNode_evalCapsule: error reading codeobj homeaddr\n");
      panCapsuleEnv_destroy(cenv);
      return;
    }

  {
    panCritRegion_ENTER();	/* This is a BIG critical region,
				 * but better to be a little on the safe
				 * side as we're playing with lots of
				 * potential races here. */
    cenv->codeobj = panCodeCache_lookup(panNode_CodeCache, &coname);
    
  /* See if the codeobj is available.  If so, evaluate.
   * Otherwise, add a new code object, send a request, and
   * save this environment for when the code is available.  */
  if (cenv->codeobj)
    {
      /* If another code object is already running, put this
       * code object into the pending queue, rather than starting it up. 
       * Otherwise, set the current capsule to this one.  */
      if (panNode_CurrentCapsule || panCapsuleRT_Lock)
	{
	  panPrint_debug("panNode_evalCapsule: a <%s> capsule is already"
			 " running.  Placing into pending queue.\n",
			 panNode_CurrentCapsule?
			 (panNode_CurrentCapsule->codeobj?
			  panNode_CurrentCapsule->codeobj->symname:
			  "EEEK!  Env had no codeobj!!!"):
			 "__CAPSULE_RT_LOCK__");
	  panCapsuleEnvList_insert(panNode_PendingCapsules, cenv);
	  panCritRegion_LEAVE();
	  return;
	}
      else
	{	  
	  panCapsuleRT_Lock = 1; /* Claim the right to execute a capsule */
	  panCritRegion_LEAVE();
	  panCapsuleEnv_evaluate(cenv);
	}
    }
  else
    {
      /* Couldn't find code object in code cache.  Request the code
       * object and save this environment to be evaluated when
       * a code object is available. */

      cenv->codeobj = panCodeobj_create(&coname, &cenv->homeaddr);
      if (!cenv->codeobj)
	{ panPrint_warn("panNode_evalCapsule: couldn't create codeobj\n");
	panCritRegion_LEAVE();
	return; }
      panCodeCache_insert(panNode_CodeCache, cenv->codeobj);      

      cenv->status = panCapsuleEnv_STATUS_WAITING;

      /* Make this capsule wait on the availability of the code object */
      if (cenv->codeobj->waitingcapsules
	  || (cenv->codeobj->waitingcapsules = panCapsuleEnvList_create()))
	{
	  panCapsuleEnvList_insert(cenv->codeobj->waitingcapsules, cenv);
	}
      panCritRegion_LEAVE();

      panCodeobj_demandLoad(cenv->codeobj);      
    }
  } /* End of the critical region block */

  /* Evaluate any other capsules that may have arrived recently... 
   * We may want to just wake up the kernel thread and let it do this.
   * Doing it here minimizes latency, however. */
  panCapsuleEnvList_evalAll(panNode_PendingCapsules);

}


/* Runs a method (init or finalize) of a code object which happens
 * immediately.
 * There should *never* be a running capsule when this happens
 * in a multithreaded kernel node.  If so, complain, loudly.
 * This is because init and finalize methods should only be called
 * from the kpand thread which should never preempt a running capsule, right?
 * If this is a bad assumption, we need to do something pretty
 * hairy like having the caller wait if we can't execute
 * this method right away.  Lets see how it goes.
 */
void
panCodeobj_evalMethod(panCodeobj *co, int method)
{
  panCapsuleEnv  *cenv;		/* The capsule env we're evaluating within. */

  cenv = panCapsuleEnv_create();
  if (!cenv) 
    {
      panPrint_error("panNode_evalCapsule: couldn't create capsule env\n");
      return;
    }
 
  cenv->caplength = 0;
  cenv->applink   = NULL;
  cenv->method    = method;
  cenv->codeobj   = co;

#ifdef __KERNEL__
  {
    panCritRegion_ENTER();
    if (panCapsuleRT_Lock || panNode_CurrentCapsule)
      {
	panPrint_error("panCodeobj_evalMethod: EEEEEEK!  A capsule is already"
		       " running!  Bad assumption at line %d of %s.\n",
		       __LINE__, __FILE__);
	panCritRegion_LEAVE();
	return;
      }
    panCapsuleRT_Lock = 1;
    panCritRegion_LEAVE();    
  }
#endif
  panCapsuleEnv_evaluate(cenv);

}


/* Evaluates a capsule environment by executing the appropriate
 * method from the specified code object. 
 * The code object must be available.
 * If another capsule is executing, it is suspended until
 * this one finishes.  (This is to allow things like init methods
 * to executing immediately without going onto the pending queue.)
 * In the kernel node, panCapsuleRT_Lock MUST be set and 
 * panNode_CurrentCapsule MUST be null.
 */
void
panCapsuleEnv_evaluate(panCapsuleEnv *cenv)
{
  panStream       ps;
  panCapsuleEnv  *ce_suspended;	/* A previously running capsule.
				 * Restore when we exit. */
  int returnval;

  if (!panCodeobj_STATUS_IS_READY_P(cenv->codeobj->status))
    {
      panPrint_error("panCapsuleEnv_evaluate: "
		     "code object <%s> isn't ready to run (has status %d)!\n",
		     cenv->codeobj->symname, cenv->codeobj->status);
      /* TODO: maybe we should queue the capsule up here instead? */
      panCapsuleEnv_destroy(cenv);
      panCapsuleRT_Lock = 0;
      return;
    }
 
  {
    panCritRegion_ENTER();
#ifdef __KERNEL__		
    /* Within the kernel, capsule execution should *NEVER* nest
     * because the kernel is multithreaded making this just way too hairy.
     */
    if (panNode_CurrentCapsule)
      {
	panPrint_error("panCodeobj_evaluate: EEEEEEK!  A capsule is already"
		       " running!  Bad assumption at line %d of %s.\n",
		       __LINE__, __FILE__);
	panCritRegion_LEAVE();
	return;
      }
#endif    
    ce_suspended = panNode_CurrentCapsule;
    panNode_CurrentCapsule = cenv;
    cenv->status = panCapsuleEnv_STATUS_RUNNING;
    cenv->codeobj->numaccept++;  
    panCritRegion_LEAVE();
  }
  
  /* Set a jump buffer for use when aborting execution
   * then invoke the accept method. */
  if (0 == (returnval = panJump_setjmp(cenv->exitjump)))
    {
      switch (cenv->method)
	{

	case panCapsuleEnv_METHOD_ACCEPT:
	  if (!cenv->codeobj->fn_accept)
	    panCapsuleEnv_abort(cenv, "capsule has no accept method");	  
	  panStream_dup(&ps, &cenv->capstart);
	  cenv->codeobj->fn_accept(&ps, cenv->caplength);
	  break;	 

	case panCapsuleEnv_METHOD_INIT:
	  if (!cenv->codeobj->fn_init)
	    panCapsuleEnv_abort(cenv, "capsule has no init method");
	  cenv->codeobj->fn_init();
	  break;
	  
	case panCapsuleEnv_METHOD_FINALIZE:
	  if (!cenv->codeobj->fn_finalize)
	    panCapsuleEnv_abort(cenv, "capsule has no finalize method");
	  cenv->codeobj->fn_finalize();
	  break;
	
	default:
	  panCapsuleEnv_abort(cenv, "unknown method type");
	}
    }
  else
    {
      panPrint_warn("panCapsuleEnv_evaluate: "
		    "evaluation of codeobj <%s> method type %d aborted "
		    "with status %d\n", cenv->codeobj->symname, cenv->method,
		    returnval);
    }

  {
    panCritRegion_ENTER();
#ifdef __KERNEL__
    panNode_CurrentCapsule = ce_suspended;
#else
    panNode_CurrentCapsule = NULL;
#endif
    panCapsuleRT_Lock = 0;
    panCritRegion_LEAVE();
  }

  panCapsuleEnv_destroy(cenv);

}




/********** panCapsuleEnvList routines.... **********/


/* Creates a new panCapsuleEnvList object and returns it.  */
panCapsuleEnvList *
panCapsuleEnvList_create(void)
{
  panCapsuleEnvList *cel;

  cel = panMemory_alloc(sizeof(panCapsuleEnvList));
  if (!cel) 
    { panPrint_error("panCapsuleEnvList_create: couldn't alloc memory\n"); 
    return NULL; }
  cel->head = cel->tail = NULL;

  return cel;
}

/* Inserts an element into the tail of a panCapsuleEnvList.
 * An error happens if the element is already in another list. */
void
panCapsuleEnvList_insert(panCapsuleEnvList *celist, panCapsuleEnv *cenv)
{
  panCritRegion_ENTER();
  if (!celist || !cenv)
    { panPrint_error("panCapsuleEnvList_insert: called with null arg\n");
    panCritRegion_LEAVE();
    return; }
  if (cenv->waitnext)
    { panPrint_error("panCapsuleEnvList_insert: cenv already in a list\n");
    panCritRegion_LEAVE();
    return; }

  if (!celist->tail)
    {
      celist->tail = celist->head = cenv;
    }
  else
    {
      celist->tail->waitnext = cenv;
      celist->tail           = cenv;
    }
  panCritRegion_LEAVE();
}


/* Evaluates all of the panCapsuleEnv's in a panCapsuleEnvList in the
 * order they were added to the list if no capsules are currently
 * running.  Also removes them from the list after executing them.
 */
void
panCapsuleEnvList_evalAll(panCapsuleEnvList *celist)
{
  panCapsuleEnv *cenv;

  if (!celist) return;

  /* While no capsules are running, go through the list and consume
   * and evaluate as many capsules as we can.
   */
  while (1)		
    {
      panCritRegion_ENTER();
      if (celist->head && !panCapsuleRT_Lock && !panNode_CurrentCapsule)
	{
	  cenv = celist->head;
	  celist->head = cenv->waitnext;
	  if (celist->tail == cenv) celist->tail = NULL;
	  panCapsuleRT_Lock = 1;
	  panCritRegion_LEAVE();
	  panCapsuleEnv_evaluate(cenv);
	}
      else
	{
	  panCritRegion_LEAVE();
	  return;
	}      
    }
}


/* Frees the memory associated with a panCapsuleEnvList,
 * and also destroys all capsule environments in the list.  */
void
panCapsuleEnvList_destroy(panCapsuleEnvList *celist)
{

  panCapsuleEnv *cenv, *nextcenv;

  for (cenv = celist->head; cenv; cenv = nextcenv)
    {
      nextcenv = cenv->waitnext;
      panCapsuleEnv_destroy(cenv);
    }
  
  if (celist) panMemory_free(celist);
}


/* Takes one panCapsuleEnvList and appends it to the
 * end of another panCapsuleEnvList.  Also removes
 * them from the source panCapsuleEnvList. */
void
panCapsuleEnvList_transferList(panCapsuleEnvList *cel_dst,
			       panCapsuleEnvList *cel_src)
{
  panCritRegion_ENTER();

  if (!cel_dst || !cel_src)
    { panPrint_error("panCapsuleEnvList_insert: called with null arg\n");
    panCritRegion_LEAVE();
    return; }

  if (!cel_dst->tail)
    {
      cel_dst->tail = cel_src->tail;
      cel_dst->head = cel_src->head;
    }
  else
    {
      cel_dst->tail->waitnext = cel_src->head;
      cel_dst->tail           = cel_src->tail;
    }
  cel_src->head = NULL;
  cel_src->tail = NULL;

  panCritRegion_LEAVE();
}


