/* $Header: /home/dl2/nygren/cvs/pan/src/node/nodecodeobj.c,v 1.11 1997/12/18 02:33:26 nygren Exp $
 *
 * FILE:    nodecodeobj.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for loading and linking PAN code objects into the PAN node.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/node.h>
#include <node/environment.h>
#include <node/binloader/loader.h>

/* Prototypes */
static void panCodeobj_failLoad(panCodeobj *co);


/*
 *  Creates a new, empty list of code objects and returns a pointer to it.
 */
panCodeobjList *
panCodeobjList_create(void)
{
  panCodeobjList *col;
  
  col = panMemory_alloc(sizeof(panCodeobjList));
  if (!col) 
    { panPrint_error("alloc failed in panCodeobjList_create\n"); return NULL; }
  col->head = col->tail = NULL;
  return col;  
}


/*
 * Adds a panCodeobj to the tail of a CodeobjList.
 */
void
panCodeobjList_addToTail(panCodeobjList *list, panCodeobj *co)
{
  panCodeobjListEl *el = panMemory_alloc(sizeof(panCodeobjListEl));

  if (el == NULL) {
    panPrint_error("Couldn't allocate panCodeobjListEl\n");  return;
  }
  {
    panCritRegion_ENTER();
    el->next = NULL;
    el->co   = co;
    if (!list->tail)
      { 
	list->head = list->tail = el;
	el->prev = NULL;
      }
    else
      {
	el->prev = list->tail;
	list->tail->next = el;
	list->tail = el;
      }
    panCritRegion_LEAVE();
  }
}


/*
 * Destroys a panCodeobjList (and all associated data structures)
 * but does not modify the code objects in the list.
 */
void
panCodeobjList_destroy(panCodeobjList *col)
{
  panCodeobjListEl *e1, *e2;

  if (!col) return;
  for (e1 = col->head; e1; e1 = e2) 
    {
      e2 = e1->next;
      panMemory_free(e1);
    }
  panMemory_free(col);
}


/*
 * Looks up a panCodeobj by a panCodeobjName in a panCodeobjList.
 */
panCodeobj *
panCodeobjList_lookupByName(panCodeobjList *col, panCodeobjName *con)
{
  int i;
  panCodeobjListEl *el;

  if (!col || !con) return NULL;
  {
    panCritRegion_ENTER();
    for (el = col->head; el; el = el->next)
      {
	if (el->co)
	  {
	    /* Compare the code object names and return the 
	     * code object if they match */
	    for (i=0; i<panCodeobjName_NUMWORDS; i++)
	      if (el->co->name.data[i] != con->data[i])
		goto lookup_failed;
	    panCritRegion_LEAVE();
	    return el->co;	/* If the lookup didn't fail, they match */
	  lookup_failed: (void)0;
	  }
      }
    panCritRegion_LEAVE();
  }
  return NULL;
}


/*
 * Removes a code object from a list of code objects.
 * (does no locking!)
 */
inline void
__panCodeobjList_removeEl(panCodeobjList *col, panCodeobjListEl *el)
{
  assert (col && el);

  if (el->prev) el->prev->next = el->next;
  if (el->next) el->next->prev = el->prev;
  if (col->head == el) col->head = el->next;
  if (col->tail == el) col->tail = el->prev;
  panMemory_free(el);
}


/*
 * Removes a code object from a list of code objects.
 */
void
panCodeobjList_removeEl(panCodeobjList *col, panCodeobjListEl *el)
{
  panCritRegion_ENTER();
  __panCodeobjList_removeEl(col, el);
  panCritRegion_LEAVE();
}

/*
 * Removes a code object from a list of code objects.
 */
void
panCodeobjList_remove(panCodeobjList *col, panCodeobj *co)
{
  panCodeobjListEl *el;

  panCritRegion_ENTER();
  for (el = col->head; el; el = el->next)
    {
      if (co == el->co)
	__panCodeobjList_removeEl(col, el);
    }
  panCritRegion_LEAVE();
}


/*************************************************************/
/*************************************************************/
/*************************************************************/
/*************************************************************/


/*
 * Creates and returns a new code object but does not place 
 * it into the code cache.  Initializes its fields to sane
 * values and creates needed structures.  Should *NOT* be
 * called directly.  (Is called by panCodeobj_load).
 * The code object's homeaddr is set to homeaddr (which is
 * where the code object and its dependencies may be found).
 */
panCodeobj *
panCodeobj_create(panCodeobjName *con, panNodeAddr *homeaddr)
{
  panCodeobj *co;
  
  assert(con);
  co = panMemory_alloc(sizeof(panCodeobj));
  if (!co) 
    { panPrint_error("panCodeobj_create: couldn't malloc co\n"); return NULL; }
  
  memcpy(&(co->name), con, sizeof(panCodeobjName));
  co->status        = panCodeobj_STATUS_NEED_CODEOBJ;
  co->numref        = 0;
  co->numaccept     = 0;
  co->fn_accept     = NULL;
  co->fn_init       = NULL;
  co->fn_finalize   = NULL;
  co->symname_need_free = 0;
  co->symname       = "__WAITING_FOR_CODEOBJ__";
  co->codeobj       = NULL;
  co->cosects       = NULL;
  co->symtab        = panSymTab_create();
  co->depend        = panCodeobjList_create();
  co->waitingonthis = panCodeobjList_create();
  co->sring         = panSwsegRing_create(16);
  if (homeaddr) panNodeAddr_dup(&co->homeaddr, homeaddr);
  co->waitingcapsules = NULL;

  return co;
}


/* Takes a panStream that points to the start of a pan code object in
 * memory and loads it into a new panCodeobj.  Looks up the code
 * object in the code cache and uses one from there, otherwise,
 * creates a new one and adds it to the code cache.  Doesn't try to
 * load or link any executable code (requires that dependencies be
 * loaded first).  Looks the dependencies up in the code cache and
 * creates them if they don't exist.  Also links the code if all
 * dependencies are satisfied.   Additionally takes cohomeaddr which
 * is where to look for unavailable code object dependencies.
 * Sets the code object to having failed to load if it fails to load.
 * Returns NULL on error, the code object on success.  
 */
panCodeobj *
panCodeobj_load(panStream *startps, panCodeCache *cache, 
		panNodeAddr *cohomeaddr)
{
  panSwseg *cosws;		/* A software segment containing the
				 * code object. */
  int       totalsize;		/* The total size of the code object */
  panStream _ps, *ps=&_ps;
  panStream _pssrc, *pssrc=&_pssrc;
  int       sectnum;
  panCodeobjName coname;
  panCodeobj *co;
  
  assert(startps && cache);

  /* Test the size of the code object */
  totalsize = panCodeobj_getLength(startps);
  if (totalsize < panCodeobjHead_SIZE)
    { panPrint_warn("panCodeobj_load: short code object"
		     " (got %d, need at least %d for header)\n",
		     totalsize, panCodeobjHead_SIZE); return NULL; }

  /* Get the code object's name */
  panStream_dup(ps, startps);
  if ((panStream_seekFwd(ps, panCodeobjHead_NAME_OFF)
       != panCodeobjHead_NAME_OFF)
      || !panCodeobjName_read(&coname, ps, 1))
    { panPrint_warn("panCodeobj_load: error reading code object name.\n");
    return NULL; }

  /* Look at the name and see if it's already in the code cache.
   * Otherwise, create and insert into the code cache. */
  co = panCodeCache_lookup(cache, &coname);
  if (co)
    {
      if (panCodeobj_STATUS_FAILED_TO_LOAD_P(co->status))
	{
	  panPrint_warn("panCodeobj_load: already tried to load <%s> "
			 "and failed.  Won't try again.\n", co->symname);
	  return NULL;
	}
      else if (panCodeobj_STATUS_IS_READY_P(co->status))
	{
	  panPrint_debug("panCodeobj_load: <%s> already loaded"
			 " (status %d)\n", 
			 co->symname, co->status);
	  /* Note that no verification is done, other than comparison 
	   * of the names. */
	  return co;
	}
      else if (co->status != panCodeobj_STATUS_NEED_CODEOBJ)
	{
	  panPrint_error("panCodeobj_load: <%s> in weird state"
			 " (status %d)\n", 
			 co->symname, co->status);
	  return NULL;
	}
    }
  else
    {
      /* Verify the code object's integrity */
      panStream_dup(ps, startps);
      if (!panCodeobjName_verify(&coname, ps, totalsize))
	{ panPrint_warn("panCodeobj_load: error verifying hash name\n");
	return NULL;}

      /* Create a new code object */
      co = panCodeobj_create(&coname, cohomeaddr);
      if (!co)
	{ panPrint_warn("panCodeobj_load: couldn't create codeobj\n");
	return NULL; }
      panCodeCache_insert(cache, co);      
    }

  /* Allocate a single contiguous software segment to store the code object */ 
  cosws = panSwseg_create_byMalloc(totalsize);
  if (!cosws)
    { panPrint_error("panCodeobj_load: couldn't create cosws\n");
    panCodeobj_failLoad(co);
    return NULL; }
  panSwsegRing_add(co->sring, cosws);

  /* Duplicate the code object we're provided into cosws */
  panStream_init(ps, cosws, 0);
  panStream_dup(pssrc, startps);
  panStream_copy(ps, pssrc, totalsize);

  co->codeobj = cosws;

  /* Read in the header */
  panStream_init(ps, cosws, 0);
  if (!panCodeobjHead_read(&(co->cohead), ps, 1))
    { panPrint_warn("panCodeobj_load: invalid header\n"); 
    panCodeobj_failLoad(co); return NULL; }

  /* Read in the sections */
  co->cosects = panMemory_alloc(sizeof(panCodeobjSection) 
				* co->cohead.num_sections);
  if (!co->cosects)
    { panPrint_warn("panCodeobj_load: couldn't malloc memory for cosects!\n");
    panCodeobj_failLoad(co);
    return NULL; }
  if (!panCodeobjSection_readArray(co->cosects, ps, co->cohead.num_sections))
    { panPrint_warn("panCodeobj_load: error reading in sections\n");
    panCodeobj_failLoad(co);
    return NULL;}

  /* Update the status of the code object */
  co->status = 0;		/* Waiting on 0 dependencies */

  co->symname_need_free = 0;
  co->symname           = "__UNNAMED__";

  /* Process the sections */
  for (sectnum=0; sectnum<co->cohead.num_sections; sectnum++)
    {
      panCodeobjSection *cursect = &(co->cosects[sectnum]);
      
      if (cursect->type == panCodeobjSection_SYMBOLICNAME)
	{
	  co->symname = panMemory_alloc(cursect->length);
	  if (!co->symname) 
	    { panPrint_error("panCodeobj_load: malloc of symname failed!\n");
	    panCodeobj_failLoad(co);
	    return NULL; }
	  co->symname_need_free = 1;
	  panStream_dup(ps, &(cursect->body));
	  panStream_read((u_int8_t*)co->symname, ps, cursect->length);
	  co->symname[cursect->length-1] = '\0';
	}

      else if (cursect->type == panCodeobjSection_DEPENDENCIES)
	{
	  int numdep = cursect->length / panCodeobjName_SIZE;
	  int curdep;
	  panCodeobjName depname;
	  panCodeobj     *depco;

	  if (cursect->length%panCodeobjName_SIZE)
	    { panPrint_warn("panCodeobj_load: depend section is invalid size"
			    " -- ignoring.\n"); 
	    continue; }
	  
	  /* For each dependency, look it up in the code cache.  
	   * If it's there, add to dependency list and up its
	   * refcount and the 
	   * If it's not there, create the new code object,
	   * add it to the code cache, and up its refcount.
	   */
	  panStream_dup(ps, &(cursect->body));
	  for (curdep=0; curdep<numdep; curdep++)
	    {
	      if (!panCodeobjName_read(&depname, ps, 1))
		{ panPrint_warn("panCodeobj_load: invalid dependency name\n");
		panCodeobj_failLoad(co); 
		return NULL; }
	      depco = panCodeCache_lookup(cache, &depname);

	      if (depco)
		{
		  /* It's in the code cache, but is it ready? */
		  if (panCodeobj_STATUS_FAILED_TO_LOAD_P(depco->status))
		    {
		      panPrint_warn("panCodeobj_load: dependency <%s> of <%s>"
				    " failed to load\n", depco->symname,
				    co->symname);
		      panCodeobj_failLoad(co); 
		      return NULL;
		    }
		  else if (!panCodeobj_STATUS_IS_READY_P(depco->status))
		    {
		      co->status++;
		      panCodeobjList_addToTail(depco->waitingonthis, co);
		      panPrint_debug("panCodeobj_load: need unavailable "
				     "dependency found in code cache\n");
		      panCodeobj_demandLoad(depco);
		    }
		}
	      else   /* Not in code cache */
		{
		  depco = panCodeobj_create(&depname, cohomeaddr);
		  if (!depco)
		    { panPrint_error("panCodeobj_load: couldn't create "
				     "depco\n"); 
		    panCodeobj_failLoad(co); return NULL; }
		  panCodeCache_insert(cache, depco);
		  co->status++;
		  panCodeobjList_addToTail(depco->waitingonthis, co);
		  panPrint_debug("panCodeobj_load: need unavailable "
				 "dependency not found in code cache\n");

		  panCodeobj_demandLoad(depco);	 	  
		}	      
	      depco->numref++;
	      panCodeobjList_addToTail(co->depend, depco);
     	    } /* end foreach dependency */
	} /**** end panCodeobjSection_DEPENDENCIES ****/

      else if (cursect->type == panCodeobjSection_BINCODE)
	{
#define MALLOC_BINCODE /* This seems to be needed to keep dynamically loaded
			* code from segfaulting.  No idea why... 
			* TODO -- find out what is causing this! */
#ifdef MALLOC_BINCODE
	  co->bincode    = panMemory_alloc(cursect->length);
	  if (!co->bincode)
	     { panPrint_error("panCodeobj_load: couldn't malloc bincode\n");
	     panCodeobj_failLoad(co);
	     return NULL; }
	  memcpy(co->bincode, cursect->body.pos, cursect->length);
	  co->bincode_sz = cursect->length;
#else /* !MALLOC_BINCODE */
	  co->bincode    = cursect->body.pos;
	  co->bincode_sz = cursect->length;
#endif
	  if (cursect->body.rem < co->bincode_sz)
	    { panPrint_warn("panCodeobj_load: short bincode section "
			     "(has %d, need at least %d)!\n",
			     cursect->body.rem, co->bincode_sz);
	    panCodeobj_failLoad(co);
	    return NULL; }

	} /**** end panCodeobjSection_BINCODE ****/

      else
	{
	  panPrint_warn("panCodeobj_load: ignoring section of unknown type"
			" 0x%04x\n", cursect->type);
	}
    }
  
  panPrint_debug("panCodeobj_load: loaded <%s> but not yet linked "
		 "(waiting on %d)\n", co->symname, co->status);

  /* If all dependencies are available, go ahead and link */
  if (co->status == 0)
    {
      int retval;
      retval = panCodeobj_link(co);
      if (retval == 0)
	{
	  panPrint_warn("panCodeobj_load: error linking <%s>\n", co->symname);
	  panCodeobj_failLoad(co);
	  return NULL;
	}
      else return co;
    }

  return co;
}


/* Loads in binary code and links symbols together.  Requires that all
 * dependencies have been loaded.  Also runs the "init" method for the
 * code object, if it has one.  Finally, goes ahead and links any code
 * objects that had been waiting on this one.  Returns 0 on failure
 * and 1 on success.  
 */
int
panCodeobj_link(panCodeobj *co)
{
  panCodeobjListEl *col_el;
  panSymbol        *sym, thissym;
  int               retval;

  panPrint_debug("panCodeobj_link: linking <%s>\n", co->symname);

  /* Guarentee that all dependencies have been satisfied */
  if (co->status != 0)
    { panPrint_error("panCodeobj_link: called with invalid status %d\n",
		     co->status);
    return 0; }

  for (col_el=co->depend->head; col_el; col_el = col_el->next)
    {
      if (!panCodeobj_STATUS_IS_READY_P(col_el->co->status))
	{ panPrint_error("panCodeobj_link: called and dependency "
			 "unavailable (dependency <%s> has status %d)\n", 
			 co->symname, co->status);
	return 0; }
    }

  /* Put the pniCodeobj_This symbol into the symtab... */
  co->thisco = co;
  thissym.name = "pniCodeobj_This";  
  thissym.addr = (void (*)())&co->thisco;
  panSymTab_insert(co->symtab, &thissym);  
  panPrint_debug("XXXXXXX pniCodeobj_This == %p\n", co);

  panPrint_debug("panCodeobj_link: calling panBinLoader_load for <%s>\n",
		 co->symname);
  retval = panBinLoader_load(co);
  if (!retval)
    {
      panPrint_warn("panCodeobj_link: panBinLoader_load failed for <%s>\n",
		    co->symname);
      return 0;
    }

  panPrint_debug("panCodeobj_link: done with panBinLoader_load for <%s>\n",
		 co->symname);

  /* Look up methods used by the system */
  sym = panSymTab_lookupInContext(co, "accept");
  if (sym) co->fn_accept = (void(*)())sym->addr;
  sym = panSymTab_lookupInContext(co, "init");
  if (sym) co->fn_init = (void(*)())sym->addr;
  sym = panSymTab_lookupInContext(co, "finalize");
  if (sym) co->fn_finalize = (void(*)())sym->addr;

  co->status = panCodeobj_STATUS_IS_READY;

  /* Run the code object's init method if it has one */
  if (co->fn_init)
    {
      panPrint_debug("panCodeobj_link: running init method for <%s>\n",
		     co->symname);
      panCodeobj_evalMethod(co, panCapsuleEnv_METHOD_INIT);
    }

  panPrint_debug("panCodeobj_link: done loading and linking <%s>\n",
		 co->symname);

  /* Transfer any waiting capsule environments to the global list
   * for processing. */
  
  {
    panCritRegion_ENTER();
    if (co->waitingcapsules)
      {
	if (co->waitingcapsules->head)
	  panPrint_debug("panCodeobj_link: transfering pending capsules\n");
	panCapsuleEnvList_transferList(panNode_PendingCapsules,
				       co->waitingcapsules);
	co->waitingcapsules = NULL;
      }
    panCritRegion_LEAVE();
  }

  /* Notify any code objects waiting on this one */
  for (col_el=co->waitingonthis->head; col_el; col_el = col_el->next)
    {
      if (col_el->co->status <= 0)
	{ panPrint_error("panCodeobj_link: code object waiting for this "
			 "(<%s>) has invalid status %d\n", 
			 col_el->co->symname, col_el->co->status);
	continue; }

      col_el->co->status--;
      if (col_el->co->status == 0)
	{
	  retval = panCodeobj_link(col_el->co);
	  if (!retval) 
	    {
	      panPrint_warn("panCodeobj_link: error linking <%s>\n",
			    co->symname);
	      panCodeobj_failLoad(col_el->co);
	    }
	}
    }
  /* Destroy the waiting list now that we're done with it */
  panCodeobjList_destroy(co->waitingonthis);
  co->waitingonthis = NULL;

  return 1;
}

/* Causes a code object to fail to load 
 * (and causes all code objects waiting on it to fail to load).
 */
static void
panCodeobj_failLoad(panCodeobj *co)
{
  panCodeobjListEl *dep;

  if (!co || panCodeobj_STATUS_FAILED_TO_LOAD_P(co->status)) return;

  panPrint_debug("panCodeobj_failLoad: code object <%s> failed to load\n",
		 co->symname);
  co->status = panCodeobj_STATUS_FAILED_TO_LOAD;
  if (co->waitingonthis)
    for (dep = co->waitingonthis->head; dep; dep=dep->next)
      {
	panCodeobj_failLoad(dep->co);
      }

  {
    panCritRegion_ENTER();
    if (co->waitingcapsules)
      {
	panCapsuleEnvList_destroy(co->waitingcapsules);
	co->waitingcapsules = NULL;
      }
    panCritRegion_LEAVE();
  }
}


/* Destroys a code object and frees and unloads 
 * all associated code and data. */
void
panCodeobj_destroy(panCodeobj *co)
{
  panCodeobjListEl *el;

  if (!co) return;
  if (co->numref)
    { panPrint_error("panCodeobj_destroy: can't destroy <%s> -- %d code "
		     "depend on it\n", co->symname, co->numref); 
    return; }

  panPrint_debug("panCodeobj_destroy: unloading and destroying <%s>\n",
		 co->symname);

  {
    panCritRegion_ENTER();
    if (co->waitingcapsules)
      {
	panCapsuleEnvList_destroy(co->waitingcapsules);
	co->waitingcapsules = NULL;
      }
    panCritRegion_LEAVE();
  }

  if (co->waitingonthis)
  {
    if (!panCodeobjList_isEmptyP(co->waitingonthis))
      {
	panPrint_error("panCodeobj_destroy: can't destroy <%s> -- has "
		       "code objects waiting on it!\n", co->symname);
	return;
      }
    panCodeobjList_destroy(co->waitingonthis);
  }
      
  /* Run the finalization method */
  if (co->fn_finalize && panCodeobj_STATUS_IS_READY_P(co->status))
    panCodeobj_evalMethod(co, panCapsuleEnv_METHOD_FINALIZE);

  if (co->loader_priv)
    panBinLoader_unload(co);

  if (co->cosects)           panMemory_free(co->cosects);
  if (co->symtab)            panSymTab_destroy(co->symtab);

  /* Unreference dependencies */
  if (co->depend)     
    {
      for (el = co->depend->head; el; el = el->next)
	if (el->co)
	  {
	    el->co->numref--;	
	    if (el->co->waitingonthis)
	      panCodeobjList_remove(el->co->waitingonthis, co);
	  }
      panCodeobjList_destroy(co->depend);
    }

#ifdef MALLOC_BINCODE
  if (co->bincode) panMemory_free(co->bincode);
#endif

  if (co->symname_need_free) panMemory_free(co->symname);  

  if (co->sring) panSwsegRing_destroy(co->sring);

  panMemory_free(co);
}


/*
 * Requests that a code object be demand-loaded.  The
 * code object should already have been created and placed into
 * the code cache.  The code object should also have 
 * a homeaddr specified. 
 */
void
panCodeobj_demandLoad(panCodeobj *co)
{
  panSwseg *capsws;
  int       caplength;
  panStream _ps, *ps = &_ps;
  panNodeAddr localaddr;
  
#if 0
  {
    char coname_str[panCodeobjName_STRINGIFY_STRLEN];
    
    panCodeobjName_stringify(coname_str, &co->name);
    panPrint_msg("*TO*DO* panCodeobj_demandLoad: load codobj %s <%s> "
		 "from homeaddr ",
		 coname_str, co->symname);
    panNodeAddr_print(&co->homeaddr);
    panPrint_msg("\n");
  }
#endif

  /* Still need to implement a better way of loading this in... */
  if (!panCodeobj_DynamicLoadCodeobj)
    { panPrint_error("panCodeobj_demandLoad: "
		     "dynamic load code object not available!\n");
    return; }

  panNodeAddr_getLocalAddr(&localaddr);

  /* A first cut at doing demand loading... (Construct a code object
   * for dynamic loading, then evaluate it). */
  caplength = 2*panCodeobjName_SIZE + 3*panNodeAddr_SIZE + 4;
  capsws = panSwseg_create_byMalloc(caplength);
  if (!capsws)
    { panPrint_error("panCodeobj_demandLoad: couldn't create swseg!\n");
    return; }
  panStream_init(ps, capsws, 0);
  panCodeobjName_write(ps, &panCodeobj_DynamicLoadCodeobj->name);
  panNodeAddr_write(ps, &localaddr);
  panNodeAddr_write(ps, &co->homeaddr);
  panNodeAddr_write(ps, &localaddr);
  panStream_write_net_int32(ps, 0);
  panCodeobjName_write(ps, &co->name);

  panStream_init(ps, capsws, 0);
  panNode_acceptCapsule(ps, caplength, NULL);
}


/* Loads a code object from a contiguous region of memory.  Used
 * primarily for loading base code objects that were statically linked
 * into the text section of the executable (for example, code objects
 * for dynamic loading and inserting other code objects).  */
panCodeobj *
panCodeobj_loadFromMemory(char *costart, int colength,
			  panCodeCache *cache, panNodeAddr *cohomeaddr)
{
  panSwseg  *cosws;
  panStream _ps, *ps=&_ps;

  cosws = panSwseg_create_static(costart, colength);
  if (!cosws) 
    { panPrint_error("panCodeobj_loadFromMemory: couldn't create swseg\n");
    return NULL; }
  panStream_init(ps, cosws, 0);  
  return panCodeobj_load(ps, cache, cohomeaddr);
}

