/* $Header: /home/dl2/nygren/cvs/pan/src/node/nodecodeobj.h,v 1.8 1997/12/18 02:33:27 nygren Exp $
 *
 * FILE:    nodecodeobj.h
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for code objects loaded into a PAN node
 *
 */

#ifndef NODE_NODECODEOBJ_H_INC
#define NODE_NODECODEOBJ_H_INC

#include <common/pan-copyright.h>
#include <common/pancodeobj.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/symtab.h>
#include <common/nodeaddr.h>


/* A panCodeobjList is a list of code objects that code objects can be
 * added to, removed from, and searched through. */
typedef struct _panCodeobjList {
  struct _panCodeobjListEl *head;
  struct _panCodeobjListEl *tail;
} panCodeobjList;
typedef struct _panCodeobjListEl {
  struct _panCodeobjListEl *next;
  struct _panCodeobjListEl *prev;
  struct _panCodeobj       *co;
} panCodeobjListEl;


/* A panCodeobj record contains information about a code
 * object loaded into a PAN node.  This includes the
 * name of the code object, information about its status,
 * its symbol table, function pointers to symbols that
 * need to be accessed quickly (ie for accepting a packet),
 * etc.
 */
typedef struct _panCodeobj {
  panCodeobjName     name;	/* The unique name of the code object.
				 * This is usually a unique cryptographic hash
				 * of the code object.  The name is 
				 * also used for looking the code object
				 * up in the code cache.  Therefore, a
				 * substring of it should be a good
				 * universal hash.  */

  int                status;	/* The status of this code object.
				 * Valid values include:
				 *    -1 == loaded, linked, and ready to run
				 *    -2 == waiting for code object 
				 *    -3 == we tried to load this before
				 *          and it failed to load
				 *  >= 0 == number of dependancies waiting for
				 */
#define panCodeobj_STATUS_IS_READY         -1
#define panCodeobj_STATUS_NEED_CODEOBJ     -2
#define panCodeobj_STATUS_FAILED_TO_LOAD   -3
#define panCodeobj_STATUS_IS_READY_P(x)     ((x)==panCodeobj_STATUS_IS_READY)
#define panCodeobj_STATUS_NEED_CODEOBJ_P(x) ((x)==panCodeobj_STATUS_NEED_CODEOBJ)
#define panCodeobj_STATUS_FAILED_TO_LOAD_P(x) ((x)==panCodeobj_STATUS_FAILED_TO_LOAD)
#define panCodeobj_STATUS_NEED_DEPENDS_P(x) ((x)>=0)

  int                numref;	/* The number of code objects depending
				 * on this one. */

  int                numaccept;	/* The number of capsules that this code
				 * object has accepted.  */
				
  void             (*fn_accept)(panStream *capsule, int caplength); 
				/* A function pointer to a routine to
				 * accept a capsule.  Takes a pointer to a
				 * panStream starting a capsule
				 * and the number of bytes in the capsule. */

  void             (*fn_init)(void); 
				/* A function pointer to a routine to
				 * run when this code object becomes
				 * ready to run. */

  void             (*fn_finalize)(void); 
				/* A function pointer to a routine to
				 * run before finalizing this code object */

  int                symname_need_free;  /* Was the symname alloced? */
  char              *symname;	/* The symbolic name of this code object
				 * (for debugging purposes). */

  panSwseg          *codeobj;	/* A software segment containing the
				 * on-the-wire code object (so it can
				 * be sent on to other sites that
				 * request it).  
				 * *** THIS MUST BE A CONTIGUOUS REGION
				 * *** OF MEMORY!!! 
				 */

  panNodeAddr        homeaddr;	/* The home address of the code object */
  
  void              *loader_priv;  /* Private information stored by
				    * the code loader */

  u_int8_t          *bincode;   /* The start of the bincode region 
				 * in memory. */

  int                bincode_sz; /* The size of the bincode region 
				  * in memory. */

  panCodeobjHead     cohead;	/* The header for the code object. */

  panCodeobjSection *cosects;	/* A list of cohead.num_sects sections
				 * (with references into cohead). */
  
  panSymTab         *symtab;	/* The symbol table for this code object */

  panCodeobjList    *depend;	/* A list of code objects that this code
				 * object depends on. */

  panCodeobjList    *waitingonthis;/* A list of code objects that depend on
				    * this code object (and are waiting
				    * for it to load). */

  struct _panCapsuleEnvList *waitingcapsules; /* A list of capsules waiting to
					       * be evaluated when this code 
					       * object becomes available. */

  panSwsegRing      *sring;	/* Software segments that are associated
				 * with this code object and which should 
				 * be released when it is destroyed. */
  struct _panCodeobj *thisco;	/* A pointer to this structure.
				 * This pointer is used for the 
				 * pniCodeobj_This symbol. */

} panCodeobj;


#include <node/codecache.h>

#define panCodeobjList_isEmptyP(col) ((!col) || ((col)->head == NULL))

extern panCodeobjList *
panCodeobjList_create(void);

extern void
panCodeobjList_addToTail(panCodeobjList *col, panCodeobj *co);

extern void
panCodeobjList_remove(panCodeobjList *col, panCodeobj *co);

extern void
panCodeobjList_removeEl(panCodeobjList *col, panCodeobjListEl *el);

extern void
panCodeobjList_destroy(panCodeobjList *col);

extern panCodeobj *
panCodeobjList_lookupByName(panCodeobjList *col, panCodeobjName *con);


extern panCodeobj *
panCodeobj_create(panCodeobjName *con, panNodeAddr *homeaddr);

extern panCodeobj *
panCodeobj_load(panStream *startps, panCodeCache *cache, 
		panNodeAddr *cohomeaddr);

extern int
panCodeobj_link(panCodeobj *co);

extern void
panCodeobj_demandLoad(panCodeobj *co);

extern void
panCodeobj_destroy(panCodeobj *co);

/* Code in node/symtab.c */
extern panSymbol *
panSymTab_lookupInContext(panCodeobj *context, char *name);

extern panCodeobj *
panCodeobj_loadFromMemory(char *costart, int colength,
			  panCodeCache *cache, panNodeAddr *cohomeaddr);

extern int
pniCodeobj_load_immediate(panStream *startps, panNodeAddr *homeaddr);

#endif /* NODE_NODECODEOBJ_H_INC */
