/* $Header: /home/dl2/nygren/cvs/pan/src/node/pni/pniswseg.c,v 1.5 1997/10/14 21:29:56 nygren Exp $
 *
 * FILE:    pniswseg.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Implements swseg routines that may be called from executing
 * capsules through the PAN Node Interface.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include <pni/pniswseg.h>
#include <pni/pnicodeobj.h>



/* Creates a software segment and adds it to the
 * keyring of the current environment.  
 * Aborts the capsule's execution if either fails.
 * Only returns NULL if there's no current environment.
 */
pniSwseg *
pniSwseg_create_byMalloc(unsigned length)
{
  pniSwseg *sws;

  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniSwseg_create_byMalloc: "
		     "called with no running capsule!\n");
      return NULL;
    }

  sws = panSwseg_create_byMalloc(length);
  if (!sws)
    {
      panCapsuleEnv_abort(panNode_CurrentCapsule,
			  "pniSwseg_create_byMalloc: couldn't get a swseg");
    }
  panSwsegRing_add(panNode_CurrentCapsule->sring, sws);
  return sws;
}

/* Creates an overlay swseg and adds it to the
 * keyring of the current environment.  
 * Aborts the capsule's execution if either fails.
 * Only returns NULL if there's no current environment.
 */
pniSwseg *
pniSwseg_overlay(pniSwseg *sws, unsigned off, unsigned length)
{
  pniSwseg *newsws;

  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniSwseg_overlay: "
		     "called with no running capsule!\n");
      return NULL;
    }

  newsws = panSwseg_overlay(sws, off, length);
  if (!newsws)
    {
      panCapsuleEnv_abort(panNode_CurrentCapsule,
			  "pniSwseg_overlay: couldn't get a swseg");
    }
  panSwsegRing_add(panNode_CurrentCapsule->sring, newsws);
  return newsws;
}

/* Makes a software segment persistent by adding it to the 
 * code object's key ring.  (So that it will get flushed 
 * only when the code object gets flushed).
 * Also ups the reference count on the swseg.
 */
void
pniSwseg_makePersistent(pniCodeobj *thisco, pniSwseg *sws)
{
  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniSwseg_makePersistent: "
		     "called with no running capsule!\n");
      return;
    }
  if (!thisco)
    {
      panPrint_error("pniSwseg_makePersistent: "
		     "called with NULL thisco!\n");
      return;      
    }

  panSwseg_ref(sws);
  panPrint_debug("pniSwseg_makePersistent: making persistent (now %d refs)\n",
		 sws->numref);
  panSwsegRing_add(thisco->sring, sws);
}


/* Releases a persistent software segment by removing it from the 
 * code object's key ring.
 * Also downs the reference count on the swseg.
 */
void
pniSwseg_releasePersistent(pniCodeobj *thisco, pniSwseg *sws)
{
  int refcount, numremoved;

  if (!panNode_CurrentCapsule 
      || panNode_CurrentCapsule->status != panCapsuleEnv_STATUS_RUNNING)
    {
      panPrint_error("pniSwseg_releasePersistent: "
		     "called with no running capsule!\n");
      return;
    }
  if (!thisco)
    {
      panPrint_error("pniSwseg_releasePersistent: "
		     "called with NULL thisco!\n");
      return;      
    }
  refcount = sws->numref;
  numremoved = panSwsegRing_remove(thisco->sring, sws, 0);
  panPrint_debug("pniSwseg_releasePersistent: "
		 "released %d persistent (%d refs)\n",
		 numremoved, refcount-numremoved);
}

