/* $Header: /home/dl2/nygren/cvs/pan/src/node/pni/pniutil_hashtab.c,v 1.3 1998/01/23 07:43:50 nygren Exp $
 *
 * FILE:    pniutil_hashtab.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Utility routines provided through the PAN Node Interface for
 * a hash table.  This set of utility routines implements a data structure
 * that, in theory, doesn't *need* to be part of the node interface.
 * It's just provided here for convenience.
 *
 * The hashtab is implemented as a panSwsegContainer with a data
 * body that contains a pniHashTabHeader.  The header contains
 * function pointers and accounting info and a pointer to the
 * hash table itself.  The hash table contains a fixed number of
 * elements where each contains a pointer to a linked list.
 * The linked list contains hash table keys and pointers to 
 * swsegs.  Keys are byte arrays that are application-dependant.
 *
 *  *TO*DO*  Make the linked lists be sorted by key values.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <node/nodecodeobj.h>
#include <node/codecache.h>
#include <node/environment.h>
#include <node/node.h>

#include <pni/pniswseg.h>
#include <pni/pniutil_hashtab.h>



typedef struct _pniHashTabEntry {
  struct _pniHashTabEntry *prev, *next;
  int    origsize;		/* The original size at time of insertion */
  pniSwseg *value;
  char   key[4];		/* MUST be last!!! (Actually longer than
				 * 4 bytes in most cases). */
} pniHashTabEntry;


typedef struct _pniHashTabHeader {
  int magic;			/* Magic value for hashtab */
#define pniHashTabHeader_MAGIC 0x6f78e56c
  int numslots;			/* Number of slots in the table */
  int keysize;			/* The number of bytes in a key */
  pniHashTabEntry **tab;	/* The hash table itself (with numslots) */
  unsigned (*hashfn)(char *);	/* Takes a pointer to a key of keysize
				 * bytes and returns a value between 0
				 * and numslots-1. */
  int cursize;			/* The current estimated size of the hash table
				 * (in bytes). */
} pniHashTabHeader;


static void 
pniHashTab_finalize(panSwseg *sws);

static int
pniHashTab_getSize(panSwseg *sws);

static int
pniHashTab_release(panSwseg *sws, int numbytes);


/* Creates a new hash table with the specified number of slots,
 * keysize (in bytes), and hash function. */
pniHashTab *
pniHashTab_create(int numslots, int keysize,
		  unsigned (*hashfn)(char *))
{
  panSwsegContainer *ht;
  pniHashTabHeader  *hth;

  if ((numslots<=0) || (keysize<=0) || !hashfn)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_create: called with invalid args\n");
  
  ht = (panSwsegContainer*)panSwsegContainer_create(sizeof(pniHashTabHeader));
  if (!ht) 
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_create: container creation failed\n");
  hth = (pniHashTabHeader*)ht->sws.data;
  
  hth->tab = (pniHashTabEntry**)panMemory_calloc(numslots, 
						    sizeof(pniHashTabEntry*));
  hth->magic    = pniHashTabHeader_MAGIC;
  hth->numslots = numslots;
  hth->keysize  = keysize;
  hth->hashfn   = hashfn;
  hth->cursize  = 0;		/* Should base size include table size??? 
				 * Probably, but right now it doesn't. */
  ht->getsize   = pniHashTab_getSize;
  ht->finalize  = pniHashTab_finalize;
  ht->release   = pniHashTab_release;

  return (pniHashTab*)ht;
}


/* Looks up a hashtable entry.  Returns the swseg matching the key, otherwise 
 * returns NULL.  If a swseg is returned and a capsule is running,
 * the swseg is referenced and added the the capsule's key ring 
 * before being returned.
 */
pniSwseg *
pniHashTab_lookup(pniHashTab *ht, char *key)
{
  unsigned int tabidx;
  pniHashTabHeader  *hth;
  pniHashTabEntry   *entry;
  
  if (ht->type != panSwseg_TYPE_CONTAINER || !ht->data || !key)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_lookup: passed invalid argument\n");
  hth = (pniHashTabHeader*)ht->data;
#if 0
  if (hth->magic != pniHashTabHeader_MAGIC)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_lookup: passed container w/ bad magic\n");
#endif
    
  tabidx = ((hth->hashfn)(key)) % hth->numslots;

  for (entry = hth->tab[tabidx]; entry; entry = entry->next)
    {
      if (0 == memcmp(key, &(entry->key[0]), hth->keysize))
	{
	  if (panNode_CurrentCapsule)
	    {
	      panSwseg_ref(entry->value);
	      panSwsegRing_add(panNode_CurrentCapsule->sring, entry->value);
	    }			       
	  return entry->value;
	}
    }
  return NULL;
}


/* Removes and unreferences a hashtable entry.  
 * Returns 1 if found and removed, else 0. */
int
pniHashTab_remove(pniHashTab *ht, char *key)
{
  unsigned int tabidx;
  pniHashTabHeader  *hth;
  pniHashTabEntry   *entry;
  
  panPrint_debug("pniHashTab_remove: entered\n");

  if (ht->type != panSwseg_TYPE_CONTAINER || !ht->data || !key)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_remove: passed invalid argument\n");
  hth = (pniHashTabHeader*)ht->data;
  if (hth->magic != pniHashTabHeader_MAGIC)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_remove: passed container w/ bad magic\n");
  
  tabidx = ((hth->hashfn)(key)) % hth->numslots;

  for (entry = hth->tab[tabidx]; entry; entry = entry->next)
    {
      if (0 == memcmp(key, &(entry->key[0]), hth->keysize))
	{
	  /* Found, now remove... */
	  if (entry->next) entry->next->prev = entry->prev;
	  if (entry->prev)
	    entry->prev->next = entry->next;
	  else
	    hth->tab[tabidx] = NULL;
	  hth->cursize -= entry->origsize;
	  panSwseg_unref(entry->value);
	  panMemory_free(entry);
	  return 1;
	}
    }
  return 0;
}


/* Inserts an entry into the hashtable with the specified
 * key and references the value.  If the key is already
 * in the table, replaces the old value with this new value.
 * Returns 1 if the key was already in the table, else 0.
 */
int
pniHashTab_insert(pniHashTab *ht, char *key, pniSwseg *value)
{
  unsigned int tabidx;
  pniHashTabHeader  *hth;
  pniHashTabEntry   *entry;
  
  if (ht->type != panSwseg_TYPE_CONTAINER || !ht->data || !value || !key)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_insert: passed invalid argument\n");
  hth = (pniHashTabHeader*)ht->data;
  if (hth->magic != pniHashTabHeader_MAGIC)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_insert: passed container w/ bad magic\n");

  tabidx = ((hth->hashfn)(key)) % hth->numslots;

  for (entry = hth->tab[tabidx]; entry; entry = entry->next)
    {
  
      if (0 == memcmp(key, &(entry->key[0]), hth->keysize))
	{
	  /* Found, now replace the value */
	  panSwseg *oldval = entry->value;

	  panSwseg_ref(value);
	  entry->value = value;
	  panSwseg_unref(oldval);
	  hth->cursize -= entry->origsize;
	  entry->origsize = panSwseg_getSize(value);
	  hth->cursize += entry->origsize;

	  return 1;
	}
    }

  /* Not found.  Create and insert a new entry (4 bytes larger than we need) */
  entry = panMemory_alloc(sizeof(pniHashTabEntry)+hth->keysize);
  if (!entry)
    panCapsuleEnv_abort(panNode_CurrentCapsule,
			"pniHashTab_insert: malloc failed\n");

  memcpy(&(entry->key[0]), key, hth->keysize);

  panSwseg_ref(value);
  entry->value = value;
  entry->next  = NULL;
  entry->origsize = panSwseg_getSize(value);
  hth->cursize += entry->origsize;
  if (hth->tab[tabidx])
    {
      entry->next = hth->tab[tabidx];
      hth->tab[tabidx]->prev = entry;      
    }
  entry->prev = NULL;
  hth->tab[tabidx] = entry;

  return 0;
}


/* Finalizes a hashtab and frees and unreferences its contents. */
static void 
pniHashTab_finalize(panSwseg *sws)
{
  pniHashTabHeader  *hth;
  pniHashTabEntry   *entry, *nextentry;
  int i;

  if (sws->type != panSwseg_TYPE_CONTAINER || !sws->data)
    { 
      panPrint_error("pniHashTab_finalize: passed invalid hashtab\n");
      return;
    }

  hth = (pniHashTabHeader*)sws->data;
  if (hth->magic != pniHashTabHeader_MAGIC)
    {
      panPrint_error("pniHashTab_finalize: passed container w/ bad magic\n");
      return;
    }

  for (i=0; i < hth->numslots; i++)
    {
      for (entry=hth->tab[i]; entry; entry=nextentry)
	{
	  nextentry = entry->next;
	  panSwseg_unref(entry->value);
	  panMemory_free(entry);
	}
    }
  panMemory_free(hth->tab);
}


/* Returns the size of the contents of a hashtab. */
static int
pniHashTab_getSize(panSwseg *sws)
{
  pniHashTabHeader  *hth;

  if (sws->type != panSwseg_TYPE_CONTAINER || !sws->data)
    { 
      panPrint_error("pniHashTab_getSize: passed invalid hashtab\n");
      return 0;
    }
  hth = (pniHashTabHeader*)sws->data;
  if (hth->magic != pniHashTabHeader_MAGIC)
    {
      panPrint_error("pniHashTab_getSize: passed container w/ bad magic\n");
      return 0;
    }

  return hth->cursize;
}


/* Releases numbytes worth of data from the hash table. */
static int
pniHashTab_release(panSwseg *sws, int numbytes)
{
  panPrint_error("pniHashTab_release: not yet implemented!!!\n");
  return 0;
}

