/* $Header: /home/dl2/nygren/cvs/pan/src/node/symtab.c,v 1.3 1997/12/10 06:58:31 nygren Exp $
 *
 * FILE:    symtab.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Implementation of symbol table that maps string symbol
 * names to pointers.
 *
 */

#include <common/pan-copyright.h>
#include <node/nodecodeobj.h>
#include <node/symtab.h>

panSymTab *panSymTab_GlobalSymTab = NULL;

static void
panSymTab_initGlobal(void);

/*
 *  Creates a new, empty symbol table and returns a pointer to it.
 */
panSymTab *
panSymTab_create(void)
{
  panSymTab *symtab;
  
  symtab = panMemory_alloc(sizeof(panSymTab));
  if (!symtab)
    { panPrint_error("couldn't alloc memory for symtab in panSymTab_create\n");
    return NULL; }
  symtab->head = NULL;
  return symtab;
}

/* 
 * Inserts a symbol into a symbol table.  Copies the contents
 * of the panSymbol sym.  (but does not copy data pointed to by sym).
 */
void
panSymTab_insert(panSymTab *symtab, panSymbol *sym)
{
  panSymTabEntry *entry;
  
  entry = panMemory_alloc(sizeof(panSymTabEntry));
  if (!entry)
    { panPrint_error("couldn't alloc memory for entry in panSymTab_insert\n");
    return; }
  entry->next = symtab->head;
  symtab->head = entry;
  entry->sym   = *sym;
}

/* 
 * Looks up a symbol by a name and returns it
 * if found (or returns NULL).
 */
panSymbol *
panSymTab_lookup(panSymTab *symtab, char *name)
{
  panSymTabEntry *entry;
  
  assert(symtab && name);
  for (entry = symtab->head; entry; entry = entry->next)
    {
      if (0 == strcmp(name, entry->sym.name))
	return &(entry->sym);
    }
  return NULL;
}

/* 
 * Destroys a symbol table and frees all associated memory
 */
void
panSymTab_destroy(panSymTab *symtab)
{
  panSymTabEntry *entry, *tmpentry;
  
  if (!symtab) return;
  for (entry = symtab->head; entry; entry = tmpentry)
    {
      tmpentry = entry->next;
      panMemory_free(entry);
    }
  panMemory_free(symtab);
}

/* 
 * Looks up a symbol by a name and returns it
 * if found (or returns NULL).  Looks the name
 * up in "context", meaning that the symbol will first be
 * searched for in this code object, then in the dependencies specified
 * in the code object (in the order they were specified), and
 * finally in table of global exported symbols.
 */
panSymbol *
panSymTab_lookupInContext(panCodeobj *context, char *name)
{
  panSymbol        *sym;
  panCodeobjListEl *coel;
  
  assert(context && name);

  if (sym = panSymTab_lookup(context->symtab, name))
    return sym;

  for (coel = context->depend->head; coel; coel = coel->next)
    {
      if (!panCodeobj_STATUS_IS_READY_P(coel->co->status))
	{
	  panPrint_error("panSymTab_lookupInContext: "
			 "found unready dependency while looking up <%s>\n",
			 name);
	  continue;
	}
      if (sym = panSymTab_lookup(coel->co->symtab, name))
	return sym;
    }
  if (!panSymTab_GlobalSymTab)
    {
      panSymTab_initGlobal();
    }
  if (panSymTab_GlobalSymTab)
    {
      if (sym = panSymTab_lookup(panSymTab_GlobalSymTab, name))
	return sym;
    }
  
  return NULL;
}


/* 
 * Initializes the global symbol table for the first time
 * with the contents of panSymTab_GlobalSymbols_PlatformIndependent
 * and panSymTab_GlobalSymbols_PlatformDependent.
 */
static void
panSymTab_initGlobal(void)
{
  int i;

  if (panSymTab_GlobalSymTab)
    {
      panPrint_error("panSymTab_initGlobal: called and panSymTab_GlobalSymTab"
		     " already exists.\n");
      return;
    }
  panPrint_msg("panSymTab_initGlobal: initializing panSymTab_GlobalSymTab\n");

  panSymTab_GlobalSymTab = panSymTab_create();
  assert(panSymTab_GlobalSymTab);

  for (i=0; panSymTab_GlobalSymbols_PlatformIndependent[i].name; i++)
    {
      panSymTab_insert(panSymTab_GlobalSymTab, 
		       &(panSymTab_GlobalSymbols_PlatformIndependent[i]));
    }

  for (i=0; panSymTab_GlobalSymbols_PlatformDependent[i].name; i++)
    {
      panSymTab_insert(panSymTab_GlobalSymTab, 
		       &(panSymTab_GlobalSymbols_PlatformDependent[i]));
    }
}
