/* $Header: /home/dl2/nygren/cvs/pan/src/panotool/old/pancodeobj_util.c,v 1.2 1997/12/10 06:58:45 nygren Exp $
 *
 * FILE:    pancodeobj.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Routines for reading in parts of a pan code object.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <common/pancodeobj.h>
#include "misc.h"
#include "pancodeobj_util.h"

#define CAST_BUF_POS_TYPE(buf,pos,type) (*((type*)&((buf)[(pos)])))

/* Reads the header of a code object from a FILE.  Seeks to
 * the beginning of the code object.
 * If testvalidity is true, also does some simple tests on the validity
 * of the code object being read in (magic number, etc) and returns NULL
 * it they fail.  Also return NULL on any error.
 */
PANcodeobjheader *
panotool_codeobj_readheader(FILE *infile, char *infilename, int testvalidity)
{
  PANcodeobjheader *coh;
  u_int8_t inbuf[PANCODEOBJHEAD_SIZE];
  int i;
  u_int16_t tmpu16;

  if (!infile) return NULL;
  if (!(coh = malloc(sizeof(PANcodeobjheader))))
    { fprintf(stderr, "couldn't malloc PANcodeobjheader\n"); return NULL; }

  fseek(infile, 0L, SEEK_SET);

  if (PANCODEOBJHEAD_SIZE != fread(inbuf, PANCODEOBJHEAD_SIZE, 1, infile))
    {
      if (feof(infile))
	fprintf(stderr, "Premature end of file reading %s\n", infilename);
      else
	fprintf(stderr, "Error reading %s\n", infilename);
      free(coh);
      return NULL;
    }

  for (i=0; i<4; i++)
    {
      coh->magic[i] = (int8_t)(inbuf[PANCODEOBJHEAD_MAGIC_OFF+i]);
      if (testvalidity && coh->magic[i] != PANCODEOBJHEAD_MAGIC[i])
	{ fprintf(stderr, "Magic number of %s does not match!\n", infilename);
	free(coh); return NULL; }
    }

  if (!panotool_codeobjname_readfrombuf(&(coh->name), 
					inbuf+PANCODEOBJHEAD_NAME_OFF,
					testvalidity))
    { free(coh); return NULL; }

  tmpu16 = CAST_BUF_POS_TYPE(inbuf, PANCODEOBJHEAD_CODETYPE_OFF, u_int16_t);
  coh->code_type = htons(tmpu16);

  tmpu16 = CAST_BUF_POS_TYPE(inbuf, PANCODEOBJHEAD_CODEVERS_OFF, u_int16_t);
  coh->code_vers = htons(tmpu16);

  tmpu16 = CAST_BUF_POS_TYPE(inbuf, PANCODEOBJHEAD_APITYPE_OFF, u_int16_t);
  coh->api_type = htons(tmpu16);

  tmpu16 = CAST_BUF_POS_TYPE(inbuf, PANCODEOBJHEAD_APIVERS_OFF, u_int16_t);
  coh->api_vers = htons(tmpu16);

  tmpu16 = CAST_BUF_POS_TYPE(inbuf, PANCODEOBJHEAD_NUMSECT_OFF, u_int16_t);
  coh->num_sections = htons(tmpu16);

  tmpu16 = CAST_BUF_POS_TYPE(inbuf, PANCODEOBJHEAD_PAD_OFF, u_int16_t);
  coh->padding = htons(tmpu16);

  /* May want to do some verification of version numbering... */

  return coh;
}

/* Copies a code object name out of a buffer into a PANcodeobjname.
 * If testvalidity is true, does a simple validity test. 
 * Returns 0 on error.  Otherwise returns 1.
 */
int
panotool_codeobjname_readfrombuf(PANcodeobjname *coname, u_int8_t *buf,
				 int testvalidity)
{
  u_int16_t tmpu16;
  int i;

  if (!coname || !buf) return 0;

  coname->type[0] = (int8_t)buf[0];
  coname->type[1] = (int8_t)buf[1];
  if (testvalidity && (coname->type[0] != 'S' || coname->type[0] != 'H'))
    { fprintf(stderr, "Currently only SHA1 code object names are supported\n");
    return 0; }
  
  tmpu16 = CAST_BUF_POS_TYPE(buf,3,u_int16_t);
  coname->vers = ntohs(tmpu16);
  if (testvalidity && coname->vers != PANCODEOBJNAME_VERS_SHA1)
    fprintf(stderr, "Warning SHA1 version mismatch (got 0x%4hx, current is 0x%4hx\n", coname->vers, PANCODEOBJNAME_VERS_SHA1);
  
  for (i=0; i<PANCODEOBJNAME_SIZE; i++)
      coname->hash[i] = buf[i];

  return 1;
}

/* Copies a code object name to a buffer from a PANcodeobjname. */
void
panotool_codeobjname_writetobuf(u_int8_t *buf, PANcodeobjname *coname)
{
  int i;

  if (!coname || !buf) return;
  buf[0] = (u_int8_t)(coname->type[0]);
  buf[1] = (u_int8_t)(coname->type[1]);
  CAST_BUF_POS_TYPE(buf,3,u_int16_t) = htons(coname->vers);
  for (i=0; i<PANCODEOBJNAME_SIZE; i++)
    buf[i] = coname->hash[i];
}

