/* $Header: /home/dl2/nygren/cvs/pan/src/panotool/panotool.c,v 1.6 1997/12/13 05:15:07 nygren Exp $
 *
 * FILE:    panotool.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Tool for manipulating .po files (which contain PAN code objects).
 * This version currently only works with ix86 binary code objects.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>
#include <common/pancodeobj.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>

#include "misc.h"

int
panStream_write_toFile(FILE *fp, panStream *ps, unsigned length);
void
panotoolCodeobj_create(char *outfilename, char *infilename, 
		       stringlist *incpath, stringlist *dependfilenames);
int
panotoolCodeobj_test(char *infilename);
panSwseg *
panotoolCodeobj_getCodeobjName(char *filename, stringlist *incpath);


#define ADDTOKEYRING(x) if (0)

#define PANCODEOBJECT_EXTENSION ".po"

static char usage_string[] = 
"\nUsage for %s:\n"
"\n"
"  panotool -T [-v] (infile.po)* \n"
"  panotool -C [-v] [-I pathname] [-o outfile.po] infile.o (dependfile.po)*\n"
"  panotool -X [-v] [-o outfile.o] infile.po \n"
"  panotool -h \n"
"\n"
"     -T         Causes a .po file to be read in and tested for integrity. \n"
"     -C         Creates a new .po file from a .o file.  Any additional \n"
"                     .po files listed are incorporated as dependencies. \n"
"     -X         Extracts a .o file out of a .po file. \n"
"     -h         Prints out this usage message. \n"
"     -v         Verbosely prints out messages. \n"
"     -I dir     Adds <dir> to the path of directories to be searched for\n"
"                     dependencies.\n"
"     -o outfile Outputs to outfile.  By default, generates the output \n"
"                     from the input filename. \n"
"\n";

#define MODE_TEST    1		/* The -T option was specified */
#define MODE_CREATE  2		/* The -C option was specified */
#define MODE_EXTRACT 3		/* The -X option was specified */

int opt_verbose = 0;

void
print_usage_and_exit(char *progname, char *errmsg, char *errarg)
{
  if (errmsg) panPrint_error( "%s: ", progname);
  if (errarg && errmsg) panPrint_error( errmsg, errarg);
  else if (errmsg) panPrint_error( errmsg);
  panPrint_error( usage_string, progname);
  exit(-1);
}

void 
main(int argc, char *argv[])
{
  char *infilename = NULL;	/* The input filename */
  char *outfilename = NULL;	/* The output filename */
  int   panotool_mode = 0;	/* What action this program should take */
  stringlist *incpath,		/* Include path */
    *dependfilenames,		/* Dependancy filenames */
    *testfiles;			/* Files to test */
  int ac;			/* Argument counter */

  incpath = stringlist_create();
  dependfilenames = stringlist_create();
  testfiles = stringlist_create();
  assert(incpath && dependfilenames);

  /* Process command line arguments... */
  for (ac = 1; ac < argc; ac++)
    {
      if (!argv[ac][0])
	{
	  print_usage_and_exit(argv[0], 
			       "Empty command-line option\n", NULL);
	}
      else if (argv[ac][0] == '-')  /* Options starting with "-" */
	{
	  switch (argv[ac][1])
	    {
	    case 'T':
	      panotool_mode = MODE_TEST;
	      break;
	    case 'C':
	      panotool_mode = MODE_CREATE;
	      break;
	    case 'X':
	      print_usage_and_exit(argv[0], 
				   "Extraction mode is not yet supported\n",
				   NULL);
	      panotool_mode = MODE_EXTRACT;
	      break;
	    case 'h':
	      print_usage_and_exit(argv[0], NULL, NULL);	      
	      break;
	    case 'v':
	      opt_verbose = 1;
	      break;
	    case 'I':
	      if (argv[ac][2])
		stringlist_addtotail(incpath, &(argv[ac][2]), 1); /* dup&add */
	      else
		{		  
		  if (++ac < argc)
		    stringlist_addtotail(incpath, argv[ac], 1); /* dup&add */
		  else
		    print_usage_and_exit(argv[0], "No argument passed to -I\n",
					 NULL);
		}
	      break;
	    case 'o':
	      if (argv[ac][2])
		outfilename = &(argv[ac][2]);
	      else
		{
		  if (++ac < argc)
		    outfilename = argv[ac];
		  else
		    print_usage_and_exit(argv[0], "No argument passed to -I\n",
					 NULL);
		}
	      break;

	    default:
	      print_usage_and_exit(argv[0], 
				   "Invalid command line option '%s'\n",
				   argv[ac]);
	    }
	}
      else  /* Options not starting with "-", ie filenames. */
	{
	  if (!panotool_mode) 	      
	    print_usage_and_exit(argv[0], "Must specify -T, -C, or -X before filenames.\n", NULL);
	  if (panotool_mode == MODE_CREATE)
	    {	      
	      if (string_extension_compare(argv[ac], ".o"))
		infilename = argv[ac];
	      else if (string_extension_compare(argv[ac], 
						PANCODEOBJECT_EXTENSION))
		stringlist_addtotail(dependfilenames, argv[ac], 1); /*dup&add*/
	      else
		print_usage_and_exit(argv[0], 
				     "Invalid command line argument '%s'.\n",
				     argv[ac]);
	    }
	  else if (panotool_mode == MODE_TEST)
	    {
	      if (string_extension_compare(argv[ac], PANCODEOBJECT_EXTENSION))
		stringlist_addtotail(testfiles, argv[ac], 1); /*dup&add*/
	      else
		print_usage_and_exit(argv[0], 
				     "Invalid command line argument '%s'.\n",
				     argv[ac]);
	    }
	  else if (panotool_mode == MODE_EXTRACT)
	    {
	      if (string_extension_compare(argv[ac], PANCODEOBJECT_EXTENSION))
		infilename = argv[ac];
	      else
		print_usage_and_exit(argv[0], 
				     "Invalid command line argument '%s'.\n",
				     argv[ac]);
	    }
	}
    }
  
  /* Done parsing command line arguments.  Now verify sanity and run things. */

  if (panotool_mode == MODE_CREATE)
    {
      stringlistel *sel;

      if (!infilename || !string_extension_compare(infilename, ".o"))
	print_usage_and_exit(argv[0], "No input filename specified!\n", NULL);

      /* If no outfilename specified, generate it from the infilename */
      if (!outfilename && string_extension_compare(infilename, ".o"))
	{
	  int ifnlen, poextlen;
	  ifnlen = strlen(infilename);
	  poextlen = strlen(PANCODEOBJECT_EXTENSION);

	  outfilename = malloc(ifnlen-strlen(".o")+poextlen+1);
	  if (!outfilename)
	    { panPrint_error( "Couldn't malloc output filename!\n");
	    exit(-1); }
	  
	  strncpy(outfilename, infilename, ifnlen);
	  strncpy(outfilename+ifnlen-2, PANCODEOBJECT_EXTENSION, poextlen+1);
	}

      if (opt_verbose)
	{
	  printf("<%s> => <%s>\n", infilename, outfilename);
	  for (sel = stringlist_gethead(dependfilenames); sel; sel=sel->next)
	    printf("\tdepend: <%s>\n", sel->str);
	  for (sel = stringlist_gethead(incpath); sel; sel=sel->next)
	    printf("\tincpath: <%s>\n", sel->str);
	}

      stringlist_addtotail(incpath, ".", 1); /* dup&add */      

      panotoolCodeobj_create(outfilename, infilename, 
			     incpath, dependfilenames);
    }
  else if (panotool_mode == MODE_TEST)
    {
      int status=1;
      stringlistel *sel = stringlist_gethead(testfiles);
      
      if (!sel)
	print_usage_and_exit(argv[0], "No input files specified\n", NULL);

      while (sel)
	{
	  status = panotoolCodeobj_test(sel->str) && status;
	  sel = sel->next;
	}
      if (!status) exit(-1);
    }
  else if (panotool_mode == MODE_EXTRACT)
    {
      print_usage_and_exit(argv[0], "-X not yet supported.\n", NULL);
    }
  else
    {
      print_usage_and_exit(argv[0], "Must specify either -T, -C, or -X.\n",
			   NULL);
    }
  exit(0);
}


/* Creates a pan code object named "outfilename" from the 
 * object file named "infilename".  Returns 
 */
void
panotoolCodeobj_create(char *outfilename, char *infilename, 
		       stringlist *incpath, stringlist *dependfilenames)
{
  FILE     *outf;		/* Output file */
  int       infd;		/* Input file descriptor */
  panSwseg *binseg;		/* Duplicate in malloced memory of mmbinseg
				 * (needed if mmbinseg isn't 4-byte word
				 * aligned). */
  panSwseg *mmbinseg;		/* swseg containing contents of infilename 
				 *   (actually mmapped into place) */
  panSwseg *headseg;		/* Software segment for codeobj header */
  panSwseg *depseg;		/* Software segment for contents of
				 *   dependency section */
  panSwseg *symnameseg;		/* Symbolic name section which essentially
				 *   contains the source filename. */
  struct stat infstatbuf;	/* Stat info about input file */
  unsigned depsectsize, binsectsize, symnamesectsize, basenamelen, totalsize;
  char     *infbasename;	/* The basename of the input file. */
  panStream ps;			/* panStreams for reading/writing swsegs */
  int i;
  stringlistel *curdep;		/* Current dependancy */
  int numsect=0;		/* Number of sections */

  /* We will create a swseg chain containing the codeobject.
   * The layout is as follows: 
   *
   *     CODEOBJ HEADER      \___  In headseg (malloced)
   *     BINSECT HEADER      /  
   *     BINSECT BODY        >---  In binseg (mmapped from infilename)
   *     SYMNAMESECT         >---  In symnameseg (malloced)
   *     DEPSECT             >---  In depseg (malloced)   
   */

  /**********************************************************/
  /*********** BINCODE SECTION (except header) **************/
  /**********************************************************/

  /* Get the size of the input file */
  if (0 != stat(infilename, &infstatbuf))
    { perror("stat");
      panPrint_error("panotool: couldn't stat input file %s\n", 
		     infilename);  exit(-1);  }
  binsectsize = ADDR_ROUND_UP(infstatbuf.st_size);

  infd = open(infilename, O_RDONLY);
  if (-1 == infd)
    { perror("open");
      panPrint_error("panotool: couldn't open file %s for reading\n", 
		     infilename);  exit(-1);  }

  mmbinseg = panSwseg_create_byMmap(binsectsize,
				  PROT_READ, MAP_PRIVATE, infd, 0);
  REQUIRE_NOT_PRED(!mmbinseg, "creation of mmbinseg failed!");
  ADDTOKEYRING(mmbinseg);
  close(infd);			/* doesn't need to stay open to 
				 * keep the mapping */

#if 0 /* This doesn't seem to be needed */
  /* If the size on disk of mmbinseg isn't aligned, we need
   * to duplicate it in memory to prevent problems with hashing.
   */
  if (ADDR_NOT_ALIGNED_P(infstatbuf.st_size))
    {
      panStream pssrc, psdst;
      
      printf(">>> doing dup\n");
      binseg = panSwseg_create_byMalloc(binsectsize);
      REQUIRE_NOT_PRED(!binseg, "creation of binseg failed!");
      ADDTOKEYRING(binseg);

      panStream_init(&pssrc, mmbinseg, 0); 
      panStream_init(&psdst, binseg, 0); 
      panStream_copy(&psdst, &pssrc, binsectsize);
    }
  else
    {
      binseg = mmbinseg;
    }
#else 
  binseg = mmbinseg;
#endif  /* if 0 */

  /* Create code object header with 4 bytes at the end for the 
   * header of the BINCODE section.  Chain the bincode section
   * after the header. */
  headseg = panSwseg_create_byMalloc(panCodeobjHead_SIZE
				     + panCodeobjSectionHead_SIZE);

  ADDTOKEYRING(headseg);
  REQUIRE_NOT_PRED(!headseg, "creation of headseg failed!");

  panSwseg_chain(headseg, binseg);
  numsect += 1;

  /******************************************/
  /*********** SYMNAME SECTION **************/
  /******************************************/
   
  /* Create the symname section from the filename. */
  infbasename = string_file_basename(infilename);
  basenamelen = strlen(infbasename)+1;
  symnamesectsize = ADDR_ROUND_UP(basenamelen);

  symnameseg = panSwseg_create_byMalloc(symnamesectsize + 4);
  ADDTOKEYRING(symnameseg);
  REQUIRE_NOT_PRED(!symnameseg, "creation of symnameseg failed!");
  panSwseg_chain(binseg, symnameseg);
  numsect += 1;

  /* Write the header then the body of the symname section */
  panStream_init(&ps, symnameseg, 0); 
  panStream_write_net_uint16(&ps, panCodeobjSection_SYMBOLICNAME);
  panStream_write_net_uint16(&ps, symnamesectsize);
  panStream_write(&ps, (u_int8_t*)infbasename, basenamelen);
  for (i=0; i < (symnamesectsize-basenamelen); i++)
    panStream_write_uint8(&ps, 0);

  /******************************************/
  /************ DEPEND SECTION **************/
  /******************************************/

  depsectsize = 0;

  /* This contains just the header */
  depseg = panSwseg_create_byMalloc(4);
  ADDTOKEYRING(depseg);
  REQUIRE_NOT_PRED(!depseg, "creation of depseg failed!");

  /* Add swsegs for dependencies */
  for (curdep = stringlist_gethead(dependfilenames); 
       curdep; curdep = curdep->next)
    {
      panSwseg *newconame;
      newconame = panotoolCodeobj_getCodeobjName(curdep->str, incpath);
      if (!newconame)
	{ panPrint_error("panotool: couldn't load depenency.\n"); exit(-1); }
      panSwseg_chain(depseg, newconame);
      depsectsize += panCodeobjName_SIZE;
    }

  /* Only chain in the dependency section if it contains elements */
  if (depsectsize != 0)
    {
      /* Write the header */
      panStream_init(&ps, depseg, 0); 
      panStream_write_net_uint16(&ps, panCodeobjSection_DEPENDENCIES);
      panStream_write_net_uint16(&ps, depsectsize);

      numsect += 1;
      panSwseg_chain(binseg, depseg);
    }

  /***************************************************************/
  /************ MAIN HEADER AND BIN SECTION HEADER  **************/
  /***************************************************************/

  panStream_init(&ps, headseg, 0); 

  /* The magic number */
  panStream_write(&ps, (u_int8_t*)panCodeobjHead_MAGIC, 4); 
  /* Go back and fill in the name later */
  panStream_seekFwd(&ps, panCodeobjName_SIZE);
  /* The type and version information */
  panStream_write_net_uint16(&ps, panCodeobjHead_CODETYPE_IX86BIN);
  panStream_write_net_uint16(&ps, panCodeobjHead_CODEVERS_IX86BIN);
  panStream_write_net_uint16(&ps, panCodeobjHead_APITYPE_PAN);
  panStream_write_net_uint16(&ps, panCodeobjHead_APIVERS_PAN);
  /* The number of sections that follow */
  panStream_write_net_uint16(&ps, numsect);
  /* Padding junk */
  panStream_write_net_uint16(&ps, 0);
  
  /* Bin section header */
  panStream_write_net_uint16(&ps, panCodeobjSection_BINCODE);
  panStream_write_net_uint16(&ps, binsectsize);

  totalsize =  panCodeobjHead_SIZE+numsect*4
    +depsectsize+binsectsize+symnamesectsize;


  /**** Everything but the codeobj name should now have
   ****    been generated.  Calculate the hash then go back and store it.
   ****/
  {
    sha1_ctx sha1context;
    u_int8_t digestbuf[SHA1_SIZE];

    panStream_init(&ps, headseg, panCodeobjHead_STARTHASH_OFF); 
    sha1_init(&sha1context);
    panStream_sha1_update(&sha1context, &ps, 
			  totalsize-panCodeobjHead_STARTHASH_OFF); 
    sha1_final(digestbuf, &sha1context);
    REQUIRE_NOT_PRED(!panStream_eofP(&ps), "not at eof after sha1 hashing\n");

    panStream_init(&ps, headseg, panCodeobjHead_NAME_OFF); 
    panStream_write_net_uint16(&ps, panCodeobjName_TYPE_SHA1);
    panStream_write_net_uint16(&ps, panCodeobjName_VERS_SHA1);
    panStream_write(&ps, digestbuf, SHA1_SIZE);
  }

  /* Now write it all out to a file  */

  outf = fopen(outfilename, "w");
  if (!outf)
    { perror("fopen");
    panPrint_error("panotool: couldn't open file %s for writing\n", 
		   outfilename);  exit(-1);  }

  panStream_init(&ps, headseg, 0); 
  if (totalsize != panStream_write_toFile(outf, &ps, totalsize))
    {
      panPrint_error("panotool: error writing to %s\n", outfilename);
      exit(-1);
    }

  if (0 != fclose(outf))
    {
      perror("fclose");
      panPrint_warn("panotool: error closing %s\n", outfilename);
    }
}




/* Opens up a .po file named filename (after searching through
 * incpath) and returns a swseg containing the CodeobjName
 * region out of the file.  Returns NULL in the case of an error.
 */
panSwseg *
panotoolCodeobj_getCodeobjName(char *filename, stringlist *incpath)
{
  stringlistel *sel;
  char          fullfn[FILENAME_MAX+4];
  FILE         *inf;
  int           pathlen;
  panSwseg     *sws;
  u_int8_t      magicbuf[4];
  int           i;

  for (sel = stringlist_gethead(incpath); sel; sel = sel->next)
    {
      strncpy(fullfn, sel->str, FILENAME_MAX);
      pathlen = strlen(fullfn);
      fullfn[pathlen] = '/';
      fullfn[pathlen+1] = '\0';
      strncat(fullfn, filename, FILENAME_MAX-pathlen-2);
      inf = fopen(fullfn, "r");
      if (inf != NULL) break;
    }

  if (inf == NULL)
    {
      panPrint_warn("panotool: couldn't find %s in search path\n", filename);
      return NULL;
    }

  sws = panSwseg_create_byMalloc(panCodeobjName_SIZE);
  ADDTOKEYRING(sws);
  REQUIRE_NOT_PRED(!sws, "creation of sws failed!");
  
  /* Read and check the magic number */
  if (1 != fread(magicbuf, 4, 1, inf))
    {
      fclose(inf);
      panPrint_warn("panotool: premature end of file in %s\n", fullfn);
      return NULL;
    }
  for (i=0; i<4; i++) if (magicbuf[i] != panCodeobjHead_MAGIC[i])
    {
      fclose(inf);
      panPrint_warn("panotool: invalid magic number in file %s\n", fullfn);
      return NULL;
    }

  /* We can only read directly because the swseg was created with
   * panSwseg_create_byMalloc which creates a contiguous region */
  if (1 != fread(sws->data, panCodeobjName_SIZE, 1, inf))
    {
      fclose(inf);
      panPrint_warn("panotool: premature end of file in %s\n", fullfn);
      return NULL;
    }

  if (opt_verbose)
    {
      panPrint_msg("Loaded dependency file %s with code object name:\n\t[ ",
		   fullfn);
      for (i=0; i<panCodeobjName_SIZE; i++)
	panPrint_msg("%02X%s", (sws->data)[i], ((i-3)%4)?"":" ");
      panPrint_msg("]\n");
    }
  

  fclose(inf);

  /* If we've gotten this far, return sws */
  return sws;
}


/* Writes length bytes from a swseg stream into a FILE using fwrite.
 * Returns the number of bytes written (which may be less than
 * length if the stream is shorter than length).  Seeks forward the ps
 * stream by length.  Returns 0 on error.
 */
panStream_GEN_SCAN_FN(
      panStream_write_toFile(FILE *fp, panStream *ps, unsigned length),
      { if (!fp || !ps) return 0; },
      {	if (numscan != fwrite(ps->pos, 1, numscan, fp))
	{ perror("fwrite"); return 0; }
      })




/* Opens up the code object named by filename and verifies its
 * integrity.  Prints out information about the contents of
 * the code object to stdout.  Returns 1 on success or 0 if there's
 * an error while reading the code object.
 */
int
panotoolCodeobj_test(char *infilename)
{
  int       infd;		/* Input file descriptor */
  struct stat infstatbuf;	/* Stat info about input file */
  unsigned totalsectssize, totalsize;
  panSwseg  *cosws;		/* Software segment for code object */
  panStream  ps;		/* panStreams for reading/writing swsegs */
  int i, cursect;  
  panCodeobjHead cohead;	/* Code object header */
  panCodeobjSection *cosects;	/* Array of codeobject sections */

  panPrint_msg("%s:\n", infilename);

  /* Get the size of the input file */
  if (0 != stat(infilename, &infstatbuf))
    { perror("stat");
      panPrint_error("\t*** couldn't stat input file %s\n", 
		     infilename);  return(0);  }
  totalsize = infstatbuf.st_size;
  if (ADDR_NOT_ALIGNED_P(totalsize))
    { panPrint_error("\t*** invalid code object (size of is not 4-byte word aligned).\n"); return(0); }

  /* Now open and create a swseg by mmapping in the file */
  infd = open(infilename, O_RDONLY);
  if (-1 == infd)
    { perror("open"); panPrint_error("\t*** couldn't open file for reading\n"); return(0); }

  cosws = panSwseg_create_byMmap(totalsize,
				  PROT_READ, MAP_PRIVATE, infd, 0);
  REQUIRE_NOT_PRED(!cosws, "creation of binseg failed!");
  ADDTOKEYRING(cosws);
  close(infd);			/* doesn't need to stay open to 
				 * keep the mapping */
  
  /* Read in the code object's header */
  panStream_init(&ps, cosws, 0);
  if (!panCodeobjHead_read(&cohead, &ps, 1))
    { panPrint_warn("\t*** has invalid header\n"); return(0); }

  /* Read in the sections */
  cosects = malloc(sizeof(panCodeobjSection)*cohead.num_sections);
  REQUIRE_NOT_PRED(!cosects, "couldn't malloc memory for cosects!");
  if (!panCodeobjSection_readArray(cosects, &ps, cohead.num_sections))
    { panPrint_warn("\t*** error reading in sections\n"); return(0); }
  
  /* Check sizes */
  for (i=0, totalsectssize=0; i<cohead.num_sections; i++)
    totalsectssize += panCodeobjSectionHead_SIZE + cosects[i].length;
  if (totalsize != totalsectssize + panCodeobjHead_SIZE)
    { panPrint_warn("\t*** file size (%d) does not equal code object size (%d)\n", totalsize, totalsectssize + panCodeobjHead_SIZE); return(0); }
  
  /* If in verbose mode, print info about the code object */
  if (opt_verbose)
    {
      char  hashstrbuf[panCodeobjName_STRINGIFY_STRLEN];

      panCodeobjName_stringify(hashstrbuf, &(cohead.name));
      panPrint_msg("\tname:      [ %s ]\n",hashstrbuf);
      panPrint_msg("\tname type: %s\n", 
		   panCodeobjName_type2string(panCodeobjName_getType(&(cohead.name))));
      panPrint_msg("\tcode type: 0x%04x ==> %s\n", cohead.code_type,
		   panCodeobjHead_codetype2string(cohead.code_type));
      panPrint_msg("\tcode vers: 0x%04x ==> %d.%d\n", cohead.code_vers,
		   (cohead.code_vers)>>8, (cohead.code_vers)&0x03);
      panPrint_msg("\tapi type:  0x%04x ==> %s\n", cohead.api_type,
		   panCodeobjHead_apitype2string(cohead.api_type));
      panPrint_msg("\tapi vers:  0x%04x ==> %d.%d\n", cohead.api_vers,
		   (cohead.api_vers)>>8, (cohead.api_vers)&0x03);
      panPrint_msg("\tnum sects: %d\n", cohead.num_sections);
      panPrint_msg("\tsections:\n");

      /* Print section contents */
      for (cursect=0; cursect < cohead.num_sections; cursect++)
	{
	  if (cosects[cursect].type == panCodeobjSection_BINCODE)
	    {
	      FILE *filefile;
	      panPrint_msg("\t--- BinCode Section\n");
	      panPrint_msg("\t    type:    0x%04x ==> bincode\n", 
			   cosects[cursect].type);
	      panPrint_msg("\t    length:  %d\n", cosects[cursect].length);
	      panPrint_msg("\t    bintype: (feeding into \"file\")\n");
	      panPrint_msg("\t     ");
	      fflush(stdout);
	      filefile = popen("file -", "w");
	      if (filefile)
		{
		  panStream_dup(&ps, &(cosects[cursect].body));
		  panStream_write_toFile(filefile, &ps,
					 cosects[cursect].length<100 
					 ? cosects[cursect].length : 100);
		  pclose(filefile);
		}
	      else
		panPrint_warn("*** couldn't popen file command\n");
	    }
	  else if (cosects[cursect].type == panCodeobjSection_DEPENDENCIES)
	    {
	      panCodeobjName con;
	      panPrint_msg("\t--- Dependencies Section\n");
	      panPrint_msg("\t    type:    0x%04x ==> dependencies\n", 
			   cosects[cursect].type);
	      panPrint_msg("\t    length:  %d\n", cosects[cursect].length);
	      panStream_dup(&ps, &(cosects[cursect].body));
	      if ((cosects[cursect].length % panCodeobjName_SIZE))
		{ panPrint_error("\t    *** dependency section size invalid\n");
		return(0); }
	      for (i=0; i<(cosects[cursect].length)/panCodeobjName_SIZE; i++)
		{
		  panCodeobjName_read(&con, &ps, 1);
		  panCodeobjName_stringify(hashstrbuf, &con);
		  panPrint_msg("\t    depend:  [ %s ]\n",
			       hashstrbuf);
		}
	    }
	  else if (cosects[cursect].type == panCodeobjSection_SYMBOLICNAME)
	    {
	      panPrint_msg("\t--- Symbolic Name Section\n");
	      panPrint_msg("\t    type:    0x%04x ==> symbolicname\n", 
			   cosects[cursect].type);
	      panPrint_msg("\t    length:  %d\n", cosects[cursect].length);
	      panPrint_msg("\t    symname: \"");
	      panStream_dup(&ps, &(cosects[cursect].body));
	      panStream_write_toFile(stdout, &ps, cosects[cursect].length),
	      panPrint_msg("\"\n");
	    }
	  else
	    {
	      panPrint_msg("\t--- Unknown Section\n");
	      panPrint_msg("\t    type:    0x%04x ==> unknown\n", 
			   cosects[cursect].type);
	      panPrint_msg("\t    length:  %d\n", cosects[cursect].length);
	    }
	}
    }

  /* Verify the code object's integrity */
  panStream_init(&ps, cosws, 0);
  if (!panCodeobjName_verify(&(cohead.name), &ps, totalsize))
    { panPrint_warn("\t*** error verifying hash name\n"); return(0); }
  panPrint_msg("\thash verification succeeded\n");

  return(1);
}

