/* $Header: /home/dl2/nygren/cvs/pan/src/test/swseg/swseg_test.c,v 1.3 1997/12/10 06:58:47 nygren Exp $
 *
 * FILE:    swseg_test.c
 * AUTHORS: Erik Nygren (nygren@mit.edu)
 *
 * Code for testing software segment data structures and routines.
 * See PAN/doc/software-segments.txt for a detailed description
 * of software segments.
 *
 */

#include <common/pan-copyright.h>
#include <common/sysconfig.h>
#include <common/swseg.h>
#include <common/swsstream.h>

#define TEST_PRED(pred,printstatement) \
   do { if ((pred)) { printf("%s:%d: ", __FILE__, __LINE__); \
                      printstatement; \
		      exit(-1); } } while(0)		

void test_swsring(void);

void
main(int argc, char **argv)
{
  panSwseg *sa1, *sa2, *sa3;
  panSwseg *sb1, *sb2, *sb3;
  panSwseg *so1;		/* Overlay */
  panStream psa, psb, pso;
  u_int8_t  buf[8*3];
  unsigned u;
  int i;


  /* Create three 8 byte segments (sa1,sa2,sa3) and chain them 
   * together and fill them with increasing byte values.
   */
  sa1 = panSwseg_create_byMalloc(8);
  sa2 = panSwseg_create_byMalloc(8);
  sa3 = panSwseg_create_byMalloc(8);
  panSwseg_chain(sa1, sa2);
  panSwseg_chain(sa2, sa3);

  /* Test panSwseg_getChainLength and panSwseg_getBytesInChain */

  TEST_PRED((panSwseg_getChainLength(sa1) != 3),
	    printf("got chain length of %d for sa1\n", 
		   panSwseg_getChainLength(sa1)));
  TEST_PRED((panSwseg_getChainLength(sa3) != 1),
	    printf("got chain length of %d for sa3\n", 
		   panSwseg_getChainLength(sa3)));

  TEST_PRED((panSwseg_getBytesInChain(sa1) != 8*3),
	    printf("got bytes in chain = %d for sa1\n", 
		   panSwseg_getBytesInChain(sa1)));
  TEST_PRED((panSwseg_getBytesInChain(sa3) != 8*1),
	    printf("got bytes in chain = %d for sa3\n", 
		   panSwseg_getBytesInChain(sa3)));

  /* Test panSwstream_getBytesRemaining */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  TEST_PRED(panStream_getBytesRemaining(&psa) != 8*3,
	      printf("error with getBytesRemaining\n"));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test writing bytes from a buffer (*/
  for (i=0; i<8*3; i++) buf[i] = 8*3-i;
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  TEST_PRED(8*3 != panStream_write(&psa, buf, 8*3), printf("short write\n"));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned bytes from the start */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=0; i<8*3; i++)
    TEST_PRED(panStream_read_uint8(&psa) != (u_int8_t)(8*3-i),
	      printf("error at stream read u8 1 at i=%d\n",i));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Now write out bytes to the three chained swsegs.
   * Each byte contains its zero-indexed byte offset into the chain. 
   */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=0; i<8*3; i++)
    panStream_write_uint8(&psa, (u_int8_t)i);
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading bytes into a buffer */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  TEST_PRED(8*3 != panStream_read(buf, &psa, 8*3), printf("short read\n"));
  for (i=0; i<8*3; i++)
    TEST_PRED(buf[i] != (u_int8_t)i,
	      printf("error at stream read u8 1 at i=%d\n",i));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned bytes from an offset */
  panStream_init(&psa, sa1, 4);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=4; i<8*3; i++)
    TEST_PRED(panStream_read_uint8(&psa) != (u_int8_t)i,
	      printf("error at stream read u8 2 at i=%d\n",i));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u16's from the start (msb-first) */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=0; i<8*3; i+=2)
    {
      u_int16_t x = panStream_read_msbfirst_uint16(&psa);
      TEST_PRED((((x>>8)&0xff) != (u_int16_t)i)
		 || (((x>>0)&0xff) !=  (u_int16_t)(i+1)),
		printf("error@streamread u16 3 at i=%d (got 0x%04x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u16's from the start (lsb-first) */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=0; i<8*3; i+=2)
    {
      u_int16_t x = panStream_read_lsbfirst_uint16(&psa);
      TEST_PRED((((x>>0)&0xff) != (u_int16_t)i)
		 || (((x>>8)&0xff) !=  (u_int16_t)(i+1)),
		printf("error@streamread u16 4 at i=%d (got 0x%04x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u16's from an offset (lsb-first) */
  panStream_init(&psa, sa1, 4);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=4; i<8*3; i+=2)
    {
      u_int16_t x = panStream_read_lsbfirst_uint16(&psa);
      TEST_PRED((((x>>0)&0xff) != (u_int16_t)i)
		 || (((x>>8)&0xff) !=  (u_int16_t)(i+1)),
		printf("error@streamread u16 5 at i=%d (got 0x%04x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u16's from a larger starting offset (lsb-first).
   * Causes init to have to seek forward a segment.  */
  panStream_init(&psa, sa1, 12);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=12; i<8*3; i+=2)
    {
      u_int16_t x = panStream_read_lsbfirst_uint16(&psa);
      TEST_PRED((((x>>0)&0xff) != (u_int16_t)i)
		 || (((x>>8)&0xff) !=  (u_int16_t)(i+1)),
		printf("error@streamread u16 6 at i=%d (got 0x%04x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u16's from a seeked offset that
   * is segment-aligned.  Tests seekfwd and eofP.  */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  TEST_PRED(8 != panStream_seekFwd(&psa, 8), printf("seek not complete\n"));
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=8; i<8*3; i+=2)
    {
      u_int16_t x = panStream_read_lsbfirst_uint16(&psa);
      TEST_PRED((((x>>0)&0xff) != (u_int16_t)i)
		 || (((x>>8)&0xff) !=  (u_int16_t)(i+1)),
		printf("error@streamread u16 7 at i=%d (got 0x%04x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u16's from a seeked offset that
   * is in the second segment.  Tests seekfwd and eofP.  */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  TEST_PRED(12 != panStream_seekFwd(&psa, 12), printf("seek not complete\n"));
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=12; i<8*3; i+=2)
    {
      u_int16_t x = panStream_read_lsbfirst_uint16(&psa);
      TEST_PRED((((x>>0)&0xff) != (u_int16_t)i)
		 || (((x>>8)&0xff) !=  (u_int16_t)(i+1)),
		printf("error@streamread u16 8 at i=%d (got 0x%04x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u32's from the start (lsb-first) */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=0; i<8*3; i+=4)
    {
      u_int32_t x = panStream_read_lsbfirst_uint32(&psa);
      TEST_PRED((((x>>0)&0xff) != (u_int32_t)i)
		 || (((x>>8)&0xff) !=  (u_int32_t)(i+1))
		 || (((x>>16)&0xff) !=  (u_int32_t)(i+2))
		 || (((x>>24)&0xff) !=  (u_int32_t)(i+3)),
		printf("error@streamread u32 9 at i=%d (got 0x%08x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Test reading unsigned u32's from the start (msb-first) */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(panStream_eofP(&psa), printf("at eof already\n"));
  for (i=0; i<8*3; i+=4)
    {
      u_int32_t x = panStream_read_msbfirst_uint32(&psa);
      TEST_PRED((((x>>24)&0xff) != (u_int32_t)i)
		 || (((x>>16)&0xff) !=  (u_int32_t)(i+1))
		 || (((x>>8)&0xff) !=  (u_int32_t)(i+2))
		 || (((x>>0)&0xff) !=  (u_int32_t)(i+3)),
		printf("error@streamread u32 10 at i=%d (got 0x%08x)\n",i,x));
    }
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));

  /* Create two 12 byte segments (sb1,sb2) and chain them together. */
  sb1 = panSwseg_create_byMalloc(12);
  sb2 = panSwseg_create_byMalloc(4);
  sb3 = panSwseg_create_byMalloc(8);
  panSwseg_chain(sb1, sb2);
  panSwseg_chain(sb2, sb3);

  /* Copy bytes from sa to sb and then compare */
  panStream_init(&psa, sa1, 0);
  panStream_init(&psb, sb1, 0);
  panStream_copy(&psb, &psa, 8*3);
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof of src\n"));
  TEST_PRED(!panStream_eofP(&psb), printf("not at eof of dest\n"));
  /* Test reading unsigned bytes from the start */
  panStream_init(&psb, sb1, 0);
  for (i=0; i<8*3; i++)
    TEST_PRED(panStream_read_uint8(&psb) != (u_int8_t)i,
	      printf("error at copy verify 1 at i=%d\n",i));

  /* Copy bytes from sa to sb and then compare */
  panStream_init(&psa, sa1, 4);
  panStream_init(&psb, sb1, 0);
  panStream_copy(&psb, &psa, 8*3-4);
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof of src\n"));
  TEST_PRED(panStream_eofP(&psb), printf("at eof of dest\n"));
  TEST_PRED(4 != panStream_seekFwd(&psb, 4), printf("seek not complete\n"));
  TEST_PRED(!panStream_eofP(&psb), printf("not at eof of dest\n"));
  /* Test reading unsigned bytes from the start */
  panStream_init(&psb, sb1, 0);
  for (i=4; i<8*3; i++)
    TEST_PRED(panStream_read_uint8(&psb) != (u_int8_t)i,
	      printf("error at copy verify 1 at i=%d\n",i));

  /* Test panStream_getState */
  panStream_init(&psa, sa1, 0);
  TEST_PRED(12 != panStream_seekFwd(&psa, 12), printf("seek not complete\n"));
  TEST_PRED(panStream_getState(&psa, &u) != sa2,
	    printf("getstate returned wrong segment\n"));
  TEST_PRED(u != 4, printf("getstate returned wrong offset\n"));

  /* Try a seek that goes past the end of the segment chain */
  panStream_init(&psa, sa1, 4);
  TEST_PRED(panStream_eofP(&psa), printf("at eof\n"));
  TEST_PRED(3*8-4 != panStream_seekFwd(&psa, 3*8+10),
	    printf("seek returned wrong value\n"));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));
  TEST_PRED(0 != panStream_seekFwd(&psa, 3*8+10),
	    printf("seek returned wrong value (should have returned 0)\n"));
  TEST_PRED(!panStream_eofP(&psa), printf("not at eof\n"));


  /*************** Test overlay creation code *************/
  so1 = panSwseg_overlay(sa1, 4, 3*8-8);
  TEST_PRED(NULL == so1, printf("overlay creation returned NULL\n"));

  /* Test reading unsigned bytes from the start */
  panStream_init(&pso, so1, 0);
  TEST_PRED(panStream_eofP(&pso), printf("at eof already\n"));
  for (i=4; i<8*3-4; i++)
    TEST_PRED(panStream_read_uint8(&pso) != (u_int8_t)i,
	      printf("error at stream read u8 of overlay at i=%d\n",i));
  TEST_PRED(!panStream_eofP(&pso), printf("not at eof\n"));

  /* Test panSwseg_getChainLength and panSwseg_getBytesInChain */

  TEST_PRED((panSwseg_getChainLength(so1) != 3),
	    printf("got chain length of %d for so1\n", 
		   panSwseg_getChainLength(so1)));

  TEST_PRED((panSwseg_getBytesInChain(so1) != 8*3-8),
	    printf("got bytes in chain = %d for so1\n", 
		   panSwseg_getBytesInChain(so1)));

  test_swsring();

  printf("passed.\n");
  exit(0);
}

/* Tests swseg ring code */
void 
test_swsring(void)
{
  panSwseg *s1, *s2, *s3, *s4, *s5, *s6, *s7, *s8;
  panSwsegRing *ring;

  s1 = panSwseg_create_byMalloc(8);
  s2 = panSwseg_create_byMalloc(8);
  s3 = panSwseg_create_byMalloc(8);
  s4 = panSwseg_create_byMalloc(8);
  s5 = panSwseg_create_byMalloc(8);
  s6 = panSwseg_create_byMalloc(8);
  s7 = panSwseg_create_byMalloc(8);
  s8 = panSwseg_create_byMalloc(8);

  ring = panSwsegRing_create(3);
  TEST_PRED(0 != panSwsegRing_numEntries(ring),
	   printf("new ring contains a non-zero number of entries\n"));
  
  /* Add five swsegs to the ring.  This should force it to make a
   * second segment... */

  panSwsegRing_add(ring, s1);
  TEST_PRED(1 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  panSwsegRing_add(ring, s2);
  TEST_PRED(2 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  panSwsegRing_add(ring, s3);
  TEST_PRED(3 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  panSwsegRing_add(ring, s4);
  TEST_PRED(4 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  panSwsegRing_add(ring, s5);
  TEST_PRED(5 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  TEST_PRED(1 != panSwsegRing_remove(ring, s3, 0),
	   printf("panSwsegRing_remove return val != 1\n"));
  TEST_PRED(4 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  TEST_PRED(1 != panSwsegRing_remove(ring, s4, 0),
	   printf("panSwsegRing_remove return val != 1\n"));
  TEST_PRED(3 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  /* Try removing a swseg already removed... */
  TEST_PRED(0 != panSwsegRing_remove(ring, s4, 0),
	   printf("panSwsegRing_remove return val != 0\n"));
  TEST_PRED(3 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  /* Add two more back (from the free list this time) */

  panSwsegRing_add(ring, s6);
  TEST_PRED(4 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  panSwsegRing_add(ring, s7);
  TEST_PRED(5 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  /* Add one more now that the free list is done with */

  panSwsegRing_add(ring, s8);
  TEST_PRED(6 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));

  /* And clear it all... */
  panSwsegRing_clearAll(ring);
  TEST_PRED(0 != panSwsegRing_numEntries(ring),
	   printf("new ring contains invalid number of entries\n"));
}
